package CoreSim::ALT::ALTHMetrics;

use strict;

sub new(){
	my $invocant = shift;
	my ($topology, $altHTopology,$ipUtility, $aggregationIndex,$outputFilePath, $_DEBUG) = @_;
	my $class = ref($invocant) || $invocant;
	
	my $numberOfQueriesHash={};
	
	my $self={
		"topology" => $topology, 
		"altHTopology" => $altHTopology, 
		"ipUtility" => $ipUtility,
		"aggregationIndex" => $aggregationIndex,
		"numberOfQueriesHash" => $numberOfQueriesHash,
		"outputFilePath" => $outputFilePath,
		"_DEBUG" => $_DEBUG,
	};
	
	
	bless($self, $class);
	
	return $self;
}

sub getALTLookupDelayFromALTRouter(){
	my $self= shift;
	my ($altRouterPrefix, $searchedIP) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	my $altHTopology = $self->{"altHTopology"};
	
	my $altRouter = $altHTopology->getALTRouterForPrefix($altRouterPrefix);
	
	#the current router is responsible for the IP
	if ($ipUtility->ipInsidePrefix($searchedIP, $altRouter->getPrefix()  )){
		printf("The first ALTRouter %s is responsible for the IP %s! There is no search latency!\n", $altRouter->getPrefix(),$searchedIP)
			if ($self->{"_DEBUG"});
		return 0;
	}
	
	my ($responsibleALTRouter, $path) = &getResponsibleALTRouter($self, $altRouter, $searchedIP);

	#debug output: used to see the prefix path followed
	if ($self->{"_DEBUG"}>=2){
		printf("The ALT prefix path is:");
		foreach my $altR (@$path){
			printf("%s ", $altR->getPrefix());
		}
		printf("\n");
	}

	#if the responsibleALTRouter is found
	if (defined($responsibleALTRouter)){
		
		#check to see if the destination router is a virtual router or an EndorsedByChildRouter (these are not supposed to get packets)
		if (!$responsibleALTRouter->isVirtualRouter() && !$responsibleALTRouter->isEndorsedByChild() ){
			
			my $nbOfPeers = scalar(@$path) -1;
			my $pathLength = scalar(@$path);
			my $pathLatency;
			my $nodeLatencyList;
			my $notAnsweredXMLRPCQueries;
			
			#update the numberOfQueriesHash
			my $numberOfQueriesHash = $self->{"numberOfQueriesHash"};
			&updateNumberOfQueriesHash($numberOfQueriesHash, $path);
			
			#check to see that the path length is higher than 1
			if ($pathLength>1){
			
				#compute the latency
				($pathLatency, $nodeLatencyList, $notAnsweredXMLRPCQueries ) = &computePathLatency($self, $path);
				
				printf("\nIn order to compute the latency implied by the search for IP %s starting at %s a number of %s paths have been checked\n", 
						$searchedIP, $altRouter->getPrefix(), $nbOfPeers) if ($self->{"_DEBUG"});
#				printf("For %s paths the latency is unknown due to failed XMLRPC searches\n",
#						 $notAnsweredXMLRPCQueries) if ($self->{"_DEBUG"});
			
			#if it is equal to 1 then the start router is responsible for the searchedIP and the latency is 0
			}else{
				$pathLatency = 0;
				$nodeLatencyList = [];
			}
			
			
			#format the output of the function
			my $status="ok";
			my $destPOP = $responsibleALTRouter->getPOPNumber();
			my $destAS = $responsibleALTRouter->getAS();
			my $destPrefix = $responsibleALTRouter->getPrefix();
			my $lowerBoundLatency = $pathLatency;
			
			return ($status, $destPOP, $destAS, $destPrefix, $lowerBoundLatency, $nbOfPeers, $nodeLatencyList);
			
		#if it is a virtual router then the packet is discarded by the ITR
		}else{
			my $status="discard";
			my $destPrefix = $responsibleALTRouter->getPrefix();
			my $destPOP =$responsibleALTRouter->getPOPNumber();
			my $destAS = $responsibleALTRouter->getAS();
			return ($status, $destPOP, $destAS, $destPrefix);
		}
		
	}else{
		#or issue a waring message
		
		printf("(EE) The altRouter responsible for the IP %s was not found!\n", $searchedIP);
		return -1;
	}
	


}

sub getResponsibleALTRouter(){
	my ($self, $altRouter, $searchedIP) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	
	my @path;
	push(@path, $altRouter);
	
	my $endLoop = 0;
	my $responsibleALTRouter=undef;
	
	#special case for when the startAltRouter is responsible for the searchedIP
	if ($ipUtility->ipInsidePrefix($searchedIP, $altRouter->getPrefix() )){
		
		return($altRouter, \@path);
		
	}
	
	#search for the router responsible for the searchedIP
	while (!$endLoop){
		
		my $nextHopALTRouter =   $altRouter->findNextHopForIP($searchedIP);
		
		#if the returned nextHopALTRouter is the same as the current router that means that the responsible altRuter is found
		if ($nextHopALTRouter->getPrefix() eq $altRouter->getPrefix()){
			
			#this is a double check to see if the router's prefix does include the searchedIP
			if ($ipUtility->ipInsidePrefix($searchedIP, $nextHopALTRouter->getPrefix() )){
		
				$responsibleALTRouter = $nextHopALTRouter;
				$endLoop =1 ;
				
			}else{
				
				printf("(EE) Something is wrong in the ALT search algorithm!\n");
				$endLoop=1;
				
			}
		#add the new router to the path
		}else{
			push(@path, $nextHopALTRouter);
			
		}
		
		$altRouter = $nextHopALTRouter;
	}
	
	return ($responsibleALTRouter, \@path);
	
}

sub computePathLatency(){
	
		my ($self, $path) = @_;
			
		my $topology = $self->{"topology"};
		
		#compute the latency
		my $totalLatency = 0;
		
		my $notAnsweredXMLRPCQueries = 0;
		my @nodeLatencyList;
		
		#add the first element at the end of the path in order to compute the RETURN LATENCY also 
		push(@$path, $path->[0]);
		push(@nodeLatencyList, $path->[0]->getPOPIP());
		
		my $pathLength = scalar(@$path);
		
		#used to keep the IP elected when a pass from /8 to /3 is done
		my $chosenAnycastIP;
		
		for(my $i=1;$i<$pathLength; $i++){
			
			my $router1 = $path->[$i-1];
			my $router2 = $path->[$i];
			
			
			#if both routers are root layers the chosenRootAnycastIP must be used. Practically one of the anycast /3 routers
			#is chosen by the /8 before and it must be used to connect to one of the two anycast routers from the next /3 in the 
			#ALT path
			if ($router1->isRootRouter() && $router2->isRootRouter() ){
				
				my $anycastLatency;
				($anycastLatency, $chosenAnycastIP) = &anycastLatency($self, $chosenAnycastIP, $router2);
				$totalLatency += $anycastLatency;
				
				#add to nodeLatencyList
				push(@nodeLatencyList, $anycastLatency);
				push(@nodeLatencyList, $chosenAnycastIP);
			
			#if router1 is root router and router2 is not that means that we are on the descending path and we don't need anycast anymore
			}elsif ( $router1->isRootRouter() ) {
				
				my $latencyQueryResponse;
				$latencyQueryResponse = $topology->getLatencyBetweenPOPIPs1($chosenAnycastIP, $router2->getPOPIP());
				$totalLatency +=$latencyQueryResponse;
				
				#add to nodeLatencyList
				push(@nodeLatencyList, $latencyQueryResponse);
				push(@nodeLatencyList, $router2->getPOPIP());
				
			#if router2 is a root router and router1 is not that means we are on the ascending part and we need anycast for the link
			#between the /8 and the /3
			}elsif( $router2->isRootRouter()){
				
				my $anycastLatency;
				($anycastLatency, $chosenAnycastIP) = &anycastLatency($self, $router1->getPOPIP, $router2);
				$totalLatency += $anycastLatency;
				
				#add to nodeLatencyList
				push(@nodeLatencyList, $anycastLatency);
				push(@nodeLatencyList, $chosenAnycastIP);
				
			#otherwise use the the XMLRPC query
			}else{
				my $routerIP1 = $router1->getPOPIP();
				my $routerIP2 = $router2->getPOPIP();
				
				#use the topology module to find the latency between the POPs representing the routers
				my $latencyQuery = $topology->getLatencyBetweenPOPIPs1($routerIP1, $routerIP2);
				
				if ($latencyQuery == -1){
					$notAnsweredXMLRPCQueries++;
					printf("(ALTHM)(W) The latency query for %s and %s failed! The prefixes implied: %s and %s\n", 
							$routerIP1, $routerIP2, $router1->getPrefix(), $router2->getPrefix()) if($self->{"_DEBUG"}>=2);
				}else{
					$totalLatency += $latencyQuery;
				}
				
				#add to nodeLatencyList
				push(@nodeLatencyList, $latencyQuery);
				push(@nodeLatencyList, $router2->getPOPIP());
					
			}
		}
		
		
		return ($totalLatency, \@nodeLatencyList, $notAnsweredXMLRPCQueries );
		
}

#"simulates anycast" by taking the minimum latency between the two routers
sub anycastLatency(){
	my ($self, $ip, $router) = @_;
	
	my $topology = $self->{"topology"};
	
	my $ipList;
	my $chosenIP;
	
	if ($router->isRootRouter()){
		$ipList = $router->getRootIPList();
	}else{
		push(@$ipList, $router->getPOPIP());
	}
	
	my $minLatency = $topology->getLatencyBetweenPOPIPs1($ip, $ipList->[0]);
	$chosenIP = $ipList->[0] ;
	
	foreach my $anycastIP (@$ipList){
		my $latencyQuery = $topology->getLatencyBetweenPOPIPs1($ip, $anycastIP);
		
		if ($latencyQuery< $minLatency){
			$minLatency = $latencyQuery;
			$chosenIP = $anycastIP;

		}
		
	}

	return ($minLatency, $chosenIP);
}

sub updateNumberOfQueriesHash(){
	my ($numberOfQueriesHash, $path) = @_;
	
	my $pathLength = scalar(@$path);
	
	for (my $i=0;$i<$pathLength; $i++){
		$numberOfQueriesHash->{$path->[$i]->getPrefix()} += 1;
	}
}

sub DESTROY(){
	my $self=shift;
	
	my $outputFilePath = $self->{"outputFilePath"};
	my $outputFile = "$outputFilePath/altRoutersStats.txt";
	my $numberOfQueriesHash = $self->{"numberOfQueriesHash"};
	my $altHTopology = $self->{"altHTopology"};
	my @altPrefixes = keys %$numberOfQueriesHash;
	
	open(OFILE, "> $outputFile") or die printf("Error encountered while opening the output file for the ALT node stats %s:%s",
			$outputFile, $!);
	
	printf(OFILE "ALT Prefix\tALT Layer\tPOP\tAS\tQueries Routed\n");
	
	foreach my $altPrefix (@altPrefixes){
		my $altRouter = $altHTopology->getRouterWithPrefix($altPrefix);
		my $pop = $altRouter->getPOPNumber();
		my $as =  $altRouter->getAS();
		my $altLevel = $altRouter->getALTLevel();
		my $queriesCount = $numberOfQueriesHash->{$altPrefix};
		
		if (defined($pop)){
			printf(OFILE "%s\t%s\t%s\t%s\t%s\n", $altPrefix, $altLevel, $pop, $as, $queriesCount);
		
		#the root and virtual routers 
		}else{
			printf(OFILE "%s\t%s\t-1\t-1\t%s\n", $altPrefix,$altLevel, $queriesCount);
		}
		
	}	
	
	close(OFILE);
}
#return 1 at the end
1