package CoreSim::Chord::ChordRing;

use strict;

#use CoreSim::Topology;
use CoreSim::Chord::ChordNode; 
use CoreSim::Chord::ChordList;
use CoreSim::Chord::FingerTable;

sub new(){
	my $invocant = shift;
	
	my ($topology, $maxNodesInChordRing, $maxNodesInFingerTable, $readFingerTablesFromFile, $startPosition, $endPosition,
		$sortedAndFilteredPrefixListFile, $ipUtility, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	#create the chordList
	my $chordList = CoreSim::Chord::ChordList->new();
	
	my $self={
		"topology" => $topology,
		"chordList" => $chordList,
		"maxNodesInFingerTable" => $maxNodesInFingerTable,
		"maxNodesInChordRing" => $maxNodesInChordRing,
		"readFingerTablesFromFile" => $readFingerTablesFromFile,
		"startPosition" => $startPosition,
		"endPosition" => $endPosition,
		"sortedAndFilteredPrefixListFile" => $sortedAndFilteredPrefixListFile,
		"ipUtility" => $ipUtility,
		"_DEBUG" => $_DEBUG,
	};
	
	&createChordRing($self);
	
	bless($self, $class);
	
	return $self;
}

#deals at a "high level" with the creation of the ChordRing
sub createChordRing(){
	my $self = shift;
	
	my $topology = $self->{"topology"};
	my $startPosition = $self->{"startPosition"};
	my $endPosition = $self->{"endPosition"};
	my $ipUtility = $self->{"ipUtility"};
	
	my $filteredPrefixList;
	my $sortedAndFilteredPrefixList;
	
	#create the list of filtered prefixes if a file is not given
	if ( !defined($self->{"sortedAndFilteredPrefixListFile"}) ){ 
		
		printf("Creating the prefix list..\n");
		#the prefixes that need to be added to the ChordCircle
		my $prefixList = $topology->getPrefixList();
		
		#filter out specific prefixes
		printf("Starting to filter out specific prefixes\n");
		$filteredPrefixList = &eliminateSpecificPrefixes2($self,$prefixList);
		
		#sort prefix list
		printf("Sorting the filtered prefix list\n");
		$sortedAndFilteredPrefixList = &sortPrefixList($filteredPrefixList, $ipUtility);
		
		printf("The sorted and filtered prefix list \n @{$sortedAndFilteredPrefixList}\n");
		printf("The number of prefixes used now is %s\n", scalar(@$sortedAndFilteredPrefixList));
		
		$self->{"sortedAndFilteredPrefixList"} = $sortedAndFilteredPrefixList;
		
	#or read the list of prefixes from the given file
	}else{
		
		printf("Reading the prefix list from %s..\n",$self->{"sortedAndFilteredPrefixListFile"} );
		$sortedAndFilteredPrefixList = &readSortedAndFilteredPrefixesFromFile($self->{"sortedAndFilteredPrefixListFile"});
		$self->{"sortedAndFilteredPrefixList"} = $sortedAndFilteredPrefixList;
		
	}
	
	#add all the prefixes to the ChordRing
	&addPrefixes($self, $sortedAndFilteredPrefixList);
	
	#create the fingerTables
	&createFingerTables($self, $startPosition, $endPosition);
	
	&testChordRing($self, 2471755775);
}

#takes the list of sorted prefixes and calls the addNode method which creates ChordNodes and adds them to ChordRing
sub addPrefixes(){
	
	my $self = shift;
	my $sortedPrefixList = shift;
	
	#the last inserted node. It starts with undef
	my $successor;
	
	printf("Starting to add the prefixes to the ChordRing...\n");
	my $count =0;
	my $numberOfPrefixes = scalar(@$sortedPrefixList);
	
	#add the prefixes
	foreach my $prefix (@$sortedPrefixList){

		my $newNode = &addNode($self, $prefix, $successor);
		$successor = $newNode;	
		$count++;
		
		printf("The number of prefixes added until now to chord = %d from a total of %s\n", $count,$numberOfPrefixes ) 
			if ( ($count %10000 == 0)&&($self->{"_DEBUG"}) );
	}
	
	printf("Finished with the insertion of prefixes into ChordRing!\n") if ($self->{"_DEBUG"});
}

#creates a ChordNode and inserts it into ChordRing. It is called by the addPrefixes method
sub addNode(){
	
	my $self = shift;
	my ($prefix, $successor) = @_;
	
	my $topology = $self->{"topology"};
	my $chordList = $self->{"chordList"};
	my $maxNodesInFingerTable = $self->{"maxNodesInFingerTable"};
	my $maxNodesInChordRing = $self->{"maxNodesInChordRing"};
	my $ipUtility = $self->{"ipUtility"};
	
	#construct the node
	my $as = $topology->getPrefixToAS($prefix);
	my $borderPOP = $topology->getBorderPOPForAS($as);
	my $borderPOPIP = $topology->getIPForPOP($borderPOP);
	
	printf("(EE) There was an error while creating the node for prefix %s\n", $prefix)
			if ((!defined($as)) || (!defined($borderPOP)) || !(defined($borderPOPIP))) ;
	
	my $chordID = $ipUtility->getChordIDForPrefix($prefix);
	
	#now we determine the successor (in order to be able to insert the node into ChordRing) and the associated keys
	#The is done the simple way: each node has a single key (the prefix) and because the prefixes will be sorted the the successor 
	#will always be the previously inserted node(we go from last to first).
	
	my $key = $prefix; 
		
	#creates the finger table but without inserting the nodes for each finger. Each finger contains only the "start"" and "interval" values.
	my $fingerTable = CoreSim::Chord::FingerTable->new($chordID,$maxNodesInFingerTable, $maxNodesInChordRing);
	
	my $newNode = CoreSim::Chord::ChordNode->new($chordID,$borderPOPIP,$borderPOP,$as,undef,undef,$key, $fingerTable);
	

	#we add the node in the ChordList ( but with an undef successor for the first one (the ChordList will deal with that))
	$chordList->insertNode($newNode, $successor);
	return $newNode;
	
}

sub createFingerTablesProcess(){
	my $self = shift;

	my $chordList = $self->{"chordList"};
	my $startPosition = $self->{"startPosition"};
	my $endPosition = $self->{"endPosition"};
	
	my $firstNode = $chordList->getFirstNode();
	my $maxNodesInChordRing = $self->{"maxNodesInChordRing"};
	my $fingerTableSize = $self->{"maxNodesInFingerTable"};
	
	if (!defined($endPosition)){ 
		$endPosition = $self->{"chordList"}->getNumberOfNodesInChordList() - 1;
		
		printf("The endPosition was not defined, using as default the last position in the ChordList: %s\n", $endPosition) 
				if($self->{"_DEBUG"});
	}

	printf("\nCreateFingerTablesProcess. Was called with parameters: startPosition:%s and endPosition:%s\n",$startPosition, $endPosition) 
			if ($self->{"_DEBUG"}>=2);
	
	my $node =$firstNode;
	my $distance = $startPosition;
	
	#find the first node in the list of nodes which will be processed by this thread
	while($distance){
		$node = $node->getSuccessor();
		$distance--;
	}
	
	#pass throught the node list (from startPosition to the endPosition)
	for (my $i=$startPosition; $i<=$endPosition; $i++){
		
		#get the finger table and the chordID for the reference node(for which we compute the finger table)
		my $fingerTable = $node->getFingerTable();
		my $referenceChordID = $node->getChordID();
		
		#start the chord ring sweep with the successor of the reference node
		my $sweepNode = $node->getSuccessor();
		
		#this variable contains, as the name suggests, the last ocupied entry in the finger table	
		my $lastOccupiedFingerEntry = -1;
		
		printf("Done with %s nodes from %s\n", $i-$startPosition+1, ($endPosition - $startPosition+1)) 
			if (($self->{"_DEBUG"}) && (($i-$startPosition+1)%1000 == 0) );
		printf("\nDealing with node number %d. ChordID %s and POP number %s\n",  $i, $node->getChordID(), $node->getPOP()) 
				if ($self->{"_DEBUG"} >= 2);
		
		#the search through the chordRing will continue as long as notEnd != 0
		my $notEnd =1;
		
		#and for each one set the nodes in the fingerTable
		while( $notEnd ){
			
			my $sweepChordID = $sweepNode->getChordID();
			my $distance;
			
			#compute the distance between the reference node and the sweep node	
			if ($sweepChordID>$referenceChordID){
				$distance = $sweepChordID-$referenceChordID;
				
			}else{
				#the sweepChordID is smaller than the referenceChorID, which implies that a jump from the last tot the first position on the 
				#chordRing has occured
				$distance = abs( $maxNodesInChordRing + $sweepChordID-$referenceChordID);
			}
			
			#in this variable, based on the distance between the nodes, is computed the possiby occupied finger table position 
			my $nextOccupiedFingerEntry = log($distance)/log(2);
			
			if ($self->{"_DEBUG"} >=3){
				printf("\n For node with ChordID %s and POP number %s testing POP with number %s and chordID %s.\n"
				. "Completing finger table entries from lastOccupiedFingerEntry=%s, nextOccupiedFingerEntry=%s\n",
					$referenceChordID, $node->getPOP(), $sweepNode->getPOP(), $sweepNode->getChordID(), $lastOccupiedFingerEntry, $nextOccupiedFingerEntry); 
			}
			
			#from last occupied finger table entry to the next one insert the current node as the "interval successor"
			for (my $i=$lastOccupiedFingerEntry+1; $i<= $nextOccupiedFingerEntry ; $i++){
				
				printf("For the node with ChordID %s and POP number %s adding as the %d finger the node with ChordID %s and POP number %s\n", 
						$node->getChordID(), $node->getPOP(), $i, $sweepNode->getChordID(), $sweepNode->getPOP() ) if($self->{"_DEBUG"}>=3);
					
				$fingerTable->setNodeForFinger($i, $sweepNode);

			}
			
			if ($lastOccupiedFingerEntry+1 <= $nextOccupiedFingerEntry){
				$lastOccupiedFingerEntry = int( $nextOccupiedFingerEntry ); 
			}
		
			#exit if the finger table is complete
			if ($lastOccupiedFingerEntry+1 == $fingerTableSize){
				last;
			}	
			
			#pass to the next node on the chord ring or stop when the sweepNode reaches back the node for which the finger table is computed
			if ($sweepNode->getSuccessor()->getChordID() != $referenceChordID ){
				$sweepNode = $sweepNode->getSuccessor();
			}else{
				$notEnd=0;
			}	
				
		}
		
		printf("\n Adding the reference node ChordID %s and POP id %s as a finger node for the last nodes\n",$node->getChordID(), 
				$node->getPOP()) if( ($self->{"_DEBUG"}>=3)&&($lastOccupiedFingerEntry+1< $fingerTableSize));
		
		#it might happend that the node list is too short and the last finger nodes in the fingerTable are exactly the referenceNode
		for (my $i=$lastOccupiedFingerEntry+1; $i< $fingerTableSize; $i++){

			printf("For the node with ChordID %s and POP id %s adding as the %d finger the node %s\n", 
					$node->getChordID(), $node->getPOP(), $i, $node->getPOP() )	if ($self->{"_DEBUG"}>=3);
					
			$fingerTable->setNodeForFinger($i, $node);
		}
		
		$node = $node->getSuccessor();
		
	}
		
}

#creates the finger tables
sub createFingerTables(){
	
	my $self=shift;
	my ($startPosition, $endPosition) = @_;
	
	if (defined($self->{"readFingerTablesFromFile"})){
		
		printf("(CR)Reading the fingerTables from the file %s\n", $self->{"readFingerTablesFromFile"});
		&readFingerTablesFromFile($self->{"chordList"}, $self->{"readFingerTablesFromFile"});
	
	}else{
		
		printf("(CR)Creating the finger tables. This will take a while ...\n");
		&createFingerTablesProcess($self);
	}

}

#sorts the prefix list
sub sortPrefixList(){
	my ($prefixList, $ipUtility) = @_;
	
	#sorting the list in descending order
	my @list = sort{ $ipUtility->getChordIDForPrefix($b) <=> $ipUtility->getChordIDForPrefix($a) } @{$prefixList};
	
	return \@list;
}

#for a given node it computes and returns a list that contains the fingerTable (this uses the whole topology so it is not to be used
#in a dynamical context)
sub createFingerTableForNode(){
	my $self = shift;
	my $node = shift;
	
	my $chordID = $node->getChordID();
	my $chordList = $self->{"chordList"};
	
}

#get node from chordID
sub getNodeWithID(){
	my $self = shift;
	my $chordID = shift;
	
	my $nodeHash = $self->{"chordList"}->getNodeHash();
	return $nodeHash->{$chordID};
}

#getSuccessor(nodeID)
sub getFirstNodeAfterID(){
	my $self = shift;
	my $id = shift;
	
	my $firstNode = $self->{"chordList"}->getFirstNode();
	my $nodeHash = $self->{"chordList"}->getNodeHash();
	
	if (defined $nodeHash->{$id}) {
		printf("This is bad! The node $id already exists!\n");
	}
	
	my $node = $firstNode;
	
	#search for the node
	while($node->getChordID() < $id){
		$node = $node->getSuccessor();
	}
	
	return $node;
	
} 

sub getNodesInChordRing(){
	my $self = shift;
	
	return $self->{"maxNodesInChordRing"};
}

#first version of the function
sub eliminateDuplicateChordIDs(){
	
	my ($list, $ipUtility) = @_;
	my @newlist=();
	
	my $listSize = scalar(@$list);
	my $endLoop = 0;
	my $i=0;
	
	
	while( ! $endLoop){
		
		my $currentPrefix = $list->[$i];
		my $currentChordID = &computeChordID($currentPrefix);
		
		my $generalMask = $ipUtility->getMaskFromPrefix($currentPrefix);
		my $generalPrefix = $currentPrefix;
	
		#find from all the consecutive prefixes with the same ChordID the one with the most general mask
		if ($i != $listSize-1){
			while ( ($i!=$listSize-1)&& ($currentChordID == &computeChordID($list->[$i+1]) ) ){
				
				my $prefix = $list->[$i+1];
				my $mask = $ipUtility->getMaskFromPrefix($prefix);
				
				if ($generalMask > $mask){
					$generalPrefix = $prefix;
					$generalMask = $mask; 
				}
				
				$i++;
			}
		}
		
		push(@newlist, $generalPrefix);
		
		$i++;
		if($i == $listSize ){
			$endLoop = 1;
		}
	}
	return \@newlist;
}

sub eliminateSpecificPrefixes(){
	
	my $self= shift;
	my $list = shift;
	
	my $ipUtility = $self->{"ipUtility"};
	my $checked;
	my @filteredList;
	
	my $listSize = scalar(@$list);
	my $endLoop = 0;
	my $i=0;
	
	while (! $endLoop){
		
		printf("Scanned the list of unfiltered prefixes until the index %s of %s\n", $i, $listSize) if ( ($self->{"_DEBUG"}) && ($i %10000 == 0) );
		
		#see if it was checked before
		if ( defined($checked->[$i]) ){
			$i++;
		}else{
			my $prefix = $list->[$i];
			
			my $foundAtLeastOneBetter=0;
			my $endSearch=0;
			
			my $j=$i+1;
			
			#check to see if we can't find another prefix "more general" than this one
			while (($j<$listSize)&&(!$endSearch)){
				
				#skip all the checked ones
				while( defined($checked->[$j]) && ($j< $listSize) ){
					$j++;
#					printf("prefix already checked %s\n", $list->[$j]);
				}
				
				if($j!=$listSize){
					
					my $tmpPrefix = $list->[$j];
#					printf("the j=%s and the tmpPrefix=%s, i=%s, listSize=%s\n", $j, $tmpPrefix, $i, $listSize);
					
					#check to see if the $tmpPrefix includes the $prefix
					if ($ipUtility->prefixInsidePrefix($prefix, $tmpPrefix)){
						$foundAtLeastOneBetter=1;
						$endSearch=1;
					}elsif ($ipUtility->prefixInsidePrefix($tmpPrefix, $prefix)){
						$checked->[$j] =1;
					}
					
					$j++;
				}
			
			}
			
			#if we haven't found a better prefix add the current one to the list
			if (!$foundAtLeastOneBetter){
				push(@filteredList, $list->[$i]);
			}
			
			
			#ATENTIE AICI
			$checked->[$i] = 1; 
			$i++;
		}
		

		if($i==$listSize){
			$endLoop=1;
		}
	}
	
	return \@filteredList;
	
}

sub eliminateSpecificPrefixes1(){
	
	my $self= shift;
	my $list = shift;
	
	my $sortedList;
	my @filteredList;
	my $checked;
	my $ipUtility = $self->{"ipUtility"}; 
	
	#sort the prefix list
	$sortedList = &sortPrefixList($list, $ipUtility);
	
	my $listSize = scalar(@$sortedList);
	my $i=0;
	
	while ($i<$listSize){
		
		printf("Scanned the list of unfiltered prefixes until the index %s of %s\n", $i, $listSize) if ( ($self->{"_DEBUG"}) && ($i %100 == 0) );
		
		#skip the checked elements
		if (defined($checked->[$i])){
			$i++;
		}else{
			my $listElement = $sortedList->[$i];
			my $firstBits = &getFirstBits($listElement);
			
			my $j=$i+1;
			my $endLoop = 0;
			my $foundABetterPrefix = 0;
			
			#while the first bits (8 bits) of the prefix stay the same (this is used to decrese the number of comparisons) do compare prefixes 
			#that haven't been compared. Also end the loop when the first "more general" prefix is found
			while( ( &getFirstBits($sortedList->[$j]) eq $firstBits) &&(!$endLoop) ){
				
				
				#if the elements have been checked skip them(also keep the condition related to the 8 bits)
				while(  defined($checked->[$j]) && ($j<$listSize) && (&getFirstBits($sortedList->[$j]) eq $firstBits) ){
					$j++;
				}
				
				#if we haven't reached the end of the list or finished the list of prefixes which begin with the same 8 bits
				if (( $j< $listSize) && (&getFirstBits($sortedList->[$j]) eq $firstBits)) {
					
					my $tmpElement = $sortedList->[$j];
					
					#we found a "wider" prefix
					if( &prefixIncludesPrefix($tmpElement, $listElement)){
						$foundABetterPrefix=1;
						$endLoop = 1;
					}elsif( &prefixIncludesPrefix($listElement, $tmpElement ) ){
						$checked->[$j] = 1;
					}
					
					$j++;
				}
			}
			
			#if we don't find a better prefix, add the current one to the list
			if (!$foundABetterPrefix){
				push(@filteredList, $sortedList->[$i]);
			}
			
			$checked->[$i] = 1; 
			$i++;
		}
		
	}
	
	return \@filteredList;
	
}

#it uses the ideea discoverd while building ALT (until now the fastest i kept the other for reference)
sub eliminateSpecificPrefixes2(){
	
	my ($self, $prefixList) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	my $debug = $self->{"_DEBUG"}; 
	
	my $prefixHash={};
	my @filteredPrefixList;
	
	
	# $i will take all the possible values for the prefix masks. From the prefixFile I noticed that there are no prefixes more general 
	#than /8 or more specific than /30

	for (my $i=8; $i<= 32; $i++){
		
		foreach my $prefix (@$prefixList){
			
			if ($ipUtility->getMaskFromPrefix($prefix) == $i){
					
				$prefixHash->{$prefix} = $prefix;
				
				#search for the parent router (the min prefix is 8)
				my $parentPrefix = &findParentPrefixForPrefix($self, $prefix, 8, $prefixHash);
				
				#if the parent is the root node then add the altRouter to the filteredPrefixList
				if ($parentPrefix eq "0.0.0.0/0"){
					
					push(@filteredPrefixList,$prefix); 
					
				}		
			}
			
		}
	}
	
	return \@filteredPrefixList;
	
	
	
}

sub findParentPrefixForPrefix(){
	my $self=shift;
	my ($childPrefix, $minMask, $prefixHash) = @_;
	
	my $ipUtility = CoreSim::Utilities::IPUtilities->new();

	my $endLoop =0;
	my $possibleParentPrefix;
	
	my $mask = $ipUtility->getMaskFromPrefix($childPrefix);
	my $prefix = $childPrefix;
	
	#search the hierarchy for a parent prefix
	while (($mask > $minMask) && (!$endLoop)){
		
		#get a possible parent prefix with a mask decreased by 1
		$possibleParentPrefix = $ipUtility->computeParentPrefix($prefix, 1);
		
		#if the parent does exist stop searching
		if (defined($prefixHash->{$possibleParentPrefix})){
			$endLoop = 1;
		}
		
		$mask -= 1;
		$prefix = $possibleParentPrefix;
	}
	
	if ($endLoop==0){
		return "0.0.0.0/0";
	}else{
		return $prefixHash->{$possibleParentPrefix};
	}
	
}

sub getFirstBits(){
	my $prefix = shift;
	
	if (defined($prefix)){
		my @list = split(/\./,$prefix);
		return $list[0];
	}else{
		return -1;
	}
}

sub checkDuplicateChordID(){
	my ($self, $list) = shift;
	
	my $ipUtility = $self->{"ipUtility"};
	my $listSize = scalar(@$list);
	my $endLoop = 0;
	my $i=0;
	
	while( ! $endLoop){
		
		my $currentPrefix = $list->[$i];
		my $currentChordID = $ipUtility->getChordIDForPrefix($currentPrefix);
		
		my $generalMask = $ipUtility->getMaskFromPrefix($currentPrefix);
		my $generalPrefix = $currentPrefix;
		
		my $currentAS = $self->{"topology"}->getPrefixToAS($currentPrefix);
	
		#find from all the consecutive prefixes with the same ChordID the one with the most general mask
		if ($i != $listSize-1){
			while ( ($i!=$listSize-1)&& ($currentChordID == $ipUtility->getChordIDForPrefix($list->[$i+1]) ) ){
				
				my $prefix = $list->[$i+1];
				my $mask = $ipUtility->getMaskFromPrefix($prefix);
				my $as = $self->{"topology"}->getPrefixToAS($prefix);
				
				if ($as!=$currentAS){
					printf("prefix1=%s from as %s, prefix2=%s from as %s\n", $currentPrefix, $currentAS, $prefix, $as);
				}
				$i++;
			}
		}
				
		$i++;
		if($i == $listSize ){
			$endLoop = 1;
		}
	}
}

sub dropFingerTablesToFile(){
	my $self=shift;
	my $outputFilePath=shift;
	
	printf("Writing the fingerTables to the file %s!\n", $outputFilePath) if ($self->{"_DEBUG"});
	open(OFILE,"> $outputFilePath") or die printf("Error while opening the output file for storing the finger tables:%s\n", $!);
	 
	my $firstNode = $self->{"chordList"}->getFirstNode();
	my $fingerTableSize = $self->{"maxNodesInFingerTable"};
		
	my $startPosition = $self->{"startPosition"};
	my $endPosition = $self->{"endPosition"};
	
	if (!defined($endPosition)){
		$endPosition = $self->{"chordList"}->getNumberOfNodesInChordList();
	}
	
	my $distance = $startPosition;
	my $node = $firstNode;
	
	#find the first node in the list of nodes which will be dropped into the file
	while($distance){
		$node = $node->getSuccessor();
		$distance--;
	}
	
	#write the fingerTableSize and the maxNodesInChordRing	
	printf(OFILE "%s %s\n",$self->{"maxNodesInFingerTable"}, $self->{"maxNodesInChordRing"});
	
	for(my $i=$startPosition; $i<=$endPosition; $i++){
		printf(OFILE "%s ", $node->getChordID());
		for (my $i=0;$i<$fingerTableSize;$i++){
			printf(OFILE "%s ", $node->getFingerTable()->getNodeForFinger($i)->getChordID() );
		}
		printf(OFILE "\n");
		$node= $node->getSuccessor();
	}
	
	close(OFILE);
}

sub dropFancyFingerTablesToFile(){
	my $self=shift;
	my $outputFilePath=shift;
	
	printf("Writing the fingerTables to the file %s!\n", $outputFilePath) if ($self->{"_DEBUG"});
	open(OFILE,"> $outputFilePath") or die printf("Error while opening the output file for storing the finger tables:%s\n", $!);
	 
	my $firstNode = $self->{"chordList"}->getFirstNode();
	my $fingerTableSize = $self->{"maxNodesInFingerTable"};
		
	my $startPosition = $self->{"startPosition"};
	my $endPosition = $self->{"endPosition"};
	
	if (!defined($endPosition)){
		$endPosition = $self->{"chordList"}->getNumberOfNodesInChordList();
	}
	
	my $distance = $startPosition;
	my $node = $firstNode;
	
	#find the first node in the list of nodes which will be dropped into the file
	while($distance){
		$node = $node->getSuccessor();
		$distance--;
	}
	
	#write the fingerTableSize and the maxNodesInChordRing	
	#printf(OFILE "%d %d\n",$self->{"maxNodesInFingerTable"}, $self->{"maxNodesInChordRing"});
	
	for(my $i=$startPosition; $i<=$endPosition; $i++){
		#printf(OFILE "Node: %08X (PoP: %d)\nFinger table:\n", $node->getChordID(), $node->getPOP());
		printf(OFILE "Node: %08X\nFinger table:\n", $node->getChordID());
		my $first = 0;
		for (my $i=1;$i<$fingerTableSize;$i++){
			if ($node->getFingerTable()->getNodeForFinger($i)->getChordID() == $node->getFingerTable()->getNodeForFinger($i-1)->getChordID()) {
				unless ($i == ($fingerTableSize-1)) {
					next;
				}
				printf(OFILE "  %08X (%d-%d)\n", $node->getFingerTable()->getNodeForFinger($i)->getChordID(), $first, $i);
			} else {
				if ($first == ($i-1)) {
					printf(OFILE "  %08X (%d)\n", $node->getFingerTable()->getNodeForFinger($i-1)->getChordID(), $i-1);
				} else {
					printf(OFILE "  %08X (%d-%d)\n", $node->getFingerTable()->getNodeForFinger($i-1)->getChordID(), $first, $i-1);
				}
				$first = $i;
				printf(OFILE "  %08X (%d)\n", $node->getFingerTable()->getNodeForFinger($i)->getChordID(), $i) if ($i == ($fingerTableSize-1));
			}
		}
		#printf(OFILE "Predecessor:\n  %08X (PoP: %d)\n", $node->getPredecessor()->getChordID(), $node->getSuccessor()->getPOP());
		#printf(OFILE "\n");
		$node = $node->getSuccessor();
	}
	
	close(OFILE);
}

sub dropSortedAndFilteredPrefixesToFile(){
	my $self = shift;
	my ($outputFilePath) = @_;
	
	my $list=$self->{"sortedAndFilteredPrefixList"};
	
	printf("Writing sorted and filtered prefixes to file %s\n", $outputFilePath);
	open(OFILE, "> $outputFilePath") or die printf("Can't open file %s to write sorted and filtered prefixes:%s\n",$outputFilePath, $!);
		
	foreach my $prefix (@$list){
		printf(OFILE "%s\n",$prefix );
	}
	close(OFILE);
}

sub readSortedAndFilteredPrefixesFromFile(){

	my $inputFilePath = shift;
	
	open(IFILE, $inputFilePath) or die printf("Can't read file %s : %s", $inputFilePath, $!);
	
	my @list; 
	my $i=0;
	
	while(<IFILE>){
		chomp;
		$list[$i] = $_; 
		$i++;
	}
	
	close(IFILE);
	
	return \@list;
}

#reads the finger tables and inserts them into the corresponding nodes by means of the chordList
sub readFingerTablesFromFile(){

	my ($chordList,$inputFilePath) = @_;
	
	open(IFILE, $inputFilePath) or die printf("Error while opening the input file for the finger tables %s:%s\n",$inputFilePath, $!);
	
	my @firstLine = split(/ /,<IFILE>);
	my $fingerTableSize = $firstLine[0];
	my $maxNodesInChordRing = $firstLine[1];
	
	
	while(<IFILE>){
		my @line = split;
		my $node = $chordList->getNodeWithID($line[0]);
		
		if (!defined($node)){
			printf("(EE) the node %s is undefined in the ChordList. Check the compatibility between the fingerTables and the chordList\n", $line[0]);
		}else{
			for (my $i=1; $i<= $fingerTableSize ; $i++){
				my $fingerTable = $node->getFingerTable();
				my $finger = $chordList->getNodeWithID($line[$i]);
				$fingerTable->setNodeForFinger($i-1, $finger);
			}
		}
		
	}
	
	close(IFILE);
}

sub showFingerTablesPrefixesForChordID(){
	my $self=shift;
	my ($chordID) = @_;
	
	my $nodeHash = $self->{"chordList"}->getNodeHash();
	my $chordNode = $nodeHash->{$chordID};
	
	my $fingerTable = $chordNode->getFingerTable();
	
	for (my $i=0;$i<32;$i++){
		my $node = $fingerTable->getNodeForFinger($i);
		printf("%s\n", $node->getPrefix());
	}
}

sub testChordRing(){
	my $self=shift;
	my ($chordID) = @_;
	
	my $chordHash = $self->{"chordList"}->getNodeHash();
	my $chordNode = $chordHash->{$chordID};
	
	printf("The successor for node %s is %s and the predecessor is %s\n", 
		$chordNode->getChordID(), $chordNode->getSuccessor()->getChordID(), $chordNode->getPredecessor()->getChordID() );
		
}

#return 1 at the end
1
