package CoreSim::ITR::MapCache;

use strict;
use CoreSim::Utilities::MCLinkedList; 

sub new(){
	my $invocant = shift;
	
	my ($timeout, $ipUtility, $maxNbOfEntriesInCache, $outputPath, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	my $outputFile = "$outputPath/mapCache.txt";
	my $cacheDatabase = CoreSim::Utilities::MCLinkedList->new($ipUtility, $_DEBUG);
	my $prefixHash = {};
	my $entriesInCache = 0;
	
	my $self = {
		"timeout" => $timeout,
		"cacheDatabase" => $cacheDatabase,
		"prefixHash" => $prefixHash,
		"ipUtility" => $ipUtility,
		"maxNbOfEntriesInCache" => $maxNbOfEntriesInCache,
		"outputFile" => $outputFile,
		"entriesInCache" => $entriesInCache,
		"_DEBUG" => $_DEBUG,
	};
	
#	$self->{"_DEBUG"}= 4;
	
	bless($self, $class);
	
	&createOutputFile($self);

	return $self;
	
}

sub update(){
	my ($self, $newTimestamp)=@_;
	
	
#	printf("the current timestamp %d\n", defined( $self->{"currentTimestamp"} ) );
	
	if ( !defined( $self->{"currentTimestamp"} ) ){
		printf("\n(MC)This is the initialization of the mapCache. First timestamp is %s\n\n", $newTimestamp->getTimestampString()) 
				if ($self->{"_DEBUG"}>2);
		$self->{"currentTimestamp"} = $newTimestamp;
		
	}else{
		
		printf("(MC)Updating the mapCache. The newTimestamp is %s\n", $newTimestamp->getTimestampString()) if ($self->{"_DEBUG"}>2);
		
		my ($firstElementPrefix,$firstElementTimestamp) = $self->{"cacheDatabase"}->firstElement();
		
		if ( defined($firstElementPrefix)){
			
			#searching for expired entries
			while (  $newTimestamp->greaterThan($firstElementTimestamp)  ){
					
				printf("(MC)Deleting the prefix %s with timestamp %s from the mapCache because it expired %s\n", $firstElementPrefix, 
						$firstElementTimestamp->getTimestampString(), $newTimestamp->getTimestampString()) if ($self->{"_DEBUG"}>2);
						
				&deleteFirstCacheEntry($self,$firstElementPrefix, $firstElementTimestamp);
				
				($firstElementPrefix,$firstElementTimestamp) = $self->{"cacheDatabase"}->firstElement();
				
	
				#end the loop when the firstElement is undefined
				if ( !defined($firstElementPrefix) ){
					last;
				}
				
			}
		}
		
		printf("(MC)DONE with the updating\n") if ($self->{"_DEBUG"}>2);
		
	}
	
}

sub deleteFirstCacheEntry(){
	my $self=shift;
	my ($prefix, $expireTimestamp) = @_;
	
	printf("(MC) Deleting the entry for %s with timestamp %s\n", $prefix, $expireTimestamp->getTimestampString()) if ($self->{"_DEBUG"}>=3);
	
	my $timeout = $self->{"timeout"};
	$self->{"entriesInCache"} -=1;
	
	#the subValue works with ms not seconds
	my $enterTimestamp = $expireTimestamp->subValue($timeout*1000);
	
	#send this information to output file
	my $delete=-1;
	my $entriesInCache = $self->{"entriesInCache"};
	my $expireTimestampString = $expireTimestamp->getTimestampString();
	&writeToOutputFile($self, $delete, $expireTimestampString,$prefix, $entriesInCache );
	
	$self->{"cacheDatabase"}->deleteFirstElement();
	printf("(MC) We still have %s entries in the mapCache\n", $self->{"entriesInCache"}) if ($self->{"_DEBUG"}>=3);
}

#add a new entry into the Cache (this method is called by the inFlightBuffer)
sub addCacheEntry(){
	my $self=shift;
	my ($prefix, $entryTimestamp) = @_;

	my $maxNbOfEntriesInCache = $self->{"maxNbOfEntriesInCache"};
	my $sizeOfCache = $self->{"entriesInCache"};
	my $timeout = $self->{"timeout"};
	
	my $expireTimestamp = $entryTimestamp->addValue($self->{"timeout"}*1000);
	
	if ($sizeOfCache< $maxNbOfEntriesInCache){
		
		#increment the counter only if the prefix is not already in the cache
		if ( ! ($self->{"cacheDatabase"}->elementExists($prefix) )  ){
			$self->{"entriesInCache"} +=1;
			
			printf("(MC) Adding a new element to the MapCache with prefix %s and timestamp %s. It expires %s\n", $prefix, 
			$entryTimestamp->getTimestampString(), $expireTimestamp->getTimestampString() ) if ($self->{"_DEBUG"} > 2);
			$self->{"cacheDatabase"}->pushElement($expireTimestamp, $prefix);
			
			#send this information to output file
			my $add=1;
			my $entriesInCache = $self->{"entriesInCache"};
			my $entryTimestampString = $entryTimestamp->getTimestampString();
			
			#output the mapping information with the entryTimestamp (not the expireTimestamp)
			&writeToOutputFile($self, $add, $entryTimestampString,$prefix, $entriesInCache );
			
		#another reply for the same prefix
		}else{
			printf("(MC) Updating the element with prefix %s to timestamp %s\n", $prefix, $entryTimestamp->getTimestampString()) 
				if ($self->{"_DEBUG"}>2);
				
			$self->{"cacheDatabase"}->updateElement($prefix, $expireTimestamp);
		}

		
	}else{
		printf("(W)(MC) Trying to add %s with timestamp %s to mapCache but the cache is full!\n", 
				$prefix, $entryTimestamp->getTimestampString());
	}
	
}

sub updateCacheEntry(){
	my ($self, $destinationPrefix, $newTimestamp) = @_;
	
	#Compute the expire timestamp. addValue works with ms
	my $expireTimestamp = $newTimestamp->addValue($self->{"timeout"} * 1000);
	$self->{"cacheDatabase"}->updateElement($destinationPrefix, $expireTimestamp);
	
}

sub ipInCache(){
	
	my $self= shift;
	my ($destinationIP,$currentTimestamp) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	my $prefixHash = $self->{"prefixHash"};
	
	my $mask=32;
	my $destinationPrefix = "$destinationIP/$mask";
	my $oldTimestamp;
	
	while($mask>=8){
		
		$oldTimestamp = $self->{"cacheDatabase"}->getTimestamp($destinationPrefix);
		
		printf("(MC) checking for destination ip %s if the prefix %s is found in cache\n", $destinationIP, $destinationPrefix) if ($self->{"_DEBUG"}>3);
		if ( defined($oldTimestamp) ){
			last;
		}else{
			#compute the parent prefix by shrinking the mask with 1 (/32 -> /31 ... ->/8)
			$destinationPrefix = $ipUtility->computeParentPrefix($destinationPrefix,1);
			$mask--;
		}
	}
	
	if (defined($oldTimestamp)){
					
		printf("(MC) Cache hit! For the destination IP %s we found the prefix %s in cache\n",$destinationIP, $destinationPrefix ) if ($self->{"_DEBUG"});
		
		#cache hit! renew the cache entry timestamp
		&updateCacheEntry($self, $destinationPrefix, $currentTimestamp);
		
		return 1;
				
		
	#there is no entry with one of the searched for prefixes
	}else{
		return 0;
	}
	
}

sub writeToOutputFile(){
	my ($self, $action, $timestamp, $prefix, $entriesInCache) = @_;
	
	my $mapCacheFileHandle  = $self->{"mapCacheFileHandle"};
	
	#add value to cache
	if ($action == 1){
		printf($mapCacheFileHandle "add\t%s\t%s\t%s\n", $timestamp, $prefix, $entriesInCache);
	}elsif($action == -1){
		printf($mapCacheFileHandle "del\t%s\t%s\t%s\n", $timestamp, $prefix, $entriesInCache);
	}
	
}

sub createOutputFile(){
	my $self=shift;
	
	my $outputFile = $self->{"outputFile"};
	
	open(OFILE, "> $outputFile") or die printf("Error opening the file %s associated to the ITR!\n", $outputFile);
	
	my $timeout = $self->{"timeout"};
	my $maxNbOfEntriesInCache = $self->{"maxNbOfEntriesInCache"};
	
	printf(OFILE "timeout\tmaxNbOfEntriesInCache\n");
	printf(OFILE "%s\t%s\n", $timeout, $maxNbOfEntriesInCache );
	
	$self->{"mapCacheFileHandle"} = \*OFILE;

}

sub DESTROY(){
	my $self=shift;
	
	printf("(MC) Closing mapCache file handle\n");
	my $mapCacheFileHandle = $self->{"mapCacheFileHandle"};
	close($mapCacheFileHandle);	
}

#return 1
1