package CoreSim::ITR::ITRNode;

use strict;
use CoreSim::Utilities::IPUtilities;
use CoreSim::Utilities::Timestamp;
use CoreSim::Utilities::IPPacket;
use CoreSim::ITR::PacketBuffer;
use CoreSim::ITR::InFlightBuffer;
use CoreSim::ITR::DiscardBuffer;
use CoreSim::ITR::MapCache;


sub new(){
	my $invocant = shift;
	my ($mappingSystem,$associatedNode, $cacheTimeout, $outputPath, $maxNbOfEntriesInCache, $_DEBUG ) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	my $ipUtility = CoreSim::Utilities::IPUtilities->new();
	my $mapCache = CoreSim::ITR::MapCache->new($cacheTimeout, $ipUtility,$maxNbOfEntriesInCache, $outputPath, $_DEBUG );
	my $packetBuffer = CoreSim::ITR::PacketBuffer->new($ipUtility, $outputPath, $_DEBUG);
	my $inFlightBuffer = CoreSim::ITR::InFlightBuffer->new($mapCache, $packetBuffer, $ipUtility, $outputPath, $_DEBUG);
	my $discardBuffer = CoreSim::ITR::DiscardBuffer->new();
	 
	my $discardFile = "$outputPath/discard.txt";
	my $itrOutputFile = "$outputPath/itrOutput.txt";
	my $self={
		"mappingSystem" => $mappingSystem,
		"associatedNode" => $associatedNode,
		"mapCache" => $mapCache,
		"outputFile" => $itrOutputFile, 
		"cacheTimeout"=> $cacheTimeout,
		"packetBuffer" => $packetBuffer,
		"inFlightBuffer" => $inFlightBuffer,
		"maxNbOfEntriesInCache" => $maxNbOfEntriesInCache,
		"ipUtility" => $ipUtility,
		"discardFile" => $discardFile,
		"discardBuffer" => $discardBuffer,
		"_DEBUG" => $_DEBUG,
	};
	
	&createOutputFiles($self);
	
	
	bless($self, $class);
#	$self->{"_DEBUG"}=3;
	return $self;
}

sub searchForIP(){
	
	my $self=shift;
	my ($searchedIP, $hexTimestamp, $protocol, $size) = @_;
	
	my $packetBuffer = $self->{"packetBuffer"};
	my $inFlightBuffer = $self->{"inFlightBuffer"};
	my $mapCache = $self->{"mapCache"};
	my $mappingSystem = $self->{"mappingSystem"};
	my $discardBuffer = $self->{"discardBuffer"};
	
	my $hit;
	my $currentTimestamp = CoreSim::Utilities::Timestamp->new($hexTimestamp);
	my $packet = CoreSim::Utilities::IPPacket->new($searchedIP, $hexTimestamp, $protocol, $size);
	
	#use as destinationInformation the searchedIP (another option would be to use the destinationPrefix!)
	my $destinationInformation = $searchedIP;
	
	printf("\n(ITR) The destination : %s\n", $destinationInformation) if ($self->{"_DEBUG"});
		
	#prune the inflight buffer (if an entry is pruned here it will also prune the packetBuffer)
	$inFlightBuffer->update1($currentTimestamp);
	
	#check the discardBuffer to see if we drop the packet before further processing is done
	if ($discardBuffer->ipInBuffer($searchedIP)){
		
		#the packet is discarded and this is shown by writing in the discard file
		my $destPrefix = $discardBuffer->getPrefixForIP($searchedIP);
		&writeDiscardedPacket($self, $packet, $destPrefix);
	
	#if the searchedIP is not found in the discardBuffer, the program continues with the "normal" processing	
	}else{
	
		#check if a MAP-REQUEST exists for the destination prefix
		if ($inFlightBuffer->checkExistingMapRequest($destinationInformation)){
			
			printf("(ITR) For the destination %s we found an existing map request!\n", $destinationInformation) if ($self->{"_DEBUG"});
			
			#deal with TCP packets
			if ($packet->getProtocol() == 6){
				
				#since we store destination IP in inFlightBuffer, when we have an TCP inFlightHit -> same destination IP , so we drop
				#the packet
				
				#the commented case can be used if we store in the inFlightBuffer the destinationPrefix
				#we decided NOT to use this approach as it is unlikely to have all the prefixes in the internet stored on all the ITRs
#				my $bufferedPacket = $inFlightBuffer->getPacketForMapRequest($destinationInformation);
#				
#				#test if the packet is the first in the flow
#				#the two packets are from the same flow (they have the same destination IP)
#				if ($packet->getDestinationIP() eq $bufferedPacket->getDestinationIP() ){
#					#this packet should be dropped as TCP will deal with it
#					#nothing is done here, as equivalent to dropping
#				
#				#the two packets are not from the same flow (different destination IP, they have the same destination prefix) 
#				}else{
#					
#					#this packet should be buffered in the packet buffer as it is the first (SYN probably) packet in the flow
#					my $expireTimestamp = $inFlightBuffer->getExpireTimestampForMapRequest($destinationInformation);
#					$packetBuffer->addPacket($packet, $expireTimestamp);
#				}
			
			#deal with UDP packets	
			}elsif($packet->getProtocol() == 17){
					
					#the UDP policy is to buffer all packets
					my $expireTimestamp = $inFlightBuffer->getExpireTimestampForMapRequest($destinationInformation);
					$packetBuffer->addPacket($packet, $expireTimestamp);
					
			#deal with ICMP packets (could be part of the general else part, but maybe we decide to do something else with ICMP)
			}elsif($packet->getProtocol() == 1){
				
				#these are just dropped, no buffering
			
			}else{
				#drop the packet
				printf("(EE)(ITR) The protocol field of the packet is neider TCP, UDP nor ICMP!Dropping the packet!\n");
			}
		
			#for all cases we have a cache miss but different from usual cache misses due to the fact that a Map-Request exists
			$hit=-1;
			&writeITRDataToFile($self, $hexTimestamp, $hit);
			
		#if the searchedIP does not have a request associated to it in the inFlightBuffer 
		}else{
			
			printf("(ITR) For the destination %s we did not find an existing map request!\n", $destinationInformation) if ($self->{"_DEBUG"});
			
			#check the mapCache(prefixCache), to see if a prefix in the cache includes the searchedIP
			if ( $mapCache->ipInCache($searchedIP, $currentTimestamp) ){
				
				printf("(ITR) For destination %s we found an existing entry in the map cache\n", $destinationInformation) if ($self->{"_DEBUG"});
				
				#this is a cache hit
				$hit = 1;
				&writeITRDataToFile($self, $hexTimestamp, $hit);
				
			#if the searchedIP is not in the cache do a MAP-REQUEST and add an entry in the inFlightBuffer and buffer the packet
			#this is a cache miss
			}else{
				printf("(ITR) For destination %s we do a map-request\n", $destinationInformation) if ($self->{"_DEBUG"});
				
				$hit = 0;
				
				my $time1 = time;
				my ($status, $destPOP, $destAS, $destPrefix, $lowerBoundLatency, $nbOfPeers, $nodeLatencyList) = 
																	$mappingSystem->getLookupDelay( $self->{"associatedNode"}, $searchedIP);
				
				my $time2 = time;
				
				#if the exit status from the getLookupDelay is "ok" (meaning that the search in the mapping system is valid)
				if ($status eq "ok"){
					
					&writeITRDataToFile($self, $hexTimestamp, $hit, $searchedIP, $destPOP, $destAS, $destPrefix, $lowerBoundLatency, $nbOfPeers, $nodeLatencyList);
					
					#add an entry in the inFlightBuffer (will it be destinationPrefix or destinationIP (or both)... we'll see)
					my $expireTimestamp = $currentTimestamp->addValue($lowerBoundLatency);
					$inFlightBuffer->addEntry($destinationInformation,$expireTimestamp, $packet, $destPrefix);
					
					#buffer the packet
					$packetBuffer->addPacket($packet, $expireTimestamp);
					
					
					#some debug information
					printf("(ITR)The time needed for the mapping system to provide the latency for finding the ip %s was %s seconds\n\n",
							$searchedIP, $time2-$time1) if ($self->{"_DEBUG"}>=2);
					
					if (defined($lowerBoundLatency)){
						
						printf("(ITR)The searched ip %s was found in the mapping system. The latency associated to the search is %s\n", 
								$searchedIP, $lowerBoundLatency) if ($self->{"_DEBUG"}>=2);
						
						return $lowerBoundLatency;
						
					}else{
						
						printf("(EE)(ITR) The search for IP %s from node %s returned with error!\n", $searchedIP, $self->{"associatedNode"});
						return undef;
					}
					
				
				#if the status of the return is discard that means that the responsible router is a virtual router
				#this may happend if the destination IP belongs to a router about which iPlane does not know anything
				}elsif($status eq "discard"){
					&writeDiscardedPacket($self, $packet, $destPrefix);
					
					#also add this information in the discardBuffer
					$discardBuffer->addEntry($searchedIP, $destPrefix);
		
				}
				
				
			}
			
		}
	
	}
	
}

sub writeITRDataToFile(){
	
	my ($self, $timestamp, $hit, $destIP, $destPOP, $destAS, $destPrefix, $lowerBoundLatency, $nbOfPeers, $nodeLatencyList) = @_;
	
	my $itrFileHandle = $self->{"itrFileHandle"};
	
	
	if ($hit == 1){
		printf($itrFileHandle "%s\n", $timestamp);
	}elsif($hit==0){
#		printf("adding the following %s %s %s %s %s %s %s %s \n",$timestamp, $hit, $destIP, $destPOP, $destAS, $destPrefix, 
#						$lowerBoundLatency, $nbOfPeers);
		printf("(ITR) The nodeLatencyList @$nodeLatencyList\n") if ($self->{"_DEBUG"}>=2);
		printf($itrFileHandle "%s %s %s %s %s %s %s %s @$nodeLatencyList \n", $timestamp, $hit, $destIP, $destPOP, $destAS, $destPrefix, 
						$lowerBoundLatency, $nbOfPeers);
	}elsif($hit==-1){
		
		printf($itrFileHandle "%s %s\n", $timestamp, $hit);
	}
	
}

sub writeDiscardedPacket(){
	my ($self, $packet, $destPrefix) = @_;
	
	my $discardFileHandle = $self->{"discardFileHandle"};
	
	printf($discardFileHandle "%s %s %s %s %s\n", $packet->getTimestamp(), $packet->getDestinationIP(), $destPrefix, $packet->getProtocol(), 
			$packet->getSize());
		
}

sub createOutputFiles(){
	my $self=shift;
	
	my $outputFile = $self->{"outputFile"};
	my $mappingSystem = $self->{"mappingSystem"};
	my $resolverType = $mappingSystem->getResolverType();
	my $associatedNode = $self->{"associatedNode"};
	my $cacheTimeout = $self->{"cacheTimeout"};
	
	
	open(ITR, "> $outputFile") or die printf("Error while opening the file %s associated to the ITR!", $outputFile);
	
	printf(ITR "associatedNode\tresolverType\tcacheTimeout\n");
	printf(ITR "%s\t%s\t%s\n", $associatedNode, $resolverType, $cacheTimeout);
	
	$self->{"itrFileHandle"} = \*ITR;
	
	my $discardFile = $self->{"discardFile"};
	open(DISCARD, "> $discardFile") or die printf("Error while opening the file %s used for writing the discarded packets!:%s", 
		$discardFile, $! );
	printf(DISCARD "File used for the discarded packets!\n");
	
	$self->{"discardFileHandle"} = \*DISCARD;
	
	
}

sub DESTROY(){
	my $self=shift;
	
	printf("(ITR) Closing itr file handle and discard file handle\n");
	
	my $itrFileHandle = $self->{"itrFileHandle"};
	close($itrFileHandle);
	
	my $discardFileHandle = $self->{"discardFileHandle"};
	close($discardFileHandle);
}

#return 1
1
