package CoreSim::ITR::MapCache;

use strict;

sub new(){
	my $invocant = shift;
	
	my ($timeout, $ipUtility, $maxNbOfEntriesInCache, $outputPath, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	my $outputFile = "$outputPath/mapCache.txt";
	my $cacheDatabase ={};
	my $packetsInCache = 0;
	
	my $self = {
		"timeout" => $timeout,
		"cacheDatabase" => $cacheDatabase,
		"ipUtility" => $ipUtility,
		"maxNbOfEntriesInCache" => $maxNbOfEntriesInCache,
		"outputFile" => $outputFile,
		"packetsInCache" => $packetsInCache,
		"_DEBUG" => $_DEBUG,
	};
	
	bless($self, $class);
	
	&createOutputFile($self);

	return $self;
	
}

#this should be called after a cache miss(or for updating a cache entry)
sub addCacheEntry(){
	my $self=shift;
	my ($destinationInformation, $timestamp) = @_;
	
	my $cacheDatabase = $self->{"cacheDatabase"};
	my $maxNbOfEntriesInCache = $self->{"maxNbOfEntriesInCache"};
	my $sizeOfCache = scalar(keys %$cacheDatabase);
	
	if ($sizeOfCache< $maxNbOfEntriesInCache){
		
		#increment the counter only if the prefix is not already in the cache
		if (!defined($self->{"cacheDatabase"}->{$destinationInformation})){
			$self->{"packetsInCache"} +=1;
		}
		#the destination may be the destinationIP or the destinationPrefix(depends on the type of cache)
		$self->{"cacheDatabase"}->{$destinationInformation} =$timestamp;
		
		#send this information to output file
		my $add=1;
		my $packetsInCache = $self->{"packetsInCache"};
		my $eTimestamp = $timestamp->getTimestampString();
		&writeToOutputFile($self, $add, $eTimestamp,$destinationInformation, $packetsInCache );
		
	}else{
		printf("(W)(MC) Trying to add %s with timestamp %s to mapCache but the cache is full!\n", 
				$destinationInformation, $timestamp->getTimestampString());
	}
	
}

sub deleteCacheEntry(){
	my $self=shift;
	my ($destinationInformation) = @_;
	
	printf("(MC) Deleting the entry for %s\n", $destinationInformation) if ($self->{"_DEBUG"}>=3);
	
	my $cacheDatabase = $self->{"cacheDatabase"};
	my $timeout = $self->{"timeout"};
	$self->{"packetsInCache"} -=1;
	
	my $enterTimestamp = $cacheDatabase->{$destinationInformation};
	my $expireTimestamp = $enterTimestamp->addValue($timeout);
	
	#send this information to output file
	my $delete=-1;
	my $packetsInCache = $self->{"packetsInCache"};
	my $eTimestamp = $expireTimestamp->getTimestampString();
	&writeToOutputFile($self, $delete, $eTimestamp,$destinationInformation, $packetsInCache );
	
	delete $cacheDatabase->{$destinationInformation};	
	printf("(MC) We still have % entries in the mapCache\n", scalar(keys %$cacheDatabase)) if ($self->{"_DEBUG"}>=3);
}

#works when destination prefixes are stored in cache
sub ipInCache(){
	
	my $self= shift;
	my ($destinationIP,$newTimestamp) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	
	my $mask=32;
	my $destinationPrefix = "$destinationIP/$mask";
	my $oldTimestamp;
	
	while($mask>=8){
		$oldTimestamp = $self->{"cacheDatabase"}->{$destinationPrefix};
		
#		printf("checking for destination ip %s if the prefix %s is found in cache\n", $destinationIP, $destinationPrefix);
		if ( defined($oldTimestamp) ){
			last;
		}else{
			#compute the parent prefix by shrinking the mask with 1 (/32 -> /31 ... ->/8)
			$destinationPrefix = $ipUtility->computeParentPrefix($destinationPrefix,1);
			$mask--;
		}
	}
	
	if (defined($oldTimestamp)){
		
		if ( !&timedOutEntry($self, $oldTimestamp, $newTimestamp, $self->{"timeout"} )  ){
			
			printf("(MC) For the destination IP %s we found the prefix %s in cache\n",$destinationIP, $destinationPrefix ) if ($self->{"_DEBUG"});
			
			#cache hit! renew the cache entry timestamp
			&addCacheEntry($self, $destinationPrefix, $newTimestamp);
			return 1;
			
		#the cache entry has timedout, delete the cache entry
		}else{
			
			printf("(MC) For the destination IP %s we found the prefix %s in cache but it has timedout\n",$destinationIP, $destinationPrefix );
			
			&deleteCacheEntry($self, $destinationPrefix);
			return 0;
		}
		
	#there is no entry with one of the searched for prefixes
	}else{
		return 0;
	}
	
}

sub timedOutEntry(){
	my ($self, $oldTimestamp, $newTimestamp, $timeout) = @_;
	
	#the following relation is checked: timeout> newTimestamp - oldTimestamp. If this holds then we have a cache hit
	#the difference is a reference to a hash with 2 keys: seconds and fractions
	my $difference = $newTimestamp->subtract($oldTimestamp);
	
	printf("(MC) The expire timestamp %s the new current %s and the difference %s\n",$oldTimestamp->getTimestampString(), 
			$newTimestamp->getTimestampString(), $difference->{"seconds"} ) if ($self->{"_DEBUG"}>=3);
			
	if ( defined($difference) ) {
		if ( $difference->{"seconds"} < $timeout){
			#a cache hit
			return 0;
		}else{
			#the entry has timedout
			return 1;
		}
		
	}else{
		return 1;
	}
}

sub writeToOutputFile(){
	my ($self, $action, $timestamp, $prefix, $entriesInCache) = @_;
	
	my $mapCacheFileHandle  = $self->{"mapCacheFileHandle"};
	
	#add value to cache
	if ($action == 1){
		printf($mapCacheFileHandle "add\t%s\t%s\t%s\n", $timestamp, $prefix, $entriesInCache);
	}elsif($action == -1){
		printf($mapCacheFileHandle "del\t%s\t%s\t%s\n", $timestamp, $prefix, $entriesInCache);
	}
	
}

sub createOutputFile(){
	my $self=shift;
	
	my $outputFile = $self->{"outputFile"};
	
	open(OFILE, "> $outputFile") or die printf("Error opening the file %s associated to the ITR!\n", $outputFile);
	
	my $timeout = $self->{"timeout"};
	my $maxNbOfEntriesInCache = $self->{"maxNbOfEntriesInCache"};
	
	printf(OFILE "timeout\tmaxNbOfEntriesInCache\n");
	printf(OFILE "%s\t%s\n", $timeout, $maxNbOfEntriesInCache );
	
	$self->{"mapCacheFileHandle"} = \*OFILE;
	
}

sub DESTROY(){
	my $self=shift;
	
	printf("(MC) Closing mapCache file handle\n");
	my $mapCacheFileHandle = $self->{"mapCacheFileHandle"};
	close($mapCacheFileHandle);	
}

#return 1
1
