package CoreSim::Init;

use CoreSim::Chord;   
use CoreSim::Topology; 
use CoreSim::ALT; 
use CoreSim::TREE;  
use CoreSim::ITR::ITRNode;
use CoreSim::Scenario::Scenario1; 
use CoreSim::DB::DBI; 

use strict;

sub new(){
	my $invocant = shift;
	
	
	my ($xmlRPCServer, $filePath, $startPosition, $endPosition, $writeFingerTablesToFile, $readFingerTablesFromFile,
		$writeSortedAndFilteredPrefixListToFile, $readSortedAndFilteredPrefixListFromFile, $tracePath, $dbiPlaneFile, $dbEstimatorFile,
		$resolverType, $geoLiteCityFile,$path1, $path2, $simulationID, $_DEBUG ) = @_;
		
	my $class = ref($invocant) || $invocant;
	 
	my $self = {
		"xmlRPCServer" => $xmlRPCServer,
		"filePath"=> $filePath,
		"writeFingerTablesToFile"=>$writeFingerTablesToFile,
		"readFingerTablesFromFile" => $readFingerTablesFromFile,
		"startPosition"=>$startPosition, 
		"endPosition"=>$endPosition,
		"writeSortedAndFilteredPrefixListToFile"=>$writeSortedAndFilteredPrefixListToFile,
		"readSortedAndFilteredPrefixListFromFile"=>$readSortedAndFilteredPrefixListFromFile,
		"tracePath" => $tracePath,
		"dbiPlaneFile" => $dbiPlaneFile,
		"dbEstimatorFile" => $dbEstimatorFile,
		"resolverType" => $resolverType,
		"geoLiteCityFile" => $geoLiteCityFile,
		"inputFilePath" => $path1,
		"outputFilePath" => $path2,
		"simulationID" => $simulationID,
		"_DEBUG"=>$_DEBUG,
	};

	&buildDBI($self);
	&buildTopology($self);
	
	bless($self, $class);
	return $self; 	
}

sub buildTopology(){
	
	my $self= shift;
	my $xmlRPCServer = $self->{"xmlRPCServer"};
	my $filepath = $self->{"filePath"};
	my $dbiPlane = $self->{"dbiPlane"};
	my $dbEstimator = $self->{"dbEstimator"};
	my $geoLiteCityFile = $self->{"geoLiteCityFile"};
	my $_DEBUG = $self->{"_DEBUG"};
	
	my $topology = CoreSim::Topology->new($xmlRPCServer, $filepath, $dbiPlane,$dbEstimator,$geoLiteCityFile, $_DEBUG);
	
	$self->{"topology"} = $topology;
	
}

sub buildDBI(){
	my $self=shift;
	
	my $dbiPlaneFile = $self->{"dbiPlaneFile"}; 
	my $dbEstimatorFile = $self->{"dbEstimatorFile"};
	
	my $dbiPlane = CoreSim::DB::DBI->new($dbiPlaneFile);
	my $dbEstimator = CoreSim::DB::DBI->new($dbEstimatorFile);
	
	$self->{"dbiPlane"} = $dbiPlane;
	$self->{"dbEstimator"} = $dbEstimator;
}

#build a mapping system depenging on the choice expressed throught the variable $mappingSystemChoice
sub buildMappingSystem(){
	my ($self, $mappingSystemChoice) = @_;
	
	#common variables to all the mapping systems
	my $topology = $self->{"topology"};
	my $resolverType = $self->{"resolverType"};
	my $outputFilePath = $self->{"outputFilePath"};
	my $_DEBUG = $self->{"_DEBUG"};
	
	if ($mappingSystemChoice == 0){
		
		#variables common only to Chord
		my $maxNodesInFingerTable = 32; 
		my $maxNodesInChordRing = 2**$maxNodesInFingerTable;
		my $fingerTablesFile = $self->{"readFingerTablesFromFile"};
		my $startPosition = $self->{"startPosition"};
		my $endPosition = $self->{"endPosition"};
		my $sortedAndFilteredPrefixListFile = $self->{"readSortedAndFilteredPrefixListFromFile"}; 	
		
		my $chord = CoreSim::Chord->new($topology, $maxNodesInChordRing, $maxNodesInFingerTable, $fingerTablesFile, $startPosition, 
			$endPosition,$sortedAndFilteredPrefixListFile, $resolverType,$outputFilePath, $_DEBUG);
									
		$self->{"mappingSystem"} = $chord;

	}elsif($mappingSystemChoice == 1){
		
		#variables common only to ALT
		my $startAggregationLevel = 3;
		my $aggregationIndex = 5;
		
		my $alt = CoreSim::ALT->new($topology, $startAggregationLevel, $aggregationIndex,$outputFilePath, $_DEBUG);
		$self->{"mappingSystem"} = $alt;
	
	}elsif($mappingSystemChoice == 2){
		
		my $tree = CoreSim::TREE->new($topology, $outputFilePath, $resolverType, $_DEBUG);
		$self->{"mappingSystem"} = $tree;
		
	}
}

#builds an ITR independend of the mapping system and the associatedNode(which are given as parameters)
sub buildITR(){
	my ($self, $mappingSystem, $associatedNode) = @_;
	
	#These are teoretical values obtained by prof. Bonaventure (We are not using them for now)
	#timeout 	entries
	#3min		18000
	#30min		43500
	#300min		103000
	
	my $timeout = 180;
	my $maxNbOfEntriesInCache = 1800000;
	
	my $outputPath = $self->{"outputFilePath"};
	my $debug = $self->{"_DEBUG"};
	
	#for TREE i change the IP here
	$associatedNode = "158.109.244.1";
	
	
	my $itr = CoreSim::ITR::ITRNode->new($mappingSystem, $associatedNode, $timeout, $outputPath, $maxNbOfEntriesInCache, $debug);
	$self->{"itr"} = $itr;
	
	#	#this is used because the default ("best connected pop") is poorly (not at all in fact) mapped by iPlane
	#	my $newIP = "158.109.244.1";
	#	$self->{"mappingSystem"}->setNodeIP($associatedNode, $newIP );
}

#builds the mapping system and the ITR
sub buildMSAndITR(){
	my $self=shift;
	my ($mappingSystemChoice, $associatedNodePrefix) = @_;
	
	#build mapping system
	&buildMappingSystem($self, $mappingSystemChoice);
	
	my $mappingSystem = $self->{"mappingSystem"};
	
	#build the ITR
	my $associatedNode = $mappingSystem->getAssociatedNodeForPrefix($associatedNodePrefix);
	&buildITR($self, $mappingSystem, $associatedNode);
	
}

#runs the first scenario: takes packets from a trace file and "passes" them through one of the mappingSystems. Also the ITR cache
#is used
sub runScenario1(){
	my $self = shift;
	
	my $itr = $self->{"itr"};
	my $scenario1 = CoreSim::Scenario::Scenario1->new($itr, $self->{"tracePath"}, $self->{"_DEBUG"});
	$scenario1->runSimulation();
	
}

#chord specific file creation
#generate the sorted and filtered prefix list
sub generateChordSortedAndFilteredPrefixList(){
	my $self=shift;
	
	#build Chord Mapping System
	&buildMappingSystem($self, 0);
	
	my $chord = $self->{"mappingSystem"};
	my $outputFilePath = $self->{"outputFilePath"};
	
	my $outputFile = "$outputFilePath/sortedAndFilteredPrefixList.txt"; 
	$chord->dropSortedAndFilteredPrefixesToFile($outputFile);
}

#generate finger tables
sub generateChordFingerTables(){
	#...to be written
}

#debug only
sub chordAuxFilesGeneration(){
	my $self=shift;
	
	my $chord = $self->{"chord"};
	
	if (defined($self->{"writeSortedAndFilteredPrefixListToFile"}) ){
		printf("WRITING THE PREFIXES TO FILE %s!!!\n", $self->{"writeSortedAndFilteredPrefixListToFile"});
		$chord->dropSortedAndFilteredPrefixesToFile($self->{"writeSortedAndFilteredPrefixListToFile"});
		printf("DONE WITH THE WRITING OF THE PREFIXES!\n");
	}
	
	if (defined($self->{"writeFingerTablesToFile"}) ){
		printf("WRITING THE FINGER TABLE TO FILE %s!!!\n", $self->{"writeFingerTablesToFile"});
		$chord->dropFingerTablesToFile($self->{"writeFingerTablesToFile"});
		printf("DONE WITH THE WRITING OF THE FINGER TABLE!\n");
	}
	
#	$chord->dropFingerTablesToFile($fingerTablesFilePath);
#	printf("Chord. Latency for the search: %s\n",$chord->getDHTLookupDelayFromChordID(117440511,"5.0.0.0") );

#	my $traceFilePath = "/home/florin/workspace/eclipse/asociatepd/20071106-1300.ilatsim.txt";
#	my $scenario1 = iLatSim::Scenario::Scenario1->new($chord, $traceFilePath );
	
#	$scenario1->runSimulation();

}

sub topologyDebug(){
		my $topology = shift;
#	printf("Let's see:%d\n", $topology->getLatencyBetweenPOPs(116,32863));
#	printf("Let's see2:%d\n", $topology->getLatencyBetweenIPs("217.175.231.68","218.53.28.9") );
	
#	printf("Let's see2:%d\n", $topology->getLatencyBetweenIPs("211.225.21.1","66.39.173.154") );
	
#	&numberOfBorderPOPsPerAS($topology);
	
	srand(time ^ $$ ^ unpack "%L*", `ps axww ⎪ gzip`);
	my $ipList = $topology->getPOPIPList();
	my $numberOfIPs = scalar(@{$ipList});
	my $nbOfQueries = 2;
	my $answeredQueries = 0;
	my $unansweredQueries = 0;
	
	for(my $i=0;$i<$nbOfQueries;$i++){
		my $randNumber1 = int(rand($numberOfIPs));
		my $randNumber2 = int(rand($numberOfIPs));
		
		if ($randNumber1 ==  $randNumber2){
			$randNumber2--;
		}
		my $ip1 = $ipList->[$randNumber1];
		my $ip2 = $ipList->[$randNumber2];
		
		my $queryResponse = $topology->getLatencyBetweenIPs($ip1, $ip2);
		
		if ($queryResponse != -1){
			$answeredQueries++;
		}else{
			$unansweredQueries++;
		}
		
		printf("the latency between ip1=%s and ip2=%s is %s\n\n", $ip1, $ip2, $queryResponse);
	}
	
	printf("The number of queries is %s the number of answered queries %s and the number of unanswered queries %s\n", 
			$nbOfQueries, $answeredQueries, $unansweredQueries);
	
	
}

sub testTopology1(){
	
	my $topology = shift;
	my $inputfilePath = "/home/florin/workspace/eclipse/asociatepd/connectedPOPGroup.txt"; 
	
	
#	printf("my test query= %s and reversed %s \n",$topology->getLatencyBetweenIPs("213.137.187.97", "192.6.72.1"), $topology->getLatencyBetweenIPs("192.6.72.1", "213.137.187.97")  );
	
	
	srand(time);

	my $popIPList = $topology->getBorderPOPIPList();
	my $nbOfQueries = 100;
	my $numberOfPOPs = scalar(@$popIPList);
	
	my $answeredQueries=0;
	my $unansweredQueries=0;
	
#	open(OFILE, "> /home/florin/workspace/eclipse/asociatepd/ipPairs.txt") or die printf("Can't open file ipPairs.txt:%s",$! );
	open(OFILE, "> /home/florin/workspace/perl/unsolvedIPs.txt") or die printf("can't open file unsolvedIPs:%s", $!);
	
	for(my $i=0;$i<$nbOfQueries;$i++){
		my $randNumber1 = int(rand($numberOfPOPs));
		my $randNumber2 = int(rand($numberOfPOPs));
		
		if ($randNumber1 ==  $randNumber2){
			$randNumber2--;
		}
		
		my $pop1 = $popIPList->[$randNumber1];
		my $pop2 = $popIPList->[$randNumber2];
		
		my $ip1 = $topology->getIPForPOP($pop1);
		my $ip2 = $topology->getIPForPOP($pop2);
		
#		printf( OFILE "%s %s\n", $ip1, $ip2);
		
		my $queryResponse = $topology->getLatencyBetweenPOPIPs1($ip1, $ip2);
		
		if ($queryResponse != -1){
			$answeredQueries++;
		}else{
			$unansweredQueries++;
			printf(OFILE "%s\n%s\n", $ip1, $ip2);
		}
		
		printf("query number %s . The latency between ip1=%s and ip2=%s is %s\n\n",$i, $ip1, $ip2, $queryResponse);
	}
	
	printf("The number of queries is %s the number of answered queries %s and the number of unanswered queries %s\n", 
			$nbOfQueries, $answeredQueries, $unansweredQueries);
	
	close(OFILE);
	
}

sub testTopology2(){
	my $topology = shift;
	
	my $list = $topology->getBorderPOPIPList();	
}

sub numberOfBorderPOPsPerAS(){
	my $topology = shift;
	my $asBorderPOPs = $topology->getASBorderPOPs();
	foreach my $as (keys %$asBorderPOPs){
		printf("as=%s border pops=", $as);
		foreach my $pop (keys %{$asBorderPOPs->{$as} }){
			printf("%s ", $pop);
		}
		printf("\n");
	}

}

#return 1 at the end
return 1;