package CoreSim::X::XTopology;

use strict;
use CoreSim::X::XRouter;
use CoreSim::Utilities::IPUtilities;

sub new(){
	
	my $invocant = shift;
	
	my ($topology, $ipUtility, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	#this hash will hold all the routers except the root layer and have as keys the prefixes associated to them 
	my $routersHash = {};
	
	#this list contains all the routers at the root layer 
	my $rootLayer = [];
	
	#the mask for the TLD routers (10 gives 1024 TLD routers)
	my $rerMask = 8;
	
	my $numberOfRootRouters = 8;
	
	#number of IPs per root router
	my $numberOfIPsPerRootRouter = 2;
	
	my $self={
		"topology" => $topology,
		"routersHash" => $routersHash,
		"rootLayer" => $rootLayer,
		"ipUtility" => $ipUtility, 
		"rerMask" => $rerMask,
		"numberOfRootRouters" => $numberOfRootRouters,
		"numberOfIPsPerRootRouter" => $numberOfIPsPerRootRouter,
		"_DEBUG" => $_DEBUG,
	};
	
	bless($self, $class);
	
	&createTopology($self);
	
	return $self;
}

sub createTopology(){
	
	my $self=shift;
	
	#generate the 8 root servers
	&createRootLayer($self);

	#create the regional EID registries
	&createRER($self);
	
	&associateAllPOPs($self);
	
}

sub createRootLayer(){
	my $self=shift;
	
	my $rootLayer = $self->{"rootLayer"};
	my $_DEBUG = $self->{"_DEBUG"};
	my $ipUtility = $self->{"ipUtility"};
	my $routersHash = $self->{"routersHash"};
	
	my $numberOfRootRouters = $self->{"numberOfRootRouters"};
	
	printf("Creating the root layer..\n");
	#all the root nodes are the same they are used for redundancy. This creates a list of them
	for (my $i= 0; $i< $numberOfRootRouters; $i++ ){
		my $newXRouter = CoreSim::X::XRouter->new( $ipUtility, $_DEBUG);
		$newXRouter->setPrefix("0.0.0.0/0");
		$newXRouter->setRootName("rootRouter$i");
		$newXRouter->setLayer(1);
		
		#it is not a prefix but I need a way to locate the router
		$routersHash->{"rootRouter$i"} = $newXRouter;
		
		push (@$rootLayer, $newXRouter);
		
	}

}

sub createRER(){
	my $self=shift;
	
	my $routersHash = $self->{"routersHash"};
	my $ipUtility = $self->{"ipUtility"};
	my $rerMask = $self->{"rerMask"};
	my $numberOfRootRouters= $self->{"numberOfRootRouters"};
	my $rootLayer = $self->{"rootLayer"};
	my $prefixToRIR = $self->{"topology"}->getPrefixToRIR();
	my $rerLayer = {};
	
	printf("size of list %s\n", scalar(keys %$prefixToRIR));
	printf("Creating the RER layer ..\n");
	for (my $i=1; $i<2**$rerMask; $i++){
		
		#we ignore the value for i=0 as 0/8 is not a valid prefix (the ones above shoul be eliminated also, for now I don't use a
		#hardcoded value)
		my $value = $i<<(32-$rerMask);
		
		my $ip = $ipUtility->numberToIP($value);
		my $prefix = "$ip/$rerMask";
		
		if (!&firstEightBitsZero($prefix)){
			my $newXRouter = CoreSim::X::XRouter->new( $ipUtility, $self->{"_DEBUG"});
			$newXRouter->setPrefix($prefix);
			$newXRouter->setLayer(2);
			
			$routersHash->{$prefix} = $newXRouter;
			
			for (my $i=0;$i<$numberOfRootRouters;$i++){
				my $rootLayerRouter = $rootLayer->[$i];
				
				$rootLayerRouter->addChild($newXRouter);
			}
			
			printf("Added the RER node %s to the topology\n", $prefix) if ($self->{"_DEBUG"}>=2);
		}
		
	}
		
}

sub associateAllPOPs(){
	my $self=shift;
	
	#this will give IPs to the whole list of root routers
	&associatePOPsToRootRouters($self);
	
	#this will associated IPs TLD nodes , create the POP Level and associate child POPs to parent virtual routers
	&associatePOPs($self);
	
}

#this should be done by geographical criterion, for now it is done just by taking the best connected POPs and asociating them to 
#the routers 
sub associatePOPsToRootRouters(){
	my $self=shift;
	
	my $topology = $self->{"topology"};
	my $rootLayer = $self->{"rootLayer"};
	my $numberOfRootRouters = $self->{"numberOfRootRouters"};
	my $numberOfIPsPerRootRouter = $self->{"numberOfIPsPerRootRouter"};
	my $numberOfPOPsRequired = $numberOfRootRouters * $numberOfIPsPerRootRouter;
	
	my $popList = $topology->getBestConnectedNPOPs($numberOfPOPsRequired);
	
	for (my $i=0; $i<$numberOfPOPsRequired;$i+=$numberOfIPsPerRootRouter){
		for (my $j=0; $j<$numberOfIPsPerRootRouter; $j++){
			
			my $routerNumber = $i/$numberOfIPsPerRootRouter;
			my $rootRouter = $rootLayer->[$routerNumber];
			
			my $routerInfo = $popList->[$i+$j];
			$rootRouter->addRootRouterInfo($routerInfo);
			
		}
	}
	
}

sub associatePOPs(){
	my $self=shift;
	
	my $topology = $self->{"topology"};
	my $prefixList = $topology->getPrefixList();
	my $ipUtility = $self->{"ipUtility"};
	my $rerMask = $self->{"rerMask"};
	my $routersHash = $self->{"routersHash"};
	
	printf("Associating POPs to xRouters..\n");
	
	my $count  = 0 ;
	
	
	
	for ( my $maskIterator = $rerMask; $maskIterator<=32; $maskIterator++){
		
		foreach my $prefix (@$prefixList){
			
			my $mask = $ipUtility->getMaskFromPrefix($prefix);
			
			#build the topology from the smallest to the highest of masks (/8->/32)
			if ( $mask == $maskIterator ){
			
		#		printf("Working with prefix %s\n", $prefix);
				
				$count ++;
				
				if ($count % 10000 == 0) {
					printf("Done with %s prefixes\n", $count);
				}
				
					
				#if mask equals rerMask then we have an xRouter already created (it is a RER router)
				if($mask == $rerMask){
					
					#associate to the RER router the data of the child
					my $xRouter = $routersHash->{$prefix};
				
					#determine from the topology the localization information of the RER router 
					#this router is "endoresed" by the already existing border router of the AS with the RER prefix
					my $pop = $topology->getBorderPOPForPrefix($prefix);
					my $as = $topology->getASForPOP($pop);
					my $ip = $topology->getIPForPOP($pop);
					
					$xRouter->setLocalizationInformation($pop, $as, $ip);
					
					#create a new router which connects to the RER router. Practically this is a router supported by the same POP
					#as RER router but a distinction should be made here: I associated the same POP to the RER router because otherwise
					#no information is known about who will administer the routers/ their POP specific data
					my $newXRouter = CoreSim::X::XRouter->new($ipUtility, $self->{"_DEBUG"});
					$newXRouter->setPrefix($prefix);
					$newXRouter->setLayer(3);
					
					#set the localization information
					$newXRouter->setLocalizationInformation($pop, $as, $ip);
					
					#add the router to the routersHash
					$routersHash->{$prefix} = $newXRouter;
					
					#add the pop level router as a child to the TLD router
					$xRouter->addChild($newXRouter);
					
				
				#if ($mask> $rerMask)the mask is wider than the rerMask then we need to create a new xRouter and associate to whom ever 
				#is responsible for the parent prefix
				}elsif($mask>$rerMask){
					
					my $newXRouter = CoreSim::X::XRouter->new($ipUtility, $self->{"_DEBUG"});
					$newXRouter->setPrefix($prefix);
					
					#now it needs to find the parent prefix
					my $parentRouter = &findParentNodeForPrefix($self,$prefix, $rerMask);
					$parentRouter->addChild($newXRouter);
					$newXRouter->setLayer($parentRouter->getLayer() + 1);
					
					#find the pop specific information		
					my $pop = $topology->getBorderPOPForPrefix($prefix);
					my $as = $topology->getASForPOP($pop);
					my $ip = $topology->getIPForPOP($pop);
					
					#set the localization information
					$newXRouter->setLocalizationInformation($pop, $as, $ip);
					
					#add the router to the routersHash
					$routersHash->{$prefix} = $newXRouter;
					
					#if parent is virtual, the child node's pop enters in the topology as the pop responsible for the parent 
					if ($parentRouter->isVirtualRouter()){
						
						$parentRouter->setLocalizationInformation($pop, $as, $ip);
		
					}
					
				}
				
				
			}
			
		}
		
	}
}

sub findParentNodeForPrefix(){
	my ($self, $childPrefix, $rerMask) = @_;
	
	my $routersHash = $self->{"routersHash"};
	my $ipUtility = CoreSim::Utilities::IPUtilities->new();

	my $parentPrefix;
	
	my $mask = $ipUtility->getMaskFromPrefix($childPrefix);
	my $prefix = $childPrefix;
	my $endLoop=0;
	my $possibleParentPrefix;
	
	#search the hierarchy for a parent prefix
	while (($mask > 8) && (!$endLoop)){
		
		#get a possible parent prefix with a mask decreased by 1
		$possibleParentPrefix = $ipUtility->computeParentPrefix($prefix, 1);
		
		#if the parent does exist stop searching
		if (defined($routersHash->{$possibleParentPrefix})){
			$endLoop = 1;
		}
		
		$mask -= 1;
		$prefix = $possibleParentPrefix;
	}
	
#	printf("prefix = %s possible parent prefix = %s endLoop = %s\n",$childPrefix, $possibleParentPrefix, $endLoop );
	#if the parent is not found then post an error message. All the prefixes should have a parent in the RER level
	if ($endLoop==1){
		
		return $routersHash->{$possibleParentPrefix};
		
	}else{
		
		printf("(EE) There is no parent for the prefix %s!\n", $childPrefix);
		die;
	}
	
	
	
}

sub getRouterWithPrefix(){
	my $self = shift;
	my ($prefix) = @_;
	
	return $self->{"routersHash"}->{$prefix};
}

sub getRootLayer(){
	my $self=shift;
	
	return $self->{"rootLayer"};
}

sub firstEightBitsZero(){
	
	my $ip = shift;
	
	my @result = split(/\./, $ip);
	if ($result[0] == 0 ){
		return 1;
	}else{
		return 0;
	}
}

#return 1 at the end
1