package CoreSim::ALT::ALTMetrics;

use strict;

sub new(){
	my $invocant = shift;
	my ($topology, $altTopology,$ipUtility, $aggregationIndex, $_DEBUG) = @_;
	my $class = ref($invocant) || $invocant;
	
	my $localDelayMap = $altTopology->getLocalDelayMap();
	my $self={
		"topology" => $topology, 
		"altTopology" => $altTopology, 
		"ipUtility" => $ipUtility,
		"aggregationIndex" => $aggregationIndex,
		"localDelayMap" => $localDelayMap,
		"_DEBUG" => $_DEBUG,
	};
	
	bless($self, $class);
	
	return $self;
}

sub getALTLookupDelayFromALTRouter(){
	my $self= shift;
	my ($altRouterPrefix, $searchedIP) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	my $altTopology = $self->{"altTopology"};
	
	my $altRouter = $altTopology->getALTRouterForPrefix($altRouterPrefix);
	
	#the current router is responsible for the IP
	if ($ipUtility->ipInsidePrefix($searchedIP, $altRouter->getPrefix()  )){
		printf("The first ALTRouter %s is responsible for the IP %s! There is no search latency!\n", $altRouter->getPrefix(),$searchedIP);
		return 0;
	}
	
	my ($responsibleALTRouter, $path) = &getResponsibleALTRouter($self, $altRouter, $searchedIP);

	#if the responsibleALTRouter is found
	if (defined($responsibleALTRouter)){
		my $pathLength = scalar(@$path) -1;
		
		#compute the latency
		my ($pathLatency,$notFoundLatenciesForVirtualRouters,$notAnsweredXMLRPCQueries,$nbOfLocalDelayMapQueries )=&computePathLatency1($self, $path);
		
		#debug information
		printf("\nIn order to compute the latency implied by the search for IP %s starting at %s a number of %s paths have been checked\n", 
				$searchedIP, $altRouter->getPrefix(), $pathLength) if ($self->{"_DEBUG"});
		printf("For %s paths the latency is unknown due to virtual routers and for %s the path is unkown due to faild XMLRPC searches\n",
				$notFoundLatenciesForVirtualRouters, $notAnsweredXMLRPCQueries) if ($self->{"_DEBUG"});
		printf("For %s the latency is taken from the localDelayMap\n",$nbOfLocalDelayMapQueries) if ($self->{"_DEBUG"});
		
		return $pathLatency;
		
	}else{
		#or issue a waring message
		
		printf("(W) The altRouter responsible for the IP %s was not found!\n", $searchedIP) if ($self->{"_DEBUG"});
		return -1;
	}

}

#for solution 1&3
sub getResponsibleALTRouter(){
	my ($self, $altRouter, $searchedIP) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	
	my @path;
	push(@path, $altRouter);
	
	my $endLoop = 0;
	my $responsibleALTRouter=undef;

	
	#search for the router responsible for the searchedIP
	while (!$endLoop){
		
		my $nextHopALTRouter =   $altRouter->findNextHopForIP($searchedIP);
		
		#if the returned nextHopALTRouter is the same as the current router that means that the responsible altRuter is found
		if ($nextHopALTRouter->getPrefix() eq $altRouter->getPrefix()){
			
			#this is a double check to see if the router's prefix does include the searchedIP
			if ($ipUtility->ipInsidePrefix($searchedIP, $nextHopALTRouter->getPrefix() )){
		
				$responsibleALTRouter = $nextHopALTRouter;
				$endLoop =1 ;
				
			}else{
				
				printf("(EE) Something is wrong in the ALT search algorithm!\n");
				$endLoop=1;
				
			}
		#add the new router to the path
		}else{
			push(@path, $nextHopALTRouter);
		}
		
		$altRouter = $nextHopALTRouter;
	}
	
	return ($responsibleALTRouter, \@path);
	
}

#for solution 2 (not finished with the latencies)
sub getResponsibleALTRouter1(){
	
	my ($self, $altRouter, $searchedIP) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	my $altTopology = $self->{"altTopology"};
	
	my $l1Routers = $altTopology->getL1Routers();
	my @l1Prefixes = keys %$l1Routers;
	
	my $endLoop = 0;
	my $responsibleALTRouter = undef;
	my $root = $altTopology->getRoot();
	my @path;
	
	while (!$endLoop){
		
		my $nextHopALTRouter = $altRouter->findNextHopForIP($searchedIP);
		
		#if the "top"" of the ALT Topology is reached
		if ($nextHopALTRouter->getPrefix() eq $root->getPrefix() ){
			
			#search among the l1Routers for a "descending" path
			my $nextHop = undef;
			foreach my $prefix (@l1Prefixes){
				if ($ipUtility->ipInsidePrefix($searchedIP, $prefix)){
					$nextHop = $l1Routers->{$prefix};
					last;
				}
			}
			
			#if we find no prefix that includes the searchedIP ... we have a problem
			if (defined($nextHop)){
				$nextHopALTRouter = $nextHop;
			}else{
				printf("Problem with the 'routing' inside ALT at the layer 1\n");
			}
			
		}
		
		#this condition is checked only when the altRouter is responsible for that IP
		if ($nextHopALTRouter->getPrefix() eq $altRouter->getPrefix()){
			
			#this is a double check to see if the router's prefix does include the searchedIP
			if ($ipUtility->ipInsidePrefix($searchedIP, $nextHopALTRouter->getPrefix() )){
		
				$responsibleALTRouter = $nextHopALTRouter;
				$endLoop =1 ;
				
			}else{
				
				printf("(EE) Something is wrong in the ALT search algorithm!\n");
				$endLoop=1;
				
			}
		
		}else{
			push(@path, $nextHopALTRouter);
		}
		
		$altRouter = $nextHopALTRouter;
		
	}
	
	return ($responsibleALTRouter, \@path);
	
	
}

sub computePathLatency(){	
		my ($self, $path) = @_;
			
		my $localDelayMap = $self->{"localDelayMap"};
		my $topology = $self->{"topology"};
		
		#compute the latency
		my $totalLatency = 0;
		
		my $notFoundLatenciesForVirtualRouters = 0;
		my $notAnsweredXMLRPCQueries = 0;
		my $nbOfLocalDelayMapQueries = 0;
		
		my $pathLength = scalar(@$path);
		
		for(my $i=1;$i<$pathLength; $i++){
			
			my $router1 = $path->[$i-1];
			my $router2 = $path->[$i];
			
			if (&l1Router($self,$router1->getPrefix()) || &l1Router($self,$router2->getPrefix())){
				$totalLatency += $localDelayMap->{$router1->getPrefix()}{$router2->getPrefix()};
				$nbOfLocalDelayMapQueries++;
			}else{
				my $routerIP1 = $router1->getPOPIP();
				my $routerIP2 = $router2->getPOPIP();
				
				if (!defined($routerIP1) || !defined($routerIP2) ){
					printf("(W) Can't compute latency between the prefixes %s and %s. At least one of them is represented by a virtual router!\n", 
						$router1->getPrefix(), $router2->getPrefix());
						
					#increment the number of not found latencies
					$notFoundLatenciesForVirtualRouters++;
				}else{
					#use the topology module to find the latency between the POPs representing the routers
#					my $latencyQuery = $topology->getLatencyBetweenPOPIPs1($routerIP1, $routerIP2);
					
					#this done temporarily due to issues with the iplane server
					my $latencyQuery =1;
					printf("xmlrp between %s %s\n", $router1->getPrefix(), $router2->getPrefix());
					if ($latencyQuery == -1){
						$notAnsweredXMLRPCQueries++;
						printf("(W) The latency query for %s and %s failed!\n", $routerIP1, $routerIP2);
					}else{
						$totalLatency += $latencyQuery;
					}
					
					
				}
				
			}
		}
		return ($totalLatency,$notFoundLatenciesForVirtualRouters, $notAnsweredXMLRPCQueries,$nbOfLocalDelayMapQueries );
		
}

#this is rewritten to make use of the assumptions made when building the localDelayMap (for the first type of ALT Topology)
#solution 1&3
sub computePathLatency1(){
	
		my ($self, $path) = @_;
			
		my $localDelayMap = $self->{"localDelayMap"};
		my $topology = $self->{"topology"};
		
		#compute the latency
		my $totalLatency = 0;
		
		my $notFoundLatenciesForVirtualRouters = 0;
		my $notAnsweredXMLRPCQueries = 0;
		my $nbOfLocalDelayMapQueries = 0;
		
		my $pathLength = scalar(@$path);
		
		for(my $i=1;$i<$pathLength; $i++){
			
			my $router1 = $path->[$i-1];
			my $router2 = $path->[$i];
			
			#if the delay is defined in the localDelayMap use it
			if ( (defined($localDelayMap->{$router1->getPrefix()}{$router2->getPrefix()})) ) {
				$totalLatency += $localDelayMap->{$router1->getPrefix()}{$router2->getPrefix()};
				$nbOfLocalDelayMapQueries++;
				
			#otherwise use the the XMLRPC query
			}else{
				my $routerIP1 = $router1->getPOPIP();
				my $routerIP2 = $router2->getPOPIP();
				
				#use the topology module to find the latency between the POPs representing the routers
				my $latencyQuery = $topology->getLatencyBetweenPOPIPs1($routerIP1, $routerIP2);
				
				if ($latencyQuery == -1){
					$notAnsweredXMLRPCQueries++;
					printf("(W) The latency query for %s and %s failed! The prefixes implied: %s and %s\n", 
							$routerIP1, $routerIP2, $router1->getPrefix(), $router2->getPrefix());
				}else{
					$totalLatency += $latencyQuery;
				}
					

				
			}
		}
		return ($totalLatency,$notFoundLatenciesForVirtualRouters, $notAnsweredXMLRPCQueries,$nbOfLocalDelayMapQueries );
		
}

sub l1Router(){	
	my ($self,$prefix) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	
	#if the mask is equal to the aggregation index then it is a L1 router
	if ($ipUtility->getMaskFromPrefix($prefix) eq $self->{"aggregationIndex"}){
		return 1;
	}else{
		return 0;
	}
	
}

#return 1
1