package CoreSim::ALT::ALTTopology;

use strict;
use CoreSim::ALT::ALTRouter;
use CoreSim::Utilities::IPUtilities;

sub new(){
	my $invocant = shift;
	my ($topology, $aggregationIndex,$ipUtility, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	my $firstAggregationLevel = 20;
	
	#this is the root node
	my $root = CoreSim::ALT::ALTRouter->new(undef,$ipUtility, $_DEBUG);
	$root->setPrefix("0.0.0.0/0");
	
	my $altRoutersHash={};
	my $localDelayMap={};
	
	#values which are delays between the virtual L1 routers (there are no masks lower than /8 in the internet (for now)) 
	#and other L1 or L2 routers
	my $L1toL1Delay = 10;
	my $L1toL2Delay = 100;
	
	my $self={
		"topology" => $topology,
		"aggregationIndex" => $aggregationIndex,
		"altRoutersHash" => $altRoutersHash,
		"ipUtility" => $ipUtility,
		"firstAggregationLevel"=> $firstAggregationLevel,
		"localDelayMap" => $localDelayMap,
		"root" => $root,
		"L1toL1Delay" => $L1toL1Delay,
		"L1toL2Delay" => $L1toL2Delay,
		"_DEBUG" => $_DEBUG,
	};
	
	bless($self, $class);
	
	&createALTTopology($self);
	return $self;
}

sub createALTTopology(){
	my $self= shift;
	
	my $topology = $self->{"topology"};
	
	printf("Creating the ALT Topology\n") if ($self->{"_DEBUG"});
	
##########################################
#solution 1	
###	create the aggregation levels and then introduce the prefixes into this topology
#	my $firstAggregationLevel = $self->{"firstAggregationLevel"};
#	my $aggregationIndex = $self->{"aggregationIndex"};
#	my $nbOfLevels = $firstAggregationLevel/$aggregationIndex;
#	&createALTAggregators($self, $aggregationIndex, $nbOfLevels);
#	&addALTRoutersToTheTopology($self);
################################


##############################################
#solution 2
#	create the topology out of prefixes only (huge L1)
#	&createALTRouters($self);

###############################################

##############################################

#solution 3
#mixture from the first two. create 2 layers of aggregators and connect the rest of the prefixes in a "tree fashion" and use 
#and aggregationIndex of 5;
	&createMixedTopology($self, 2, 5);
	
###################################################

}

#############################################################
#this is solution 1 and it might fail due to the impossibility of *guessing* the *position* of some ALT Routers for which we do not have
#an overlapping prefix

#this variant uses aggregationIndex and aggregationLevels
sub createALTAggregators(){
	my ($self, $aggregationIndex, $nbOfLevels)=@_;
	
	my $topology = $self->{"topology"};
	my $debug = $self->{"_DEBUG"};
	
	my $filteredPrefixList = $topology->getFilteredPrefixList();
	
	my $ipUtility = $self->{"ipUtility"};
	my $altRoutersHash;
	my $root = $self->{"root"};
	
	my $localDelayMap = $self->{"localDelayMap"};

	my $previousRouterLevel;
	my $currentRouterLevel;
	my $currentAggregation=0;
	
	printf("Creating the Aggregators!\n") if ($self->{"_DEBUG"});
	
	#create the aggregators
	#for all the levels
	for (my $i=0; $i< $nbOfLevels; $i++){
		$currentAggregation += $aggregationIndex; 
		
		#the old currentRouterLevel becomes the previousRouterLevel
		$previousRouterLevel = $currentRouterLevel;
		
		printf("\n The current aggregation level %s\n", $currentAggregation) if ($self->{"_DEBUG"});

		if (defined($previousRouterLevel)){
			
			my $count = 0;
			$currentRouterLevel = ();
			
			#for all the previously introduced routers
			foreach my $element ( @{$previousRouterLevel} ){
				
				#increment their prefix value with the value determined by the currentAggregationLevel
				for (my $j = 0; $j<(2**$aggregationIndex) ; $j++){
					
					my $value = $element->{"value"} + ( $j<<(32-$currentAggregation) );
			
						
					#compute the router prefix
					my $tmpIP = $ipUtility->numberToIP($value);
					my $tmpMask = $currentAggregation;
					my $tmpPrefix = "$tmpIP/$tmpMask";
					
					#do not add prefixes of the form 0.* to the alt topology 
					if (! &firstEightBitsZero($tmpIP)){
						
						my $parentALTRouter = $element->{"node"};
						
						#create the element
						my $altRouter = CoreSim::ALT::ALTRouter->new($element->{"node"}, $ipUtility, $debug);
						$altRouter->setPrefix($tmpPrefix);
						
						#add the altRouter as a child to the parentALTRouter
						$parentALTRouter->addChild($altRouter);
						
						#add the router to the altRouterHash (we can find the routers by their prefix)
						$altRoutersHash->{$tmpPrefix} = $altRouter;
						
						#adding the node to the routerLevel
						my %tmpHash;
						$tmpHash{"node"} = $altRouter;
						$tmpHash{"value"} = $value;
						$currentRouterLevel->[$count]=\%tmpHash;
						$count++;
						
						printf("Adding prefix %s with the parent %s\n", $tmpPrefix,$parentALTRouter->getPrefix())
								if ($self->{"_DEBUG"}>=3);
						
#						#this is L2 thus the delays to the parents must be set(the parents are virtual routers)
#						if ($currentAggregation == 2*$aggregationIndex){
#							$localDelayMap->{ $tmpPrefix }{ $parentALTRouter->getPrefix() } = $L1toL2Delay;
#							$localDelayMap->{ $parentALTRouter->getPrefix() }{ $tmpPrefix } = $L1toL2Delay;
#						}
					
						#using the parent and the child complete the localDelayMap if necessary(one of the two routers is a virtual router) 
						#these are ASSUMPTIONS!
						&completeLocalDelayMap($self, $parentALTRouter, $altRouter);
					}
				}
			}
		

		
			
		#these are the routers in the first level
		}else{
			for (my $j =0; $j<(2**$aggregationIndex); $j++){
				my $value = $j<<(32-$currentAggregation);
				
				#create the element
				my $altRouter = CoreSim::ALT::ALTRouter->new($root, $ipUtility, $debug);
				
				#set the router prefix
				my $tmpIP = $ipUtility->numberToIP($value);
				my $tmpMask = $currentAggregation;
				
				my $tmpPrefix = "$tmpIP/$tmpMask";
				$altRouter->setPrefix($tmpPrefix);
				
				$root->addChild($altRouter);
				
				#add the router to the altRouterHash (we can find the routers by their prefix)
				$altRoutersHash->{$tmpPrefix} = $altRouter;
				
				#create the level
				my %tmpHash;
				$tmpHash{"node"} = $altRouter;
				$tmpHash{"value"} = $value;
				$currentRouterLevel->[$j]=\%tmpHash;
				
				printf("Adding prefix %s with the parent root\n", $tmpPrefix) if ($self->{"_DEBUG"}>=3);
				
			}
			#set the delays within the L1 layer (i call L1 the first layer "under" root)
			&createL1Mesh($self, $currentRouterLevel);
		}
		
		
	}
	
	$self->{"altRoutersHash"} = $altRoutersHash;
}

sub addALTRoutersToTheTopology(){
	my $self = shift;
	
	my $root = $self->{"root"};	
	my $altRoutersHash = $self->{"altRoutersHash"};
	my $topology = $self->{"topology"};
	my $ipUtility = $self->{"ipUtility"};
	my $debug = $self->{"_DEBUG"};
	
	my $prefixList = $topology->getFilteredPrefixList();
	my $aggregationIndex = $self->{"aggregationIndex"};
	
	printf("Creating the ALT routers out of prefixes!\n") if ($self->{"_DEBUG"});
	
	#for all the prefixes create altRouters and add them to the topology
	foreach my $prefix (@$prefixList){
		
		my $nbOfBitsToParentAggregationLevel = &findNbOfBitsToParentAggregationLevel($self, $prefix, $aggregationIndex);
		
		my $possibleParentPrefix = $ipUtility->computeParentPrefix($prefix,$nbOfBitsToParentAggregationLevel );
		my $possibleParentRouter = $altRoutersHash->{$possibleParentPrefix};
		
		#if the parent altRouter exists add the new altRouter as a child
		if (defined($possibleParentRouter)){ 
			
			#if the prefix of the parent is the same as the prefix that is to be added to the topology 
			if ($possibleParentPrefix eq $prefix){
				
				#set the pop parameters on the possibleParentRouter(which happens to overlap with the prefix from the prefixList)
				my $pop = $topology->getBorderPOPForPrefix($prefix);
				my $popIP = $topology->getIPForPOP($pop);
				
				$possibleParentRouter->setPrefix($prefix);
				$possibleParentRouter->setPOPNumber($pop);
				$possibleParentRouter->setPOPIP($popIP);
				
				printf("Prefix %s is at an aggregation level. Adding the router parameters\n", $prefix) 
						if ($self->{"_DEBUG"}>=3);
				
			}else{
				#create the new child altRouter and set its associated pop parameters
				my $altRouter = CoreSim::ALT::ALTRouter->new($possibleParentRouter, $ipUtility, $debug);
				$altRouter->setPrefix($prefix);
				$possibleParentRouter->addChild($altRouter);
				
				my $pop = $topology->getBorderPOPForPrefix($prefix);
				my $popIP = $topology->getIPForPOP($pop);
				
				$altRouter->setPOPNumber($pop);
				$altRouter->setPOPIP($popIP);
				printf("\tPrefix %s is a child of %s. Adding the router parameters\n", $prefix, $possibleParentPrefix) 
						if ($self->{"_DEBUG"}>=3);
						
				#using the parent and the child complete the localDelayMap if necessary(one of the two routers is a virtual router) 
				#these are ASSUMPTIONS!
				&completeLocalDelayMap($self, $possibleParentRouter, $altRouter);
			}
			
		}else{
			printf("(W) The prefix %s has the parent %s but no parent router was found!\n", $prefix, $possibleParentPrefix);
		}
	}
}

sub completeLocalDelayMap(){
	my ($self,$parentALTRouter, $childALTRouter) = @_;
	
	my $virtualParentExistingChildLatency = 10;
	my $existingParentVirtualChildLatency = 1;
	my $virtualParentVirtualChildLatency = 10;
	my $localDelayMap = $self->{"localDelayMap"};
	
	if ( (&virtualRouter($self, $parentALTRouter)) && (&virtualRouter($self, $childALTRouter)) ){
		
		$localDelayMap->{$parentALTRouter->getPrefix()}{$childALTRouter->getPrefix()} = $virtualParentVirtualChildLatency;
		$localDelayMap->{$childALTRouter->getPrefix()}{$parentALTRouter->getPrefix()} = $virtualParentVirtualChildLatency;
		
	}elsif( !(&virtualRouter($self, $parentALTRouter)) && (&virtualRouter($self, $childALTRouter)) ){
		
		$localDelayMap->{$parentALTRouter->getPrefix()}{$childALTRouter->getPrefix()} = $existingParentVirtualChildLatency;
		$localDelayMap->{$childALTRouter->getPrefix()}{$parentALTRouter->getPrefix()} = $existingParentVirtualChildLatency;
		
	}elsif( (&virtualRouter($self, $parentALTRouter)) && !(&virtualRouter($self, $childALTRouter)) ){
		
		$localDelayMap->{$parentALTRouter->getPrefix()}{$childALTRouter->getPrefix()} = $virtualParentExistingChildLatency;
		$localDelayMap->{$childALTRouter->getPrefix()}{$parentALTRouter->getPrefix()} = $virtualParentExistingChildLatency;
		
	}
	
}

sub virtualRouter(){
	my ($self,$altRouter) = @_;
	
	my $topology = $self->{"topology"};
	my $prefix = $altRouter->getPrefix();
	my $pop = $topology->getBorderPOPForPrefix($prefix);
	
	#if there is no pop for this prefix then this router is virtual
	if (!defined($pop)){
		return 1;
		
	#otherwise it is a real (existing, known about) router
	}else{
		return 0;
	}
}

sub findNbOfBitsToParentAggregationLevel(){
	my $self = shift;
	my ($prefix, $aggregationIndex) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	
	my $mask = $ipUtility->getMaskFromPrefix($prefix);
	my $firstAggregationLevel = $self->{"firstAggregationLevel"};
	
	if ($mask<$firstAggregationLevel){
		my $nbOfBits = $mask - $aggregationIndex* ( int($mask/$aggregationIndex) );
		return $nbOfBits;
		
	}else{
		my $nbOfBits = $mask - $firstAggregationLevel;
		return $nbOfBits;
	}
	
}

sub createL1Mesh(){
	my ($self, $routerLevel) = @_;
	
	my $nbOfRouters = scalar(@$routerLevel);
	my $localDelayMap = $self->{"localDelayMap"};
	my $L1toL1Delay = $self->{"L1toL1Delay"};
	
	for (my $i=0;$i<$nbOfRouters ; $i++){
		for (my $j=0; $j<$nbOfRouters; $j++){
			if ($i!=$j){
				my $router1 = $routerLevel->[$i]->{"node"};
				my $router2 = $routerLevel->[$j]->{"node"};
				$localDelayMap->{$router1->getPrefix()}{$router2->getPrefix()} = $L1toL1Delay;
				
				$router1->addFIBEntry($router2->getPrefix, $router2);
			}
			
		}
	}
}

sub firstEightBitsZero(){
	
	my $ip = shift;
	
	my @result = split(/\./, $ip);
	if ($result[0] == 0 ){
		return 1;
	}else{
		return 0;
	}
}


###################################################################################################
#solution 2. It tries to build the topology just from known prefixes. Practically we associate an ALT Router to each prefix
#!!!!!CAN be used as a replacement for the filtering algorithm! WAY faster!DO NOT DELETE!

#unfinished, need to add pop specific information to the existing routers
sub createALTRouters(){
	my $self = shift;
	
	my $topology = $self->{"topology"};
	
	#we use the prefix list (not the filtered one)
	my $prefixList = $topology->getPrefixList();
	
	my $ipUtility = $self->{"ipUtility"};
	my $debug = $self->{"_DEBUG"}; 
	
	my $altRoutersHash = $self->{"altRoutersHash"};
	my $l1Routers;
	
	my $root = $self->{"root"};
	
	# $i will take all the possible values for the prefix masks. From the prefixFile I noticed that there are no prefixes more general 
	#than /8 or more specific than /30

	my $prefixesWithParents = 0;
	for (my $i=8; $i<= 32; $i++){
		
		printf("Adding altRouters responsible for prefixes \/%s\n", $i);
		
		foreach my $prefix (@$prefixList){
			
			if ($ipUtility->getMaskFromPrefix($prefix) == $i){
					
				#search for the parent router (the min prefix is 8)
				my $parentALTRouter = &findParentNodeForPrefix($self, $prefix, 8, $altRoutersHash);
				
				#create alt router
				my $altRouter = CoreSim::ALT::ALTRouter->new($parentALTRouter, $ipUtility, $debug);
				$altRouter->setPrefix($prefix);
				$altRoutersHash->{$prefix} = $altRouter;
				
				#set the altRouter as a child for the parent	
				$parentALTRouter->addChild($altRouter);
				
#				printf("adding prefix %s with parent %s\n", $prefix, $parentALTRouter->getPrefix());
				
				#if the parent is the root node then add the altRouter to the l1LayerRouters
				if ($parentALTRouter->getPrefix() eq $root->getPrefix()){
					
					$l1Routers->{$prefix} = $altRouter; 
					
				}else{
					$prefixesWithParents++;
					
				}		
			}
			
		}
	}
	
	$self->{"l1Routers"} = $l1Routers;
	printf("Prefixes with parents = %s\n", $prefixesWithParents);
	
}

#some changes are needed for this to be used by the 3rd solution and I want solution 2 to still work 
sub createALTRouters1(){
	my $self = shift;
	
	my $topology = $self->{"topology"};
	
	#we use the prefix list (not the filtered one)
	my $prefixList = $topology->getPrefixList();
	my $ipUtility = $self->{"ipUtility"};
	my $debug = $self->{"_DEBUG"}; 
	
	my $altRoutersHash = $self->{"altRoutersHash"};
	my $l1Routers;
	
	my $root = $self->{"root"};
	
	# $i will take all the possible values for the prefix masks. From the prefixFile I noticed that there are no prefixes more general 
	#than /8 or more specific than /30

	my $prefixesWithParents = 0;
	for (my $i=8; $i<= 32; $i++){
		
		printf("Adding altRouters responsible for prefixes \/%s\n", $i);
		
		foreach my $prefix (@$prefixList){
			
			if ($ipUtility->getMaskFromPrefix($prefix) == $i){
				
				#this happends if the prefix is among those in the first 2 layers (no parent involved)
				if ( defined($altRoutersHash->{$prefix}) ){
					my $altRouter = $altRoutersHash->{$prefix};
					
					#set pop specific information for the altRouter
					my $pop = $topology->getBorderPOPForPrefix($prefix);
					my $popIP = $topology->getIPForPOP($pop);
					
					$altRouter->setPOPNumber($pop);
					$altRouter->setPOPIP($popIP);	
					
				#otherwise, if the altRouter added is new to the topology search for its parent
				}else{
					
					#search for the parent router (the min prefix is 8)
					my $parentALTRouter = &findParentNodeForPrefix($self, $prefix, 8, $altRoutersHash);
					
					#create alt router
					my $altRouter = CoreSim::ALT::ALTRouter->new($parentALTRouter, $ipUtility, $debug);
					$altRouter->setPrefix($prefix);
					
					#add the router to the altRoutersHash
					$altRoutersHash->{$prefix} = $altRouter;
					
					#set the altRouter as a child for the parent	
					$parentALTRouter->addChild($altRouter);
					
					#set pop specific information for the altRouter
					my $pop = $topology->getBorderPOPForPrefix($prefix);
					my $popIP = $topology->getIPForPOP($pop);
					
					$altRouter->setPOPNumber($pop);
					$altRouter->setPOPIP($popIP);	
					
					#complete the localDelayMap. If one of the routers is virtual we need to add a virtual link latency
					&completeLocalDelayMap($self,$parentALTRouter, $altRouter);
				}	
			}
			
		}
	}
	
	$self->{"l1Routers"} = $l1Routers;
	printf("Prefixes with parents = %s\n", $prefixesWithParents);
	
}

sub findParentNodeForPrefix(){
	my $self=shift;
	my ($childPrefix, $minMask, $altRoutersHash) = @_;
	
	my $ipUtility = CoreSim::Utilities::IPUtilities->new();

	my $endLoop =0;
	my $possibleParentPrefix;
	
	my $mask = $ipUtility->getMaskFromPrefix($childPrefix);
	my $prefix = $childPrefix;
	
	#search the hierarchy for a parent prefix
	while (($mask > $minMask) && (!$endLoop)){
		
		#get a possible parent prefix with a mask decreased by 1
		$possibleParentPrefix = $ipUtility->computeParentPrefix($prefix, 1);
		
		#if the parent does exist stop searching
		if (defined($altRoutersHash->{$possibleParentPrefix})){
			$endLoop = 1;
		}
		
		$mask -= 1;
		$prefix = $possibleParentPrefix;
	}
	
#	printf("prefix = %s possible parent prefix = %s endLoop = %s\n",$childPrefix, $possibleParentPrefix, $endLoop );
	#if the parent is not found then root should be the parent?
	if ($endLoop==0){
		return $self->{"root"};
	}else{
		return $altRoutersHash->{$possibleParentPrefix};
	}
	
}

##############################################################################################
#solution 3: 2 layers of virtual routers

#it creates N aggregation layers but N=2 for now as issues arise when N>2 due to "intermediate layer" unsolved problem:
# I don't know how to interleave /9 between /8 and /12 aggregation layers
sub createMixedTopology(){
	my ($self, $nbOfLevels, $aggregationIndex)= @_;
		
	#create the 2 layers of aggregators
	&createALTAggregators($self, $aggregationIndex, $nbOfLevels);
	
	#add all the routers responsible for the prefixes to the ALT Topology
	&createALTRouters1($self);
}

##############################################################################################

sub getLocalDelayMap(){
	my $self=shift;
	
	return $self->{"localDelayMap"};
}

sub getALTRouterForPrefix(){
	my $self = shift;
	my $prefix = shift;
	
	return $self->{"altRoutersHash"}->{$prefix};	
}

sub getRoot(){
	my $self= shift;
	
	return $self->{"root"};
}

sub getL1Routers(){
	my $self= shift;
	
	return $self->{"l1Routers"};
}

#return 1 at the end
1