package CoreSim::X::XTopology;

use strict;
use CoreSim::X::XRouter;
use CoreSim::Utilities::IPUtilities;

sub new(){
	
	my $invocant = shift;
	
	my ($topology, $ipUtility, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	#this hash will hold all the routers except the root layer and have as keys the prefixes associated to them 
	my $routersHash = {};
	
	#this list contains all the routers at the root layer 
	my $rootLayer = [];
	
	#the mask for the TLD routers (10 gives 1024 TLD routers)
	my $tldMask = 10;
	
	my $numberOfRootRouters = 8;
	
	#number of IPs per root router
	my $numberOfIPsPerRootRouter = 2;
	
	my $self={
		"topology" => $topology,
		"routersHash" => $routersHash,
		"rootLayer" => $rootLayer,
		"ipUtility" => $ipUtility, 
		"tldMask" => $tldMask,
		"numberOfRootRouters" => $numberOfRootRouters,
		"numberOfIPsPerRootRouter" => $numberOfIPsPerRootRouter,
		"_DEBUG" => $_DEBUG,
	};
	
	bless($self, $class);
	
	&createTopology($self);
	
	return $self;
}

sub createTopology(){
	
	my $self=shift;
	
	#generate the 8 root servers
	&createRootLayer($self);

	#create the regional EID registries
	&createRER($self);
	
	&associateAllPOPs($self);
	
}

sub createRootLayer(){
	my $self=shift;
	
	my $rootLayer = $self->{"rootLayer"};
	my $_DEBUG = $self->{"_DEBUG"};
	my $ipUtility = $self->{"ipUtility"};
	my $routersHash = $self->{"routersHash"};
	
	my $numberOfRootRouters = $self->{"numberOfRootRouters"};
	
	printf("Creating the root layer..\n");
	#all the root nodes are the same they are used for redundancy. This creates a list of them
	for (my $i= 0; $i< $numberOfRootRouters; $i++ ){
		my $newXRouter = CoreSim::X::XRouter->new( $ipUtility, $_DEBUG);
		$newXRouter->setPrefix("0.0.0.0/0");
		$newXRouter->setRootName("rootRouter$i");
		$newXRouter->setLayer(1);
		
		#it is not a prefix but I need a way to locate the router
		$routersHash->{"rootRouter$i"} = $newXRouter;
		
		push (@$rootLayer, $newXRouter);
		
	}

}

sub createRER(){
	my $self=shift;
	
	my $routersHash = $self->{"routersHash"};
	my $ipUtility = $self->{"ipUtility"};
	my $tldMask = $self->{"tldMask"};
	my $numberOfRootRouters= $self->{"numberOfRootRouters"};
	my $rootLayer = $self->{"rootLayer"};
	
	printf("Creating the TLD layer ..\n");
	for (my $i=1; $i<2**$tldMask; $i++){
		
		#we ignore the value for i=0 as 0/10 is not a valid prefix (the ones above shoul be eliminated also, for now I don't use a
		#hardcoded value)
		my $value = $i<<(32-$tldMask);
		
		my $ip = $ipUtility->numberToIP($value);
		my $prefix = "$ip/$tldMask";
		
		if (!&firstEightBitsZero($prefix)){
			my $newXRouter = CoreSim::X::XRouter->new( $ipUtility, $self->{"_DEBUG"});
			$newXRouter->setPrefix($prefix);
			$newXRouter->setLayer(2);
			
			$routersHash->{$prefix} = $newXRouter;
			
			for (my $i=0;$i<$numberOfRootRouters;$i++){
				my $rootLayerRouter = $rootLayer->[$i];
				
				$rootLayerRouter->addChild($newXRouter);
			}
			
			printf("Added the TLD node %s to the topology\n", $prefix);
		}
		
	}
		
}

sub associateAllPOPs(){
	my $self=shift;
	
	#this will give IPs to the whole list of root routers
	&associatePOPsToRootRouters($self);
	
	#this will associated IPs TLD nodes , create the POP Level and associate child POPs to parent virtual routers
	&associatePOPs($self);
	
}

#this should be done by geographical criterion, for now it is done just by taking the best connected POPs and asociating them to 
#the routers 
sub associatePOPsToRootRouters(){
	my $self=shift;
	
	my $topology = $self->{"topology"};
	my $rootLayer = $self->{"rootLayer"};
	my $numberOfRootRouters = $self->{"numberOfRootRouters"};
	my $numberOfIPsPerRootRouter = $self->{"numberOfIPsPerRootRouter"};
	my $numberOfPOPsRequired = $numberOfRootRouters * $numberOfIPsPerRootRouter;
	
	my $popList = $topology->getBestConnectedNPOPs($numberOfPOPsRequired);
	
	for (my $i=0; $i<$numberOfPOPsRequired;$i+=$numberOfIPsPerRootRouter){
		for (my $j=0; $j<$numberOfIPsPerRootRouter; $j++){
			
			my $routerNumber = $i/$numberOfIPsPerRootRouter;
			my $rootRouter = $rootLayer->[$routerNumber];
			
			my $routerInfo = $popList->[$i+$j];
			$rootRouter->addRootRouterInfo($routerInfo);
			
		}
	}
	
}

sub associatePOPs(){
	my $self=shift;
	
	my $topology = $self->{"topology"};
	my $prefixList = $topology->getPrefixList();
	my $ipUtility = $self->{"ipUtility"};
	my $tldMask = $self->{"tldMask"};
	my $routersHash = $self->{"routersHash"};
	
	printf("Associating POPs to xRouters..\n");
	my $count  = 0 ;
	
	foreach my $prefix (@$prefixList){
		
#		printf("Working with prefix %s\n", $prefix);
		
		my $mask = $ipUtility->getMaskFromPrefix($prefix);
		$count ++;
		
		if ($count % 10000 == 0) {
			printf("Done with %s prefixes\n", $count);
		}
		
		#we need to split wider prefixes (the ones with the mask shorter than the tldMask ) across more than one TLD
		if (($mask< $tldMask) && ($mask>=8) ){
			
			#create new XRouter and set the prefix the wider prefix BUT announce more than one more specific prefix 
			#to more than one TLD router. See the example below
			
			#create the router and set the prefix and layer
			my $newXRouter = CoreSim::X::XRouter->new($ipUtility, $self->{"_DEBUG"});
			$newXRouter->setPrefix($prefix);
			$newXRouter->setLayer(3);
			
			#find the pop information		
			my $pop = $topology->getBorderPOPForPrefix($prefix);
			my $as = $topology->getASForPOP($pop);
			my $ip = $topology->getIPForPOP($pop);
			
			#populate the pop information
			$newXRouter->setLocalizationInformation($pop, $as, $ip);
			
			#add the router to the routerHash
			$routersHash->{$prefix} = $newXRouter;
			
			#prepare the "split" algorithm
			my $maskDifference = $tldMask - $mask;
			
			my $prefixToNumber = $ipUtility->getPrefixToNumber($prefix);
			
			
			#algorithm to split the wider prefix into smaller ones 
			#1/8 goes to 1.0/10 1.64/10 1.128/10 1.192/10 
			#the algoritm does the following: 
			#			1/8 + 0<<22 -> 1.0/10
			#			1/8 + 1<<22 -> 1.64/10
			#			1/8 + 2<<22 -> 1.128/10
			#			1/8 + 3<<22 -> 1.192/10
			#the 4 added values (1..3) << 22 are obtained from 2**maskDifference
			
			for (my $i=0; $i< 2**$maskDifference; $i++){
				
				my $newIPNumber = $prefixToNumber + ( $i<<(32-$tldMask) );
				my $newIP = $ipUtility->numberToIP($newIPNumber);
				
#				printf("new ip %s, prefixToNumber %s,  i=%s se aduna %s\n",$newIP, $prefixToNumber, $i, $i<<(32-$tldMask));
				
				#find the parent xRouter
				my $announcedPrefix = "$newIP/$tldMask";
				my $xRouter = $routersHash->{$announcedPrefix};
				
				#"announce" the split prefix to the TLD router 
				$xRouter->addSpecialChild($newXRouter, $announcedPrefix);
				
				#also check to see if the TLD router is virtual
				if ($xRouter->isVirtualRouter() ){
					
					#if it is, let the new router's pop enter also for the child in the topology
					$xRouter->setLocalizationInformation($pop, $as, $ip);
					
				}
			}
			
		#if mask equals tldMask then we have an xRouter already created (it is a TLD router)
		}elsif($mask == $tldMask){
			
			#associate to the TLD router the data of the child
			my $xRouter = $routersHash->{$prefix};
		
			my $pop = $topology->getBorderPOPForPrefix($prefix);
			my $as = $topology->getASForPOP($pop);
			my $ip = $topology->getIPForPOP($pop);
			
			$xRouter->setPOPNumber($pop);
			$xRouter->setAS($as);
			$xRouter->setPOPIP($ip);
			
			#create a new router which connects to the TLD router
			my $newXRouter = CoreSim::X::XRouter->new($ipUtility, $self->{"_DEBUG"});
			$newXRouter->setPrefix($prefix);
			$newXRouter->setLayer(3);
			
			#set the localization information
			$newXRouter->setLocalizationInformation($pop, $as, $ip);
			
			#add the router to the routersHash
			$routersHash->{$prefix} = $newXRouter;
			
			#add the pop level router as a child to the TLD router
			$xRouter->addChild($newXRouter);
			
		
		#if ($mask> $tldMask)the mask is wider than the tldMask then we need to create a new xRouter and associate it to a TLD router
		}elsif($mask>$tldMask){
			
			my $newXRouter = CoreSim::X::XRouter->new($ipUtility, $self->{"_DEBUG"});
			$newXRouter->setPrefix($prefix);
			$newXRouter->setLayer(3);
			
			#find the pop information		
			my $pop = $topology->getBorderPOPForPrefix($prefix);
			my $as = $topology->getASForPOP($pop);
			my $ip = $topology->getIPForPOP($pop);
			
			#set the localization information
			$newXRouter->setLocalizationInformation($pop, $as, $ip);
			
			#add the router to the routersHash
			$routersHash->{$prefix} = $newXRouter;
			
			#now it needs to find the parent /tldMask prefix
			my $parentRouter = &findParentNodeForPrefix($self,$prefix, $tldMask );
			
			$parentRouter->addChild($newXRouter);
			
			#if parent is virtual, the child pop enters in the topology also in his name
			if ($parentRouter->isVirtualRouter()){
				
				$parentRouter->setLocalizationInformation($pop, $as, $ip);

			}
			
		}
	}
}

sub findParentNodeForPrefix(){
	my ($self, $childPrefix, $tldMask) = @_;
	
	my $routersHash = $self->{"routersHash"};
	my $ipUtility = CoreSim::Utilities::IPUtilities->new();

	my $parentPrefix;
	
	my $mask = $ipUtility->getMaskFromPrefix($childPrefix);
	my $prefix = $childPrefix;
	
	#search the hierarchy for a parent prefix
	while ($mask > $tldMask){
		
		#get a possible parent prefix with a mask decreased by 1
		$parentPrefix = $ipUtility->computeParentPrefix($prefix, 1);
		
		$mask -= 1;
		
		$prefix = $parentPrefix;
	
	}

	my $parentRouter = $routersHash->{$parentPrefix};
	
	if (defined($parentRouter)){
		return $parentRouter;
	}else{
		printf("(EE) The child %s has no TLD parent!\n", $childPrefix);
		return undef;
		
	}
	
}

sub getRouterWithPrefix(){
	my $self = shift;
	my ($prefix) = @_;
	
	return $self->{"routersHash"}->{$prefix};
}

sub getRootLayer(){
	my $self=shift;
	
	return $self->{"rootLayer"};
}

sub firstEightBitsZero(){
	
	my $ip = shift;
	
	my @result = split(/\./, $ip);
	if ($result[0] == 0 ){
		return 1;
	}else{
		return 0;
	}
}

#return 1 at the end
1