package CoreSim::TREE::TREEMapResolver;

use CoreSim::TREE::TREEMapResolverCache; 
use strict;

sub new(){
	my $invocant = shift;
	my ($treeTopology, $topology, $treeMetrics, $ipUtility, $outputFilePath, $resolverType, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	my $treeMapResolverCache = CoreSim::TREE::TREEMapResolverCache->new($ipUtility, $outputFilePath, $_DEBUG);
	
	my $rootLTS;
	my $numberOfQueriesHash = {};
	my $numberOfIterations = 0;
	
	my $self={
		"topology" => $topology,
		"treeTopology" => $treeTopology,
		"treeMapResolverCache" => $treeMapResolverCache,
		"rootLTS" => $rootLTS,
		"treeMetrics" => $treeMetrics,
		"numberOfQueriesHash" => $numberOfQueriesHash,
		"outputFilePath" => $outputFilePath,
		"resolverType" => $resolverType,
		"numberOfIterations" => $numberOfIterations,
		"_DEBUG" => $_DEBUG,
	};
	
	bless($self, $class);
#	$self->{"_DEBUG"} = 3;
	return $self;
}

sub searchForIP(){
	my $self=shift;
	my ($itrIP, $searchedIP) = @_;
	
	#this is a hack! storedITRIP may be used to override the itrIP with another value. 
	#this is needed when I need to change the IP of the MR (for UPC) and 
	if ($self->{"storedITRIP"}) {
		$itrIP = $self->{"storedITRIP"};
	}
	
	my $treeMapResolverCache = $self->{"treeMapResolverCache"};
	my $treeMetrics = $self->{"treeMetrics"};
	my $resolverType = $self->{"resolverType"};
	
	#for now I need to check this way (the format of the ITRNode class obliges me too, if I have time i'll change it: 
	#the map resolver will be initialized by the ITR, Chord and ALT don't really need it)
	my $rootLTS;
	if (!defined($self->{"rootLTS"})){
		$rootLTS = &findClosestRootLTS($self,$itrIP);
		$self->{"rootLTS"}= $rootLTS;
	}else{
		$rootLTS = $self->{"rootLTS"};
	} 
	
	my $responsibleLTS;
	my $peers;
	my $intermediaryLTS = $treeMapResolverCache->findIntermediaryLTS($searchedIP);
	
	#find the destination router, and the path to reach it
	if (defined($intermediaryLTS)){
		($responsibleLTS, $peers) = &findResponsibleLTS($self, $searchedIP, $intermediaryLTS);
	}else{
		($responsibleLTS, $peers) = &findResponsibleLTS($self, $searchedIP, $rootLTS);
	}
	
	#check to see if we found a responsible router
	if (defined($responsibleLTS)){
		
		if ($self->{"_DEBUG"}>=2){
			printf("(TMR) The prefix path:");
			for (my $i=0;$i< scalar(@$peers); $i++){
				printf("%s ", $peers->[$i]->getPrefix());
			}
			printf("\n");
		}
		
		#add the peers to the cache (should I add them even if the responsibleLTS is not defined?)
		#this depends on the type of resolver!!!!!
		$treeMapResolverCache->addNodes($peers, $resolverType);
		
		&updateNodeStats($self, $peers );
		
		#compute the latency associated with the path
		my ($pathLatency, $nodeLatencyList) = $treeMetrics->computePathLatency($peers, $itrIP, $resolverType);
		
		#format the output of the function
		my $status="ok";
		my $destPOP = $responsibleLTS->getPOPNumber();
		my $destAS = $responsibleLTS->getAS();
		my $destPrefix = $responsibleLTS->getPrefix();
		my $nbOfPeers = scalar(@$peers);
		
		
		return ($status, $destPOP, $destAS, $destPrefix, $pathLatency, $nbOfPeers, $nodeLatencyList);
	
	#"back-up" results for when the search fails
	}else{
		
		my $status="discard";
		my $destPrefix = "\t";
		my $destPOP = "\t";
		my $destAS =  "\t";
		return ($status, $destPOP, $destAS, $destPrefix);
		
	}
	
	
	
}

#new version of responsibleLTS (the function which can be found in treeMetrics is now changed and moved here)
#starts from a givenRouter and searches recursively(from the point of view of the algorithm) 
#for the searchedIP
sub findResponsibleLTS(){
	my ($self, $searchedIP, $startLTS) = @_;
	
	my @peers;
	push(@peers, $startLTS);
	my $endLoop = 0;
	
	my $currentLTS = $startLTS;
	do{
		my $nextHopLTS = $currentLTS->findNextHopForIP($searchedIP);
		
		if (defined($nextHopLTS)){
			
			#check to see if we have reached the last LTS [ it returns itself as the owner]
			if ($nextHopLTS->getPrefix() ne $currentLTS->getPrefix()){
			
				push(@peers, $nextHopLTS);
				$currentLTS = $nextHopLTS;
				
			}else{
				$endLoop = 1;
			}
			
		}else{
			
			#end the search and return undef for currentLTS and peers
			return (undef, undef);
		}

	}while(!$endLoop);
	
	#need to check if the $currentLTS is virtualRouter (we can't really solve this query)
	if( $currentLTS->isVirtualRouter() ){
		$currentLTS = undef;
	}
	
	return ($currentLTS, \@peers);
}

sub updateNodeStats(){
	
	my ($self, $peers) = @_;
	
	$self->{"numberOfIterations"} += 1;
	
	
	my $numberOfQueriesHash = $self->{"numberOfQueriesHash"};
	my $nbOfPeers = scalar(@$peers);
	
	for (my $i=0;$i<$nbOfPeers; $i++){

		$numberOfQueriesHash->{$peers->[$i]->getPrefix()} +=1;
	}
	
	if ($self->{"numberOfIterations"}%100000 == 0 ){
		&dropNodeStatsToFile($self);
	}
}


#finds the closest root LTS to the MR
sub findClosestRootLTS(){
	my ($self, $mrIP) = @_;
	
	my $treeTopology = $self->{"treeTopology"};
	my $topology = $self->{"topology"};
	my $rootLayer = $treeTopology->getRootLayer();
	
	my $minNotSet = 1;
	my $minLatency=10000;
	my $chosenRootLTS;
	
	foreach my $rootLTS (@$rootLayer){
		my $latency = $topology->getXMLRPCQuery($rootLTS->getPOPIP(), $mrIP);
		
		if ( ($minNotSet) && ($latency>0) ){
			$minLatency = $latency;
			$chosenRootLTS = $rootLTS;
			
		}elsif(( $latency<$minLatency)&&($latency>0) ){
			$minLatency = $latency;
			$chosenRootLTS = $rootLTS;
		}
		
		printf("(TMR) Testing with rootLTS %s. The latency is %s\n", $rootLTS->getPOPIP(), $latency) if ($self->{"_DEBUG"});
	}

	if (!defined($chosenRootLTS)){
		printf("(EE)(TR) No root server set! Program will exit!\n");
		die;
	}else{
		printf("Have chosen rootLTS with ip %s to which the MR has the latency %s\n", $chosenRootLTS->getPOPIP(), $minLatency) if ($self->{"_DEBUG"});
		return $chosenRootLTS;
	}
}

sub dropNodeStatsToFile{
	my $self=shift;
	
	#write treeRoutersStats
	my $outputFilePath = $self->{"outputFilePath"};
	my $outputFile = "$outputFilePath/treeRoutersStats.txt";
	
	my $numberOfQueriesHash = $self->{"numberOfQueriesHash"};
	my $treeTopology = $self->{"treeTopology"};
	my $layer1LTSs = $treeTopology->getLayer1LTSs();
	my @prefixes = keys %$numberOfQueriesHash;
	
	
	open(OFILE, "> $outputFile") or die printf("Error encountered while opening the output file for the TREE router stats %s:%s",
			$outputFile, $!);
	
	printf(OFILE "TREE Prefix\tX Layer\tPOP\tAS\tQueries Routed\n");
	
	foreach my $prefix (@prefixes){

		my $router;
		if ($prefix eq "0.0.0.0/0"){
			$router = $self->{"rootLTS"};
		}elsif(defined($layer1LTSs->{$prefix})){
			$router = $layer1LTSs->{$prefix};
		}else{
			$router = $treeTopology->getLTSWithPrefix($prefix);
		}
		
		my $pop = $router->getPOPNumber();
		my $as =  $router->getAS();
		my $level = $router->getLayer();
		my $queriesCount = $numberOfQueriesHash->{$prefix};
		
		printf(OFILE "%s\t%s\t%s\t%s\t%s\n", $prefix, $level, $pop, $as, $queriesCount);

		
	}	
	
	close(OFILE);
		
}

sub DESTROY(){
	my $self=shift;
	
	&dropNodeStatsToFile($self);
	
}

#return 1
1