package CoreSim::TREE::TREETopology;

use strict;
use CoreSim::TREE::LTS; 

sub new(){
	
	my $invocant = shift;
	
	my ($topology, $ipUtility,  $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	#this hash will hold all the routers except the root layer and have as keys the prefixes associated to them 
	my $ltsHash = {};

	
	#the mask for the L1 routers (10 gives 1024 TLD routers)
	my $l1Mask = 8;
		
	
	my $self={
		"topology" => $topology,
		"ltsHash" => $ltsHash,
		"ipUtility" => $ipUtility, 
		"l1Mask" => $l1Mask,
		"_DEBUG" => $_DEBUG,
	};
	
	bless($self, $class);
	
#	$self->{"_DEBUG"} = 3;
	&createTopology($self);
	
	return $self;
}

sub createTopology(){
	
	my $self=shift;
	
	#generates the root server(s)
	&createRootLayer($self);

	#create the regional EID registries. They are all the /8 prefixes except E and D classes
	#and reserved private address space
	&createLayer1($self);
	
	#associate POPs to the nodes forming the topology except for the root layer
	&parsePrefixList($self, $self->{"l1Mask"});
	
#	&outputNodesAndChildren($self);
	
}

sub createRootLayer_obsolete(){
	
	my $self = shift;
	my $ipUtility = $self->{"ipUtility"};
	my $topology = $self->{"topology"};
	my $rootServers = $topology->getRootServers();
	my $_DEBUG = $self->{"_DEBUG"};
	
	#this list contains all the routers at the root layer 
	my $rootLayer = [];
	
	printf("Creating root LTSs...\n");
	#add to root layer
	for (my $i=0; $i < scalar(@$rootServers); $i++){
		my $newLTS = CoreSim::TREE::LTS->new( $ipUtility, $_DEBUG); 
		$newLTS->setPrefix("0.0.0.0/0");
		$newLTS->setLayer(0);
		
		$newLTS->setPOPIP($rootServers->[$i]);
		$rootLayer->[$i] = $newLTS;
	}	
	
	
	$self->{"numberOfRootLTSs"} = scalar (@$rootServers);
	$self->{"rootLayer"} = $rootLayer;
}

#there is just one root pop
sub createRootLayer(){
	my $self = shift;
	my $ipUtility = $self->{"ipUtility"};
	my $topology = $self->{"topology"};
	my $_DEBUG = $self->{"_DEBUG"};
	
	#I use a larger number just to be sure there is a list from where the map resolver can choose
	my $numberOfRootLTSs = 20;
	my $popList = $topology->getBestConnectedNPOPs( $numberOfRootLTSs);
	
	#this list contains all the routers at the root layer 
	my $rootLayer = [];
	
	printf("Creating root LTSs...\n");
	
	for (my $i=0; $i<$numberOfRootLTSs; $i++){
		my $newLTS = CoreSim::TREE::LTS->new( $ipUtility, $_DEBUG); 
		$newLTS->setPrefix("0.0.0.0/0");
		$newLTS->setLayer(0);
		
		$newLTS->setLocalizationInformation($popList->[$i]->{"popNumber"}, $popList->[$i]->{"asNumber"},
											$popList->[$i]->{"ip"});
		$rootLayer->[$i] = $newLTS;
	}
	
	$self->{"numberOfRootLTSs"} = $numberOfRootLTSs;
	$self->{"rootLayer"} = $rootLayer;
}

sub createLayer1(){
	
	my $self = shift;
	
	my $ltsHash = $self->{"ltsHash"};
	my $ipUtility = $self->{"ipUtility"};
	my $numberOfRootLTSs= $self->{"numberOfRootLTSs"};
	my $rootLayer = $self->{"rootLayer"};
	my $layer1LTSs = {};
	
	#controls the layer1 prefixes
	my $l1Mask = 8;
	
	printf("Creating the layer 1 LTSs ...\n");
	
	for (my $i = 1; $i < 224; $i++){
#		if ( ($i != 10) && ($i != 127) && ($i != 192) ){
			 
			# the prefix is obtained from i << 24
			my $newIPNumber = $i<<(32 - $l1Mask);
			my $newIP = $ipUtility->numberToIP($newIPNumber);
			my $newPrefix = "$newIP/8";
			
			#create the LTS
			my $newLTS = CoreSim::TREE::LTS->new( $ipUtility, $self->{"_DEBUG"});
			$newLTS->setLayer(1);
			$newLTS->setPrefix($newPrefix);
			
			#add to ltsHash and layer1LTSs
			$ltsHash->{$newPrefix} = $newLTS;
			$layer1LTSs->{$newPrefix} = $newLTS;
			
			#add FIB entries to all the root routers
			for (my $j=0; $j<$numberOfRootLTSs; $j++){
				my $rootRouter = $rootLayer->[$j];
				$rootRouter->addChild($newLTS);
			}
			
			printf("(TT) Added the L1 node %s to the topology\n", $newPrefix) if ($self->{"_DEBUG"}>=2);
			
#		}
	}
	
	$self->{"layer1LTSs"} = $layer1LTSs;
	
}

sub parsePrefixList(){
	
	my ($self, $l1Mask) = @_;
	
	my $topology = $self->{"topology"};
	my $ltsHash = $self->{"ltsHash"};
	my $ipUtility = $self->{"ipUtility"};
	my $layer1LTSs = $self->{"layer1LTSs"};
	my $layer2LTSs = {};
	
	printf("Parsing the prefix list ...\n");
	
	#get the prefix list from the topology module
	my $prefixList = $self->{"topology"}->getFilteredPrefixList();
	
	#add nodes from /l1Mask to /32 to the TREE
	for (my $maskIterator = $l1Mask; $maskIterator<=32; $maskIterator++){
		
		#go through all the prefixes and ...
		foreach my $prefix (@$prefixList){
			
			# select those with the mask equal with the maskIterator
			if ($ipUtility->getMaskFromPrefix($prefix) == $maskIterator) {
				
				#special case for prefixes with a l1Mask. the MS will be at L1
				if ($maskIterator == $l1Mask){
					
					#find the lts 
					my $lts = $ltsHash->{$prefix};
				
					#determine from the topology module the localization information of the L1 LTS 
					my $pop = $topology->getBorderPOPForPrefix($prefix);
					my $as = $topology->getASForPOP($pop);
					my $ip = $topology->getIPForPOP($pop);
					
					$lts->setLocalizationInformation($pop, $as, $ip);
					
					#set the LTS as MS. This is because the /8's that exist in the prefixList don't have children 
					$lts->setMS();
				
				#create new leaf LTS (MS)
				}else{
	
					#create the LTS
					my $newLTS = CoreSim::TREE::LTS->new( $ipUtility, $self->{"_DEBUG"});
					$newLTS->setPrefix($prefix);
					
					#find parent router 
					my $parentLTS = &findParentLTSForPrefix($self, $prefix);
					$parentLTS->addChild($newLTS);
					$newLTS->setLayer( $parentLTS->getLayer() + 1 );
					
					#find the pop specific information		
					my $pop = $topology->getBorderPOPForPrefix($prefix);
					my $as = $topology->getASForPOP($pop);
					my $ip = $topology->getIPForPOP($pop);
					
					#set the localization information
					$newLTS->setLocalizationInformation($pop, $as, $ip);
					
					#set the LTS as MS
					$newLTS->setMS();
					
					#add to ltsHash and layer2LTSs
					$ltsHash->{$prefix} = $newLTS;
					$layer2LTSs->{$prefix} = $newLTS;
					
					#if parent is virtual, the child node's pop enters in the topology as the pop responsible for the parent 
					if ($parentLTS->isVirtualRouter()){
						
						$parentLTS->setLocalizationInformation($pop, $as, $ip);
		
					}
					
					printf("(TT) Added the L2 LTS %s to the topology with parent %s\n", $prefix, $parentLTS->getPrefix()) if ($self->{"_DEBUG"}>=3);
				}
				

			}
			
		}
		
	}
	
	$self->{"layer2LTSs"} = $layer2LTSs;
}

sub findParentLTSForPrefix(){
	my ($self, $childPrefix) = @_;
	
	my $ltsHash = $self->{"ltsHash"};
	my $ipUtility = $self->{"ipUtility"};
	
	my $mask = $ipUtility->getMaskFromPrefix($childPrefix);
	my $prefix = $childPrefix;
	
	my $parentPrefix;
	my $possibleParentPrefix;
	
	my $endLoop=0;
	
	#search the hierarchy for a parent prefix
	while (($mask > 8) && (!$endLoop)){
		
		#get a possible parent prefix with a mask decreased by 1
		$possibleParentPrefix = $ipUtility->computeParentPrefix($prefix, 1);
		
		#if the parent does exist stop searching
		if (defined($ltsHash->{$possibleParentPrefix})){
			$endLoop = 1;
		}
		
		$mask -= 1;
		$prefix = $possibleParentPrefix;
	}
	
	printf("(TT) prefix = %s possible parent prefix = %s endLoop = %s\n",$childPrefix, $possibleParentPrefix, $endLoop ) if ($self->{"_DEBUG"}>3);
	
	#if the parent is not found then post an error message. All the prefixes should have a parent in the L1 level
	if ($endLoop==1){
		
		return $ltsHash->{$possibleParentPrefix};
		
	}else{
		
		printf("(EE) There is no parent for the prefix %s!\n", $childPrefix);
		die;
	}
	
	
	
}

sub getRootLayer(){
	my $self=shift;
	return $self->{"rootLayer"};	
}

sub getLayer1LTSs(){
	my $self=shift;
	return $self->{"layer1LTSs"};
}

sub getLTSWithPrefix(){
	my $self = shift;
	my ($prefix) = @_;
	
	return $self->{"ltsHash"}->{$prefix};
}

sub outputNodesAndChildren(){
	my $self=shift;
	
	my $outputFile = "/home/florin/workspace/eclipse/simdata/output/nodesAndChildren.txt";
	open(OFILE,">$outputFile");
	
	my $rootLayer = $self->{"rootLayer"};
	
	printf(OFILE "The root layer\n");
	for(my $i=0; $i<scalar(@$rootLayer); $i++){
		my $lts =$rootLayer->[$i];
		printf(OFILE "%s \n", $lts->getPrefix());
		
		my $fib = $lts->getFIB();
		foreach my $entry (keys %$fib){
			printf(OFILE "%s ", $entry);
		}
		printf(OFILE "\n");
	}
	
	my $l1Layer = $self->{"layer1LTSs"};
	printf(OFILE "\nThe l1 layer\n");
	
	foreach my $lts (keys %$l1Layer){
		printf(OFILE "node %s \n", $l1Layer->{$lts}->getPrefix);
		
		my $fib = $l1Layer->{$lts}->getFIB();
		foreach my $entry (keys %$fib){
			printf(OFILE "%s ", $entry);
		}
		printf(OFILE "\n");
	}
	
	my $l2Layer = $self->{"layer2LTSs"};
	printf(OFILE "\nThe l2 layer\n");
	
	foreach my $lts (keys %$l2Layer){
		printf(OFILE "node %s \n", $l2Layer->{$lts}->getPrefix());
		
		my $fib = $l2Layer->{$lts}->getFIB();
		foreach my $entry (keys %$fib){
			printf(OFILE "%s ", $entry);
		}
		printf(OFILE "\n");
	}
	
	close(OFILE);
		
}

return 1