package CoreSim::TREE::TREEMetrics;

use strict;

sub new(){
	my $invocant = shift;
	my ($topology, $treeTopology,$ipUtility, $_DEBUG) = @_;
	my $class = ref($invocant) || $invocant;
	
	my $numberOfQueriesHash={};
	
	my $self={
		"topology" => $topology, 
		"treeTopology" => $treeTopology,
		"ipUtility" => $ipUtility,
		"numberOfQueriesHash" => $numberOfQueriesHash,
		"_DEBUG" => $_DEBUG,
	};
#	$self->{"_DEBUG"}=2;
	bless($self, $class);
	
	return $self;
}

sub getLookupDelayForIP(){
	my $self=shift;
	my($itrIP, $searchedIP) = @_;
	
	printf("(gLDFI) Called with itrIP = %s and searchedIP = %s\n", $itrIP, $searchedIP) if ($self->{"_DEBUG"});
	my $treeTopology = $self->{"treeTopology"};
	
	my ($rootRouter,$rootRouterIP, $latencyToRoot) = &findClosestRootRouter($self, $itrIP);
	printf("The closest Root Router has the ip %s\n", $rootRouterIP) if ($self->{"_DEBUG"});
	
	
	printf("Finding responsible router for %s starting with the rootRouter with IP %s\n", $searchedIP, $rootRouterIP) 
			if($self->{"_DEBUG"} >=2);
	my ($responsibleRouter, $peers) = &findResponsibleRouter($self, $searchedIP, $rootRouter);
	
	if (defined($responsibleRouter)){
		
		my $numberOfQueriesHash = $self->{"numberOfQueriesHash"};
		&updateNodeStats($numberOfQueriesHash, $peers );
		
		my ($pathLatency, $nodeLatencyList) = &computePathLatency($self,$itrIP,$peers, $rootRouterIP,$latencyToRoot);
		
		#format the output of the function
		my $status="ok";
		my $destPOP = $responsibleRouter->getPOPNumber();
		my $destAS = $responsibleRouter->getAS();
		my $destPrefix = $responsibleRouter->getPrefix();
		my $nbOfPeers = scalar(@$peers);
		
		return ($status, $destPOP, $destAS, $destPrefix, $pathLatency, $nbOfPeers, $nodeLatencyList);
		
	}else{
		
		my $status="discard";
		my $destPrefix = "\t";
		my $destPOP = "\t";
		my $destAS =  "\t";
		return ($status, $destPOP, $destAS, $destPrefix);
		
	}
	
	
}

sub findResponsibleRouter(){
	my ($self, $searchedIP, $rootRouter) = @_;
			
	my @peers;
	push(@peers, $rootRouter);
	
	#ask root layer
	my $rerRouter = $rootRouter->findNextHopForIP($searchedIP);
	push(@peers, $rerRouter);
	
	#ask TLD supplied by the root
	my $responsibleRouter = $rerRouter->findNextHopForIP($searchedIP);
	
	#we need to test that the responsible router exists
	if (defined($responsibleRouter)){

		push(@peers, $responsibleRouter);
	}
	
	return ($responsibleRouter, \@peers);
	
}

sub computePathLatency1(){
	my $self=shift;
	my ($peers, $itrIP) = @_;
	
	my $topology = $self->{"topology"};
	my $pathLength = scalar(@$peers);
	
	my @nodeLatencyList;
	my $latency = 0;
	
	push(@nodeLatencyList, $itrIP);
	
	for (my $i=0;$i<$pathLength; $i++){
		
		my $node = $peers->[$i];
		my $queryLatency = $topology->getLatencyBetweenPOPIPs1($itrIP, $node->getPOPIP());
		$latency += 2* $queryLatency;
		
		printf("(XM) Computed latency between ITR %s and %s with the IP %s is %s\n",$itrIP, $node->getPrefix(), $node->getPOPIP(),
				$queryLatency )if ($self->{"_DEBUG"}>=2);
		
		push(@nodeLatencyList, $queryLatency );
		push(@nodeLatencyList, $node->getPOPIP());
	}
	
	return ($latency, \@nodeLatencyList);
}

#"simulates anycast" by taking the minimum latency between the two routers
sub anycastLatency(){
	my ($self, $ip, $router) = @_;
	
	my $topology = $self->{"topology"};
	
	my $chosenIP;

	my $ipList = $router->getRootIPList();
	
	my $minLatency = $topology->getLatencyBetweenPOPIPs1($ip, $ipList->[0]);
	$chosenIP = $ipList->[0] ;
	
	foreach my $anycastIP (@$ipList){
		my $latencyQuery = $topology->getLatencyBetweenPOPIPs1($ip, $anycastIP);
		
		if ($latencyQuery< $minLatency){
			$minLatency = $latencyQuery;
			$chosenIP = $anycastIP;

		}
		
	}

	return ($minLatency, $chosenIP);
}

#returns an xRouter coresponding to a root router
sub findClosestRootLTS(){
	my $self=shift;
	my ($itrIP) = @_;
	
	printf("Finding closest root router for ip %s\n", $itrIP) if ($self->{"_DEBUG"}>=2);
	
	my $rootLayer = $self->{"treeTopology"}->getRootLayer();
	my $nbOfRootLTSs = scalar(@$rootLayer);
	
	my ($minLatency, $rootLTSIP) = &anycastLatency($self, $itrIP, $rootLayer->[0]);
	my $closestRootLTS = $rootLayer->[0];
	my $closestRootLTSIP = $rootLTSIP;
	
	for(my $i=0; $i<$nbOfRootLTSs; $i++){
		my $rootLTS = $rootLayer->[$i];
		my ($latency, $chosenIP) = &anycastLatency($self, $itrIP, $rootLTS);
		
		printf("Latency between rootLTS %s and the ITR %s is %s and the chosen IP is %s\n", $i, $itrIP, $latency, $chosenIP)
				if ($self->{"_DEBUG"}>=2);
				
		if ($latency < $minLatency){
			$minLatency = $latency;
			$closestRootLTS = $rootLTS;
			$closestRootLTSIP = $chosenIP;
		}
	}
	
	my $rootLTS = CoreSim::TREE::LTS->new($self->{"ipUtility"}, $self->{"_DEBUG"});
	
	#determine pop and as
	my $pop = $closestRootLTS->getPOPForIP($closestRootLTSIP);
	my $as = $closestRootLTS->getASForIP($closestRootLTSIP);
	
	$rootLTS->setLocalizationInformation($pop, $as, $closestRootLTSIP);
	$rootLTS->setFIB($closestRootLTS->getFIB());
	$rootLTS->setRootLTS();
	$rootLTS->setLayer(1);
	$rootLTS->setPrefix("0.0.0.0/0");
	
	printf("(xM) The closest rootLTS has the ip %s\n", $closestRootLTSIP) if ($self->{"_DEBUG"}>=2);
	return $rootLTS;
}

sub updateNodeStats(){
	my ($numberOfQueriesHash, $peers) = @_;
	
	#for the root router use the rootName instead of the prefix
	$numberOfQueriesHash->{$peers->[0]->getRootName()} +=1;
	
	my $nbOfPeers = scalar(@$peers);
	
	for (my $i=1;$i<$nbOfPeers; $i++){

		$numberOfQueriesHash->{$peers->[$i]->getPrefix()} +=1;
	}
}



#return 1 at the end
1