package CoreSim::TREE::TREETopology;

use strict;
use CoreSim::TREE::LTS; 
#use CoreSim::Utilities::IPUtilities;

sub new(){
	
	my $invocant = shift;
	
	my ($topology, $ipUtility, $_DEBUG) = @_;
	
	my $class = ref($invocant) || $invocant;
	
	#this hash will hold all the routers except the root layer and have as keys the prefixes associated to them 
	my $routersHash = {};
	
	#this list contains all the routers at the root layer 
	my $rootLayer = [];
	
	#the mask for the TLD routers (10 gives 1024 TLD routers)
	my $rerMask = 8;
	
	my $numberOfRootRouters = 8;
	
	#number of IPs per root router
	my $numberOfIPsPerRootRouter = 2;
	
	my $self={
		"topology" => $topology,
		"routersHash" => $routersHash,
		"rootLayer" => $rootLayer,
		"ipUtility" => $ipUtility, 
		"rerMask" => $rerMask,
		"numberOfRootRouters" => $numberOfRootRouters,
		"numberOfIPsPerRootRouter" => $numberOfIPsPerRootRouter,
		"_DEBUG" => $_DEBUG,
	};
	
	bless($self, $class);
	
	$self->{"_DEBUG"}=0;
	&createTopology($self);
	
	return $self;
}

sub createTopology(){
	
	my $self=shift;
	
	#generate the 8 root servers
	&createRootLayer($self);

	#create the regional EID registries
	&createRER($self);
	
	&associateAllPOPs($self);
	
}

sub createRootLayer(){
	my $self=shift;
	
	my $rootLayer = $self->{"rootLayer"};
	my $_DEBUG = $self->{"_DEBUG"};
	my $ipUtility = $self->{"ipUtility"};
	my $routersHash = $self->{"routersHash"};
	
	my $numberOfRootRouters = $self->{"numberOfRootRouters"};
	
	printf("Creating the root layer..\n");
	#all the root nodes are the same they are used for redundancy. This creates a list of them
	for (my $i= 0; $i< $numberOfRootRouters; $i++ ){
		my $newLTS = CoreSim::TREE::LTS->new( $ipUtility, $_DEBUG);
		$newLTS->setPrefix("0.0.0.0/0");
		$newLTS->setRootName("rootLTS$i");
		$newLTS->setLayer(1);
		
		#it is not a prefix but I need a way to locate the router
		$routersHash->{"rootRouter$i"} = $newLTS;
		
		push (@$rootLayer, $newLTS);
		
	}

}

sub createRER(){
	my $self=shift;
	
	my $routersHash = $self->{"routersHash"};
	my $ipUtility = $self->{"ipUtility"};
	my $rerMask = $self->{"rerMask"};
	my $numberOfRootRouters= $self->{"numberOfRootRouters"};
	my $rootLayer = $self->{"rootLayer"};
	my $prefixToRIR = $self->{"topology"}->getPrefixToRIR();
	my $rerLayer = {};
	
	printf("size = %s\n", scalar(keys %$prefixToRIR));
	printf("(XT) Creating the RER layer ..\n");
	
	foreach my $firstOctet (keys %$prefixToRIR){
		
		my $rir = $prefixToRIR->{$firstOctet};
		if (!defined($rerLayer->{ $rir })){
			
			#obtain the prefix from the number 1 << 24 -> 1/8
			my $prefixNumber = $firstOctet << 24;
			my $ip = $ipUtility->numberToIP($prefixNumber);
			my $prefix = "$ip/8";
			
			my $newLTS = CoreSim::TREE::LTS->new( $ipUtility, $self->{"_DEBUG"});
			
			#set the RER name (it will take the name of the RIR)
			$newLTS->setRERName($prefixToRIR->{$firstOctet});
			$newLTS->setLayer(2);
			$newLTS->addRERPrefix($prefixToRIR->{$firstOctet});
			
			#add to routersHash and rerLayer
			$routersHash->{$prefix} = $newLTS;
			$rerLayer->{$prefixToRIR->{$firstOctet}} = $newLTS;
			
			
			#add FIB entries for all the Root routers
			for (my $i=0;$i<$numberOfRootRouters;$i++){
				my $rootLayerRouter = $rootLayer->[$i];
				
				$rootLayerRouter->addFIBEntry($prefix, $newLTS);
			}
			
			
			printf("(XT) Added the RER node %s to the topology and prefix %s\n", $prefixToRIR->{$firstOctet}, $prefix) 
					if ($self->{"_DEBUG"}>=2);
			
		}else{
			my $rerRouter = $rerLayer->{$rir};
			
			my $prefixNumber = $firstOctet << 24;
			my $ip = $ipUtility->numberToIP($prefixNumber);
			my $prefix = "$ip/8";
			
			$rerRouter->addRERPrefix($prefixToRIR->{$firstOctet});
			$routersHash->{$prefix} = $rerRouter;
			
			#add FIB entries for all the Root routers
			for (my $i=0;$i<$numberOfRootRouters;$i++){
				my $rootLayerRouter = $rootLayer->[$i];
				
				$rootLayerRouter->addFIBEntry($prefix, $rerRouter);
			}
			
			printf("(XT) Added prefix %s to RER %s\n", $prefix, $rerRouter->getName()) if ($self->{"_DEBUG"} >= 2);
		}
	}
	
	
	$self->{"rerLayer"} = $rerLayer;
	#create all the other virtual pops
	
		
}

sub associateAllPOPs(){
	my $self=shift;
	
	#this will give IPs to the whole list of root routers
	&associatePOPsToRootRouters($self);
	
	#IPs to RERs
	&associatePOPsToRERs($self);
	
	#0 for 3 layer; 1 for 5 layer
	my $typeOfTopology = 0;
	if ($typeOfTopology == 1 ){
		#this will associated IPs TLD nodes , create the POP Level and associate child POPs to parent virtual routers
		&associatePOPs($self);
	}else{
		&associatePOPsFlat($self);
	}
}

#this should be done by geographical criterion, for now it is done just by taking the best connected POPs and asociating them to 
#the routers 
sub associatePOPsToRootRouters(){
	my $self=shift;
	
	my $topology = $self->{"topology"};
	my $rootLayer = $self->{"rootLayer"};
	my $numberOfRootRouters = $self->{"numberOfRootRouters"};
	my $numberOfIPsPerRootRouter = $self->{"numberOfIPsPerRootRouter"};
	my $numberOfPOPsRequired = $numberOfRootRouters * $numberOfIPsPerRootRouter;
	
	my $popList = $topology->getBestConnectedNPOPs($numberOfPOPsRequired);
	
	for (my $i=0; $i<$numberOfPOPsRequired;$i+=$numberOfIPsPerRootRouter){
		for (my $j=0; $j<$numberOfIPsPerRootRouter; $j++){
			
			my $routerNumber = $i/$numberOfIPsPerRootRouter;
			my $rootRouter = $rootLayer->[$routerNumber];
			
			my $routerInfo = $popList->[$i+$j];
			$rootRouter->addRootRouterInfo($routerInfo);
			
		}
	}
	
}

sub associatePOPsToRERs(){
	my $self=shift;
	
	my $topology = $self->{"topology"};
	my $numberOfRootRouters = $self->{"numberOfRootRouters"};
	my $numberOfIPsPerRootRouter = $self->{"numberOfIPsPerRootRouter"};
	my $rerLayer = $self->{"rerLayer"};
	
	#add 5 as a constant because there are 5 RER/RIRs
	my $numberOfRootPOPsRequired = $numberOfRootRouters * $numberOfIPsPerRootRouter;
	my $numberOfPOPsRequired = $numberOfRootPOPsRequired + 5;
	my $bestConnectedPOPs = $topology->getBestConnectedNPOPs($numberOfPOPsRequired);
	
	my $nbOfAssociatedRouters =$numberOfRootPOPsRequired ;
	foreach my $rer (keys %$rerLayer){
		my $rerRouter = $rerLayer->{$rer};
		
		my $pop= $bestConnectedPOPs->[$nbOfAssociatedRouters]->{"popNumber"};
		my $as = $bestConnectedPOPs->[$nbOfAssociatedRouters]->{"asNumber"};
		my $ip = $bestConnectedPOPs->[$nbOfAssociatedRouters]->{"ip"};
		
		$rerRouter->setLocalizationInformation($pop, $as, $ip);
		$nbOfAssociatedRouters++;
	}

	
}

sub associatePOPsFlat(){
	my $self=shift;
	
	my $topology = $self->{"topology"};
	my $prefixList = $topology->getFilteredPrefixList();
	my $ipUtility = $self->{"ipUtility"};
	my $rerMask = $self->{"rerMask"};
	my $routersHash = $self->{"routersHash"};
	my $numberOfRootRouters= $self->{"numberOfRootRouters"};
	my $rootLayer = $self->{"rootLayer"};
	
	printf("(XT) Associating POPs to LTSs..\n");
	
	my $count  = 0 ;
	
	#create virtual routers for all unexisting /8's; desn't go above 224(multicast)
	for (my $i=1;$i<224;$i++){

		my $number = $i<<24;
		my $ip = $ipUtility->numberToIP($number);
		my $prefix = "$ip/8";
		
		if (!defined($routersHash->{$prefix})){
			my $newLTS = CoreSim::TREE::LTS->new($ipUtility, $self->{"_DEBUG"});
			$newLTS->setPrefix($prefix);
			
			#add FIB entries for all the Root routers
			for (my $i=0;$i<$numberOfRootRouters;$i++){
				my $rootLayerRouter = $rootLayer->[$i];
				
				$rootLayerRouter->addFIBEntry($prefix, $newLTS);
			}
			
			$newLTS->setLayer(2);
			
			#add the router to the routersHash
			$routersHash->{$prefix} = $newLTS;
		}
		
	}
	
	
	
	#add the companies which have /8
	for my $prefix(@$prefixList){
		
		my $mask = $ipUtility->getMaskFromPrefix($prefix);
		if ($mask==8){
		
			my $newLTS = $routersHash->{$prefix};

			#find the pop specific information		
			my $pop = $topology->getBorderPOPForPrefix($prefix);
			my $as = $topology->getASForPOP($pop);
			my $ip = $topology->getIPForPOP($pop);
			
			#set the localization information
			$newLTS->setLocalizationInformation($pop, $as, $ip);
			
			#add the router to the routersHash
			$routersHash->{$prefix} = $newLTS;
			
		}
	}
	
	
	#add the prefixes with mask higher than 8
	for my $prefix (@$prefixList){
		
		$count ++;
		
		if ($count % 10000 == 0) {
			printf("(XT) Done with %s prefixes\n", $count);
		}
		
		if (!defined($routersHash->{$prefix}) && ($prefix ne "0.0.0.0/0")){
			
			my $newLTS = CoreSim::TREE::LTS->new($ipUtility, $self->{"_DEBUG"});
			$newLTS->setPrefix($prefix);
			
			#now it needs to find the parent prefix
			my $parentRouter = &findParentFlat($self,$prefix, $rerMask);
			$parentRouter->addChild($newLTS);
			$newLTS->setLayer(3);
			
			#find the pop specific information		
			my $pop = $topology->getBorderPOPForPrefix($prefix);
			my $as = $topology->getASForPOP($pop);
			my $ip = $topology->getIPForPOP($pop);
			
			#set the localization information
			$newLTS->setLocalizationInformation($pop, $as, $ip);
			
			#add the router to the routersHash
			$routersHash->{$prefix} = $newLTS;
			
			#if parent is virtual, the child node's pop enters in the topology as the pop responsible for the parent 
			if ($parentRouter->isVirtualRouter()){
				
				$parentRouter->setLocalizationInformation($pop, $as, $ip);
				$parentRouter->setEndorsed();
			}
			
		#this happends for companies
		}else{
			
			#not sure what to do here!
		}
	
	}
	
}

sub associatePOPs(){
	my $self=shift;
	
	my $topology = $self->{"topology"};
	my $prefixList = $topology->getPrefixList();
	my $ipUtility = $self->{"ipUtility"};
	my $rerMask = $self->{"rerMask"};
	my $routersHash = $self->{"routersHash"};
	
	printf("(XT) Associating POPs to LTSs..\n");
	
	my $count  = 0 ;
	
	for ( my $maskIterator = $rerMask; $maskIterator<=32; $maskIterator++){
		
		foreach my $prefix (@$prefixList){
			
			my $mask = $ipUtility->getMaskFromPrefix($prefix);
			
			#build the topology from the smallest to the highest of masks (/8->/32)
			if ( $mask == $maskIterator ){
			
		#		printf("Working with prefix %s\n", $prefix);
				
				$count ++;
				
				if ($count % 10000 == 0) {
					printf("(XT) Done with %s prefixes\n", $count);
				}
				
					
				#if mask equals rerMask then we have an LTS already created (it is a RER LISP server)
				if($mask == $rerMask){
					
					#associate to the RER router the data of the child
					my $lts = $routersHash->{$prefix};
				
					#determine from the topology the localization information of the RER router 
					#this router is "endoresed" by the already existing border router of the AS with the RER prefix
					my $pop = $topology->getBorderPOPForPrefix($prefix);
					my $as = $topology->getASForPOP($pop);
					my $ip = $topology->getIPForPOP($pop);
					
					$lts->setLocalizationInformation($pop, $as, $ip);
					
					#create a new router which connects to the RER router. Practically this is a router supported by the same POP
					#as RER router but a distinction should be made here: I associated the same POP to the RER router because otherwise
					#no information is known about who will administer the routers/ their POP specific data
					my $newLTS = CoreSim::TREE::LTS->new($ipUtility, $self->{"_DEBUG"});
					$newLTS->setPrefix($prefix);
					$newLTS->setLayer(3);
					
					#set the localization information
					$newLTS->setLocalizationInformation($pop, $as, $ip);
					
					#add the router to the routersHash
					$routersHash->{$prefix} = $newLTS;
					
					#add the pop level router as a child to the TLD router
					$lts->addChild($newLTS);
					
				
				#if ($mask> $rerMask)the mask is wider than the rerMask then we need to create a new lts and associate to whom ever 
				#is responsible for the parent prefix
				}elsif($mask>$rerMask){
					
					my $newLTS = CoreSim::TREE::LTS->new($ipUtility, $self->{"_DEBUG"});
					$newLTS->setPrefix($prefix);
					
					#now it needs to find the parent prefix
					my $parentRouter = &findParentNodeForPrefix($self,$prefix, $rerMask);
					$parentRouter->addChild($newLTS);
					$newLTS->setLayer($parentRouter->getLayer() + 1);
					
					#find the pop specific information		
					my $pop = $topology->getBorderPOPForPrefix($prefix);
					my $as = $topology->getASForPOP($pop);
					my $ip = $topology->getIPForPOP($pop);
					
					#set the localization information
					$newLTS->setLocalizationInformation($pop, $as, $ip);
					
					#add the router to the routersHash
					$routersHash->{$prefix} = $newLTS;
					
					#if parent is virtual, the child node's pop enters in the topology as the pop responsible for the parent 
					if ($parentRouter->isVirtualRouter()){
						
						$parentRouter->setLocalizationInformation($pop, $as, $ip);
		
					}
					
				}
				
				
			}
			
		}
		
	}
}

sub findParentNodeForPrefix(){
	my ($self, $childPrefix, $rerMask) = @_;
	
	my $routersHash = $self->{"routersHash"};
	my $ipUtility = CoreSim::Utilities::IPUtilities->new();

	my $parentPrefix;
	
	my $mask = $ipUtility->getMaskFromPrefix($childPrefix);
	my $prefix = $childPrefix;
	my $endLoop=0;
	my $possibleParentPrefix;
	
	#search the hierarchy for a parent prefix
	while (($mask > 8) && (!$endLoop)){
		
		#get a possible parent prefix with a mask decreased by 1
		$possibleParentPrefix = $ipUtility->computeParentPrefix($prefix, 1);
		
		#if the parent does exist stop searching
		if (defined($routersHash->{$possibleParentPrefix})){
			$endLoop = 1;
		}
		
		$mask -= 1;
		$prefix = $possibleParentPrefix;
	}
	
#	printf("prefix = %s possible parent prefix = %s endLoop = %s\n",$childPrefix, $possibleParentPrefix, $endLoop );
	#if the parent is not found then post an error message. All the prefixes should have a parent in the RER level
	if ($endLoop==1){
		
		return $routersHash->{$possibleParentPrefix};
		
	}else{
		
		printf("(EE) There is no parent for the prefix %s!\n", $childPrefix);
		die;
	}
	
	
	
}

sub findParentFlat(){
	my ($self, $prefix, $rerMask) = @_;
	
	my $ipUtility = $self->{"ipUtility"};
	my $routersHash = $self->{"routersHash"};
	my $mask = $ipUtility->getMaskFromPrefix($prefix);
	
	my $possibleParentPrefix = $ipUtility->computeParentPrefix($prefix, $mask-$rerMask);
	
	if( defined($routersHash->{$possibleParentPrefix}) ){
		return $routersHash->{$possibleParentPrefix};
	}else{
		printf("(XT) There is no parent prefix for prefix %s. The possible parent prefix is %s\n", $prefix, $possibleParentPrefix)
				if($self->{"_DEBUG"}>=2);
		die;
	}
}

sub getRouterWithPrefix(){
	my $self = shift;
	my ($prefix) = @_;
	
	return $self->{"routersHash"}->{$prefix};
}

sub getRootLayer(){
	my $self=shift;
	
	return $self->{"rootLayer"};
}

sub firstEightBitsZero(){
	
	my $ip = shift;
	
	my @result = split(/\./, $ip);
	if ($result[0] == 0 ){
		return 1;
	}else{
		return 0;
	}
}

sub getRERLayer(){
	my $self=shift;
	
	return $self->{"rerLayer"};
}

#return 1 at the end
1