#this module may change depending on the type of input file we use, thus, as the interface to the main program should be fairly stable
#writing this independently seems only fair

package CoreSim::Topology::DataReadModule;

use strict;

sub new(){
	my $invocant = shift;
	my $filePath = shift;
	
	my $class = ref($invocant) || $invocant;
	
	my $self = {
		"filePath" => $filePath,
	};
	
	bless($self, $class);
	
}

#reads/parses data from the inter_pop_links.txt and it returns a list of structures built from this data. 
sub readIPLData(){
	my $self = shift;
	my $inputFilePath = shift;
	
	my $inputFile;
	
	if (defined($self->{"filePath"}->[0])){
		$inputFilePath = $self->{"filePath"}->[0];
	}
	
	my $delayMap;
	my $asBorderPOPs;
	my $popToASMapping;
	my $asDelayMap;
	
	$inputFile = &readFile($inputFilePath);
	
	foreach my $line (@$inputFile){
		
		#hash of hash that keeps the delay between 2 pops: 
		#delayMap->{POP1}{POP2} = latency
		if ( ($line->[4]!=9999) && ($line->[4] > 0) ){
			$delayMap->{$line->[0]}{$line->[2]} = $line->[4];
		}
		
		#hash that keeps for a given AS a list of POPs and for each POP the number of inter-as links that pass through it:
		#asBorderPOPs->{AS}->{POP} = inter_as_links. Also check this isn't an intra-as link
		if ($line->[1]!=$line->[3]){
			$asBorderPOPs->{$line->[1]}->{$line->[0]} += 1;
			$asBorderPOPs->{$line->[3]}->{$line->[2]} += 1;
		}
		
		$popToASMapping->{$line->[0]} = $line->[1];
		$popToASMapping->{$line->[2]} = $line->[3];
		$asDelayMap->{$line->[1]}{$line->[3]} = $line->[4];
		$asDelayMap->{$line->[3]}{$line->[1]} = $line->[4];
	}
	
	undef @$inputFile;
	
	#return the list of structures. If more are needed just append the list
	return ($delayMap,$asBorderPOPs, $popToASMapping, $asDelayMap);
	
}

#reads/parses data from the ip_to_pop_mapping.txt
sub readIPMData(){
	my $self = shift;
	my $inputFilePath = shift;
	
	my $ipPOPMapping;
	my $popIPMapping;
	
	if (defined($self->{"filePath"}->[1])){
		$inputFilePath = $self->{"filePath"}->[1];
	}
	
	#read the file (get a reference to an array of array)
	my $inputFile = &readFile($inputFilePath);
	
	foreach my $line (@$inputFile){
		$ipPOPMapping->{$line->[0]} = $line->[1];
		
		#create a list of IPs for every POP
		$popIPMapping->{$line->[1]}->[ scalar( @{ $popIPMapping->{$line->[1]} } ) ] = $line->[0];
	 }
	
	undef @$inputFile;
	return ($ipPOPMapping,$popIPMapping);
}

#reads/parses data from the origin_as_mapping.txt
sub readOAData(){
	
	my $self = shift;
	my $inputFilePath = shift;
	my $prefixToAS;
	my $asToPrefix;
	
	if (defined($self->{"filePath"}->[2])){
		$inputFilePath = $self->{"filePath"}->[2];
	}
	
	#read the file (get a reference to an array of array)
	my $inputFile = &readFile($inputFilePath);
	
	foreach my $line (@$inputFile){
		
		#eliminate the casese when we have prefixes of the form <prefix> AS1_AS2 (just use AS1)
		if (index($line->[1], "_") != -1){
			my @tmp = split("_", $line->[1]);
			$line->[1] = $tmp[0];
		}
		
		$prefixToAS->{$line->[0]}=$line->[1];
		
		#there are more prefixes/AS, thus all the prefixes are enterned in a hash 
		$asToPrefix->{$line->[1]}->{$line->[0]}=1;
	}
	
	undef @$inputFile;
	
	return ($prefixToAS, $asToPrefix);
}

#read from a file the largest connected group of POPs(used in testing XMLRPC)
sub readPOPConnectedList(){
	my $self=shift;
	my $inputFilePath = shift;
	
	my $inputFile = &readFile($inputFilePath);
	my $connectedPOPList;
	
	foreach my $line (@$inputFile){
		for (my $i=0; $i< scalar(@$line); $i++){
			$connectedPOPList->[$i] = $line->[$i];
		}
	}
	
	undef @$inputFile;
	
	return($connectedPOPList);
	
}

#added such that any mapping system(for now Chord and ALT) can use the filtered prefix list
sub readFilteredPrefixList(){
	
	my $self = shift;
	my $inputFilePath;
	
	if ( defined( $self->{"filePath"}->[3]) ){
		$inputFilePath = $self->{"filePath"}->[3];
	}else{
		printf("(EE) There is a request to read the filtered prefix list but no input file is given!");
		return -1;
	}
	
	open(IFILE, $inputFilePath) or die printf("Can't read file %s : %s", $inputFilePath, $!);
	
	printf("Opening file: $inputFilePath\n");
	
	my @list;
	my $i=0;
	
	while(<IFILE>){
		chomp;
		$list[$i] = $_; 
		$i++;
	}
	
	close(IFILE);
	
	return \@list;
}

#method which reads a file and returns it as a structure (array of array)
sub readFile(){
	
	my $inputFilePath = shift;
	my $inputFile;
	
	printf("Opening file:$inputFilePath\n");
	open(IFILE, $inputFilePath) or die printf("Error opening $inputFilePath:%s\n",$!);
	
	my $count = 0;
	
	while (<IFILE>){
		my @line = split;
		#each line of the $inputFile array of array is one of the read lines
		$inputFile->[$count] = \@line; 
		$count++;	
	}
	
	close(IFILE);
	
	return $inputFile;
}

#writes the data in the structure delayMap to a file
sub dropIPLToFile(){
	my ($self, $delayMap, $outputFile) = @_;
	
	open(OFILE,"> $outputFile") or die printf("Error:%s\n",$!);
	
	for my $key1 (keys %{ $delayMap }){
		for my $key2 (keys %{ $delayMap->{$key1} } ){
					printf(OFILE "$key1 $key2\n");
		}
		
	}
	close(OFILE);
}

sub readPrefixToRIR(){
	my $self=shift;
	my $inputFile = shift;
	
	printf("Opening file: %s\n", $inputFile);
	open(IFILE, $inputFile);
	my $prefixToRIR = {};
	
	while(<IFILE>){
		chomp;
		my @line = split(/\s+/, $_);
		my @res = split(/\//, $line[0]);
		
		$prefixToRIR->{$res[0]} = $line[1];
	}
	return $prefixToRIR;
}

sub readRootServers(){
	my $self=shift;
	my $inputFile = shift;
	
	printf("Opening file: %s\n", $inputFile);
	open(IFILE, $inputFile);
	my $rootServers; 
	
	while(<IFILE>){
		chomp;
		@$rootServers = split(/\s+/, $_);
	}
	
	return $rootServers;	
}

sub DESTROY(){
	printf("The DataReadModule is destroyed!\n");
}

#return 1 at the end
1