#finds the delays between all the POPs in the topology
package CoreSim::Topology::TopologyMetrics;

use strict;
use CoreSim::Utilities::Estimator; 
#use Frontier::Client;
use XMLRPC::Lite;   
use Math::Complex;

#constructor
sub new(){
	my $invocant = shift;
	my ($storedTopology,$xmlRPCServer, $dbiPlane, $dbEstimator,$geoLiteCityFile, $_DEBUG) = @_;
	
	my $delayMap = $storedTopology->{"delayMap"};
	my $class = ref($invocant) || $invocant;
	my $estimator = CoreSim::Utilities::Estimator->new($geoLiteCityFile);
	
	my $self = {
		"xmlRPCServer" => $xmlRPCServer,
		"storedTopology"=> $storedTopology,	
		"dbiPlane" => $dbiPlane,
		"dbEstimator" => $dbEstimator,
		"estimator" => $estimator,
		"_DEBUG" => $_DEBUG,
	};

	bless($self, $class);
	return $self;
	
}

#compute delay between 2 POPs using the inter_pop_links.txt file
sub getLatencyBetweenPOPs(){
	my ($self,$popIP1, $popIP2, $pop1, $pop2) = @_;
	
	my $localQuery;
	
	if ( !(defined($pop1) && defined($pop2)) ) {
		$pop1 = $self->{"storedTopology"}->getPOPForIP($popIP1);
		$pop2 = $self->{"storedTopology"}->getPOPForIP($popIP2);
	}
	
	if ( defined( $localQuery =  $self->{"storedTopology"}->getInterPOPDelayMap()->{$pop1}{$pop2} ) ){
		printf("The latency between the two POPS(%s, %s) is stored locally!\n", $pop1, $pop2) if($self->{"_DEBUG"});
		return $localQuery; 
	}elsif( defined($localQuery =  $self->{"storedTopology"}->getInterPOPDelayMap()->{$pop2}{$pop1} ) ){
		printf("The latency between the two POPS(%s, %s) is stored locally!\n", $pop1, $pop2) if($self->{"_DEBUG"});
		return $localQuery;
	}else{
		printf("The latency between the two POPs(%s, %s) is not stored locally! Using XMLRPC\n", $pop1, $pop2) if($self->{"_DEBUG"});

		my $xmlRPCQuery = &xmlRPCQuery($self, $popIP1, $popIP2);
		return $xmlRPCQuery;
	}
}

sub getLatencyBetweenIPs(){
	my $self = shift;
	my ($ip1, $ip2) = @_;
	
	if((defined(my $pop1 = $self->{"storedTopology"}->getPOPForIP($ip1))) && (defined(my $pop2 = $self->{"storedTopology"}->getPOPForIP($ip2)))){
		printf("The two IPs(%s, %s) are associated to POPs. Trying to solve query by computing the latency between any IPs of the two POPs\n", 
				$ip1, $ip2) if( $self->{"_DEBUG"});
		return &getLatencyBetweenPOPIPs1($self, $ip1, $ip2);
	}else{
		printf("The two IPs(%s, %s) are not associated to POPs. An XMLRPC query is performed\n", $ip1, $ip2) if ($self->{"_DEBUG"});
		return &xmlRPCQuery($self, $ip1, $ip2);
	}
}

sub getLatencyBetweenPOPIPs(){
	
	my $self = shift;
	my ($ip1,$ip2) = @_; 
	
	my $pop1 = $self->{"storedTopology"}->getPOPForIP($ip1);
	my $pop2 = $self->{"storedTopology"}->getPOPForIP($ip2);

	return &getLatencyBetweenPOPs($self, $pop1, $pop2);
}

#called with 2 IPs which are known to be associated with stored POPs
sub getLatencyBetweenPOPIPs1(){
	
	my $self=shift;
	my ($ip1, $ip2) = @_;
	
	#do a basic check if the two IPs are identical (return 0) and try to solve it with the local database provided by the 
	#iPlane files
	
	#if the two IPs are equal
	if ($ip1 eq $ip2 ) {
		return 0;
	}
	
	my $pop1 = $self->{"storedTopology"}->getPOPForIP($ip1);
	my $pop2 = $self->{"storedTopology"}->getPOPForIP($ip2);
	
	my $popLatency;
	
	if (defined ($popLatency = $self->{"storedTopology"}->getInterPOPDelayMap()->{$pop1}{$pop2} ) ){
		printf("(gLBPI) solved locally!\n") if ($self->{"_DEBUG"}>=2);
		return $popLatency;
	}elsif( defined ($popLatency = $self->{"storedTopology"}->getInterPOPDelayMap()->{$pop2}{$pop1} ) ){
		printf("(gLBPI) solved locally!\n") if ($self->{"_DEBUG"}>=2);
		return $popLatency;
	}
	
	#now two methods can be used to interogate the iPlane XMLRPC server: singlePairInQuery of multiplePairsInQuery. For now
	#I use multiplePairsInQuary becaue it is faster. If it happends that the iPlane server timeouts it is recommended to decrease the
	#number of pairs per query or use the singlePairPerQuery method
	return &multiplePairsInQuery($self, $ip1, $ip2);
	

}

sub multiplePairsInQuery(){
	my ($self, $ip1, $ip2) = @_;
	
	my $estimator = $self->{"estimator"};
	my $dbiPlane = $self->{"dbiPlane"};
	my $dbEstimator = $self->{"dbEstimator"};
	
	my $pop1 = $self->{"storedTopology"}->getPOPForIP($ip1);
	my $pop2 = $self->{"storedTopology"}->getPOPForIP($ip2);
	my $ipList1 = $self->{"storedTopology"}->getIPListForPOP($pop1);
	my $ipList2 = $self->{"storedTopology"}->getIPListForPOP($pop2);
	
	my $latency;
	my $nbOfQueries;
	
	#query iPlane.db to see if the result is stored
	my $latencyQueryResult = $dbiPlane->getLatency($ip1, $ip2);
	
	return $latencyQueryResult if(defined($latencyQueryResult));
	
	#query estimatorDB next to avoid calling xmlRPC for pairs of IPs for which it does not provide a result
	$latencyQueryResult = $dbEstimator->getLatency($ip1, $ip2);
	
	return $latencyQueryResult if(defined($latencyQueryResult));
	
		
	#next query the xmlRPC. I use the eval to avoid failures of the simulator when xmlRPC fails
	eval{
		($latencyQueryResult, $nbOfQueries) = &xmlRPCComplexQuery1($self, $ipList1, $ipList2, $ip1, $ip2);
		printf("latency = %s for the pair %s %s\n", $latencyQueryResult, $ip1, $ip2) if ($self->{"_DEBUG"}>=3);
		#if the latency is defined
		if ( defined($latencyQueryResult) ){
		
			#and latency is valid
			if( ($latencyQueryResult!=9999 )&&($latencyQueryResult>=0) ){
				
				#add it to the database
				printf("add the latency %s between %s and %s to the iPlane.db\n",$ip1, $ip2, $latencyQueryResult) if ($self->{"_DEBUG"});
				$dbiPlane->addLatency($ip1, $ip2, $latencyQueryResult);
			
			#if the latency is invalid
			}else{
				
				#undef the latency such that the program will pass to the estimator
				printf("...undef the invalid latency\n") if ($self->{"_DEBUG"}>=2);
				$latencyQueryResult = undef;
			}
		}	
	};
	
	return $latencyQueryResult if(defined($latencyQueryResult));
	
	#if everything fails use the estimator 
	printf("will estimate the latency...") if ($self->{"_DEBUG"}>=2);
	
	my $time1 = time;
	$latencyQueryResult = $estimator->getLatency($ip1, $ip2);
	my $time2 = time;
	
	printf("..done in % seconds\n", $time2-$time1) if ($self->{"_DEBUG"}>=2);
	
	$dbEstimator->addLatency($ip1, $ip2, $latencyQueryResult);
	
	return $latencyQueryResult;	
	
}

#old method used to introgate iPlane with just 1 pair per query. This exists due to sometimes failure of the iPlane server to deal 
#with a high number of pairs per query 
sub singlePairInQuery(){
	my ($self, $ip1, $ip2) = @_;
	
	my $pop1 = $self->{"storedTopology"}->getPOPForIP($ip1);
	my $pop2 = $self->{"storedTopology"}->getPOPForIP($ip2);
	my $ipList1 = $self->{"storedTopology"}->getIPListForPOP($pop1);
	my $ipList2 = $self->{"storedTopology"}->getIPListForPOP($pop2);
	
	my $dbiPlane = $self->{"dbiPlane"};
	my $numberOfQueries=0;
	my $latencyQuery;
	

	if ($self->{"_DEBUG"}>=4){
		
		return &testLatencies($self, $ipList1, $ipList2, $pop1, $pop2);
		
	}else{
		
		#test to see if the latency between any 2 IPs asociated with the POPs is defined 
		foreach my $myIP1 (@{$ipList1}){
			foreach my $myIP2(@{$ipList2}){
			
				#first try to see if the latency is stored in the local database
				my $dbQuery = $dbiPlane->getLatency($myIP1, $myIP2);
				
				if ( defined( $dbQuery ) ){
					
					return $dbQuery;
				
				#try to solve it by using xmlrpc
				}else{
		
					$latencyQuery = &xmlRPCQuery($self, $myIP1, $myIP2);
					$numberOfQueries++; 
					
					#if the query failed, try reversing the IPs
					if ($latencyQuery == -1){
						$latencyQuery = &xmlRPCQuery($self, $myIP2, $myIP1);
						$numberOfQueries++;	
					}
					
					printf("queries for %s %s. The current count %s\n",$ip1, $ip2,  $numberOfQueries);
					
					
					#if a valid latency is found add it to the database
					if (($latencyQuery != -1)&&($latencyQuery!=9999)){
						printf("For IPs (%s,%s) associated with pops (%s,%s) we found the pair of IPs (%s,%s) between which the latency is defined!\n",	
								$ip1,$ip2,$pop1, $pop2, $myIP1, $myIP2) if($self->{"_DEBUG"});
						printf("Success after %s queries!\n", $numberOfQueries) if($self->{"_DEBUG"});	
						
						$dbiPlane->addLatency($myIP1, $myIP2, $latencyQuery);
							
						return $latencyQuery;
					}
				}
				
			}
		}
		
	}
	
	#if all the queries failed return -1
	return -1;
}

sub xmlRPCQuery(){
	
	my $self = shift;
	my ($src, $dst) = @_;
	
	my %query = ("src" => $src, "dst" => $dst );
	
	my @paths = ();
	push(@paths, \%query);
	
	my $url = $self->{"xmlRPCServer"};
	
	
#	my $client = Frontier::Client->new(
#			url => $url,
#			debug => 0, 
#	);
#	my @results = $client->call( 'iplane.query', @paths);	
	
	printf("This is a XMLRPC Query for latency between ip1=%s, ip2=%s\n", $src, $dst) if($self->{"_DEBUG"}>=3);
	
	
	
#####

	my @results;
	  
  	@results= XMLRPC::Lite
      -> proxy($url)
      -> call('iplane.query', @paths)
      -> result;
###

	if(!defined($results[0])) {
		printf("The result array is undefined");
	}
	
	
	foreach my $result ( @{$results[0]} ){
		my %result_hash = %{$result};
		my $src = $result_hash{"src"};
		my $dst = $result_hash{"dst"};
		my $meas_or_pred = $result_hash{"meas_or_pred"};
		my $latency = $result_hash{"latency"};
		my $loss = $result_hash{"loss"};
		my $aspath = $result_hash{"as_path"};
		my $clusterpath = $result_hash{"cluster_path"};
		my $path = $result_hash{"path"};
	
		if ($latency != 9999){
			return $latency;
		}
		
	}
		
	printf("The XMLRPC query result is not defined!\n") if($self->{"_DEBUG"}>=2);
	return -1;	
			
}

sub xmlRPCComplexQuery1(){
	my $self= shift;
	my ($ipList1, $ipList2, $sIP1, $sIP2) = @_;
	
	my $url = $self->{"xmlRPCServer"};
	
	my @paths=();
	my $pairsInQuery=0;
	my $pairsInQueryLimit = 200;
	my $nbOfQueries=0;
	
	#through this I limit the number of queries
	my $queryHighLimit = 200; 
	
	foreach my $ip1 (@{$ipList1}){
		foreach my $ip2(@{$ipList2}){
			
				my %path1 = ("src"=>$ip1, "dst"=>$ip2);
				my %path2 = ("src"=>$ip2, "dst"=>$ip1);
				push(@paths,\%path1);
				push(@paths,\%path2);
				
				$pairsInQuery+=2;
				
				#if we reach the limit for the paris in a query stop adding IPs and make the query
				if ($pairsInQuery == $pairsInQueryLimit){
					
					my ($latency, $nbOfQueriesL) = &xmlRPCIPPathsQuery($url, \@paths);
					$nbOfQueries += $pairsInQuery;
					
					if ( defined($latency) ){
						return($latency, $nbOfQueries);
					}
					
					#stop the querying when the number of queries has reached 10k (better estimate the latency than wait)
					if ($nbOfQueries >= $queryHighLimit){
						return(undef, $nbOfQueries);
					}
					
					$pairsInQuery = 0;
					@paths=();
				}
			
		}
	}
	
	#for the last IPs which haven't sumed up to pairsInQuery;
	my ($latency, $nbOfQueriesL) = &xmlRPCIPPathsQuery($url, \@paths);
	$nbOfQueries += $nbOfQueriesL;
	
	return ($latency, $nbOfQueries);
	
}

sub xmlRPCIPPathsQuery(){
	my ($url, $paths ) = @_;
	
	my $nbOfQueries = 0;
	my @results= XMLRPC::Lite
	      -> proxy($url)
	      -> call('iplane.query', @$paths)
	      -> result;     
	    
	#check the results list
	foreach my $result ( @{ $results[0] } ){
		my %result_hash = %{$result};
		my $latency = $result_hash{"latency"};
		my $src = $result_hash{"src"};
		my $dst = $result_hash{"dst"};
			
		$nbOfQueries++;
			
		#if the latency is defined add it to the database and return the result
		if(defined($latency)){
					
				#end the search
				return ($latency, $nbOfQueries);
		}	

	}
	return (undef, $nbOfQueries);
	
}

sub testLatencies(){
	
	my ($self, $ipList1, $ipList2, $pop1, $pop2) = @_;
	my @queryAnswers=();
	my $sumOfLatencies;
	
	my $numberOfQueries = 0;
	#test to see if the latency between any 2 IPs asociated with the POPs is defined 
	foreach my $myIP1 (@{$ipList1}){
		foreach my $myIP2(@{$ipList2}){
			
			$numberOfQueries++; 
			my $latencyQuery = &xmlRPCQuery($self, $myIP1, $myIP2);
			if ($latencyQuery != -1){
				$sumOfLatencies +=$latencyQuery;
				push(@queryAnswers, $latencyQuery);
			}
		}
	}
	
	my $numberOfAnswerdQueries = scalar(@queryAnswers);
	
	if ($numberOfAnswerdQueries != 0){
		#compute the variance for the results;
		my $mean = $sumOfLatencies/ $numberOfAnswerdQueries ;
		my $variance =0;
		
		foreach my $query (@queryAnswers){
			$variance += ( ($query - $mean)**2 )/ $numberOfAnswerdQueries;	
		}
		printf("For the latency between pop %s and pop %s %s queries have been answered from a total of %s queries. The standard deviation is %s\n",
				$pop1, $pop2, $numberOfAnswerdQueries, $numberOfQueries, sqrt($variance) ) if ($self->{"_DEBUG"}>=3);
		printf("The latencies: @queryAnswers\n") if ($self->{"_DEBUG"}>=3);
		return($queryAnswers[0])
	}else{
		
		#in case of failure to find the latency return -1
		printf("For the latency between pop %s and pop %s %s queries failed!\n",$pop1, $pop2, $numberOfQueries);
		return -1;	
	}
		
}

sub xmlRPCComplexQuery(){
	my $self= shift;
	my ($ipList1, $ipList2, $sIP1, $sIP2) = @_;
	
	my $url = $self->{"xmlRPCServer"};
	
	my $dbi = $self->{"dbiPlane"};
	
	my @paths=();
	my $pairsInQuery=0;
	my $pairsInQueryLimit = 200;
	my $nbOfQueries=0;
	
	#flag used to decide if invalid latencies are stored in the local database or not. If they are stored but the flag is 0 the program 
	#will try to resolve them again
	my $storeInvalidLatencies =0;
	
	foreach my $ip1 (@{$ipList1}){
		foreach my $ip2(@{$ipList2}){
			
			my $time1 = time;
			#first try to see if the latency is stored in the local database
			my $dbQuery = $dbi->getLatency($ip1, $ip2);
			
			my $time2 = time;
			
			printf("(XMLRPC-SQLite) The latency for the SQLite for the IPs %s and %s query is %s\n", $ip1, $ip2, $time2-$time1) 
					if ($self->{"_DEBUG"}>=2);
			
			if ($storeInvalidLatencies){
				
				#search for the latency in the database
				if ( defined( $dbQuery ) ){
					
					return ($dbQuery, $nbOfQueries);
				
				#if not found add the IP pair to the query
				}else{
				
					my %path1 = ("src"=>$ip1, "dst"=>$ip2);
					my %path2 = ("dst"=>$ip2, "dst"=>$ip1);
					push(@paths,\%path1);
					push(@paths,\%path2);
					
					$pairsInQuery+=2;
					
					#if we reach the limit for the paris in a query stop adding IPs and make the query
					if ($pairsInQuery == $pairsInQueryLimit){
						
						my @results= XMLRPC::Lite
					      -> proxy($url)
					      -> call('iplane.query', @paths)
					      -> result;     
					    
					    #check the results list
					    foreach my $result ( @{ $results[0] } ){
							my %result_hash = %{$result};
							my $latency = $result_hash{"latency"};
							my $src = $result_hash{"src"};
							my $dst = $result_hash{"dst"};
							
							$nbOfQueries++;
							
							#if the latency is defined add it to the database and return the result
							if(defined($latency)){
			
									$dbi->addLatency($sIP1, $sIP2, $latency);
									
									if( ( ($sIP1 ne $src) || ($sIP2 ne $dst) ) && (( $sIP1 ne $dst) || ($sIP2 ne $src)) ){
										$dbi->addLatency($src, $dst, $latency);
									}
									
									#end the search
									return ($latency, $nbOfQueries);
							}
							
				
						}
						
						$pairsInQuery = 0;
						@paths=();
					}
					
				}
				
				#for the last IPs which haven't sumed up to pairsInQuery;
				my @results= XMLRPC::Lite
			      -> proxy($url)
			      -> call('iplane.query', @paths)
			      -> result; 
			     
			    #check the results 	
				foreach my $result ( @{ $results[0] } ){
					my %result_hash = %{$result};
					my $latency = $result_hash{"latency"};
					my $src = $result_hash{"src"};
					my $dst = $result_hash{"dst"};
					
					$nbOfQueries++;
					if(defined($latency)){
						
							$dbi->addLatency($sIP1, $sIP2, $latency);
							
							if( ( ($sIP1 ne $src) || ($sIP2 ne $dst) ) && (( $sIP1 ne $dst) || ($sIP2 ne $src)) ){
								$dbi->addLatency($src, $dst, $latency);
							}
									
							
							return ($latency, $nbOfQueries);
					}
				}
				
				$dbi->addLatency($sIP1, $sIP2, -1);
				return (-1, $nbOfQueries);
				
			
			#(do not use stored invalid latencies)even if the latency is stored , if it is invalid try to "solve" it again
			}else{
				
				if ( defined( $dbQuery )&&( $dbQuery!=-1)&&($dbQuery!=-9999)&&($dbQuery!=9999) ){
					
					return ($dbQuery, $nbOfQueries);
					
				}else{
				
					my %path1 = ("src"=>$ip1, "dst"=>$ip2);
					my %path2 = ("dst"=>$ip2, "dst"=>$ip1);
					push(@paths,\%path1);
					push(@paths,\%path2);
					
					$pairsInQuery+=2;
					
					#if we reach the limit for the paris in a query stop adding IPs and make the query
					if ($pairsInQuery == $pairsInQueryLimit){
						
						my @results= XMLRPC::Lite
					      -> proxy($url)
					      -> call('iplane.query', @paths)
					      -> result;     
					    
					    #check the results list
					    foreach my $result ( @{ $results[0] } ){
							my %result_hash = %{$result};
							my $latency = $result_hash{"latency"};
							my $src = $result_hash{"src"};
							my $dst = $result_hash{"dst"};
							
							$nbOfQueries++;
							
							#if the latency is defined add it to the database and return the result
							if(defined($latency)){
								if (($latency != -1)&&($latency!=9999)&&($latency!=-9999)){
									$dbi->addLatency($src, $dst, $latency);
									$dbi->addLatency($sIP1, $sIP2, $latency);
									return ($latency, $nbOfQueries);
								}
							}
							
				
						}
						
						$pairsInQuery = 0;
						@paths=();
					}
				}
				
				
				#for the last IPs which haven't sumed up to pairsInQuery;
				my @results= XMLRPC::Lite
			      -> proxy($url)
			      -> call('iplane.query', @paths)
			      -> result; 
			      	
				foreach my $result ( @{ $results[0] } ){
					my %result_hash = %{$result};
					my $latency = $result_hash{"latency"};
					my $src = $result_hash{"src"};
					my $dst = $result_hash{"dst"};
					
					$nbOfQueries++;
					if(defined($latency)){
						
						if (($latency != -1)&&($latency!=9999)&&($latency!=-9999)){
							$dbi->addLatency($src, $dst, $latency);
							$dbi->addLatency($sIP1, $sIP2, $latency);
							return ($latency, $nbOfQueries);
						}
					}
					
				}	
				
				return (-1, $nbOfQueries);
			}
			
			
		}
	}
	

	
}

#return 1
1
